package book;

import java.io.*;

/**
 * Escape sequences for ANSI terminal outputs.
 * 
 * <p>This class, meant to be inherited or used statically, defines character
 * constants containing ANSI escape sequence that changes the color or
 * behaviour of some terminals.</p>
 *
 * <p>TODO: make methods instead of constants that allow to remove coloring if
 * the terminal is not ANSI.</p>
 *
 * @author Antoine Dutot
 * @since 19981107 creation (20000620 Passed from C++ to Java, 20010215, 20040529)
 */
public class AnsiTerminal
{
// Character values
	
	/**
	 * Reset the terminal in normal mode.
	 */
	public static final String RESET  = "[0m";
	
	/**
	 * Idem to `RESET' and also echo a new line.
	 */
	public static final String RESET_NL  = "[0m\n";

// Character style

	/**
	 * Bold mode.
	 */
	public static final String BOLD  = "[1m";
	
	/**
	 * Unserscore mode.
	 */
	public static final String UNDERSCORE  = "[4m";
	
	/**
	 * Reversed mode (background becomes foreground and vice &amp; versa).
	 */
	public static final String STANDOUT  = "[7m";

// Foreground colors

	/**
	 * Foreground color in black.
	 */
	public static final String BLACK  = "[30m";
	
	/**
	 * Foreground color in dark red.
	 */
	public static final String RED  = "[31m";
	
	/**
	 * Foreground color in dark green.
	 */
	public static final String GREEN  = "[32m";
	
	/**
	 * foreground color in dark yellow.
	 */
	public static final String YELLOW  = "[33m";
	
	/**
	 * foreground color in dark blue.
	 */
	public static final String BLUE  = "[34m";
	
	/**
	 * foreground color in dark magenta.
	 */
	public static final String MAGENTA  = "[35m";
	
	/**
	 * foreground color in dark cyan.
	 */
	public static final String CYAN  = "[36m";
	
	/**
	 * foreground color in dark white.
	 */
	public static final String WHITE  = "[37m";


// Background colors

	/**
	 * background color in black.
	 */
	public static final String BG_BLACK  = "[40m";

	/**
	 * background color in red.
	 */
	public static final String BG_RED  = "[41m";

	/**
	 * background color in green.
	 */
	public static final String BG_GREEN  = "[42m";

	/**
	 * background color in yellow.
	 */
	public static final String BG_YELLOW  = "[43m";

	/**
	 * background color in blue.
	 */
	public static final String BG_BLUE  = "[44m";

	/**
	 * background color in magenta.
	 */
	public static final String BG_MAGENTA  = "[45m";

	/**
	 * background color in cyan.
	 */
	public static final String BG_CYAN  = "[46m";

	/**
	 * background color in white.
	 */
	public static final String BG_WHITE  = "[47m";


// Light Foreground colors

	/**
	 * Foreground color in light black.
	 */
	public static final String LIGHT_BLACK  = "[30;1m";

	/**
	 * Foreground color in light red.
	 */
	public static final String LIGHT_RED  = "[31;1m";

	/**
	 * Foreground color in light green.
	 */
	public static final String LIGHT_GREEN  = "[32;1m";

	/**
	 * Foreground color in light yellow.
	 */
	public static final String LIGHT_YELLOW  = "[33;1m";

	/**
	 * Foreground color in light blue.
	 */
	public static final String LIGHT_BLUE  = "[34;1m";

	/**
	 * Foreground color in light magenta.
	 */
	public static final String LIGHT_MAGENTA  = "[35;1m";

	/**
	 * Foreground color in light cyan.
	 */
	public static final String LIGHT_CYAN  = "[36;1m";

	/**
	 * Foreground color in light white.
	 */
	public static final String LIGHT_WHITE  = "[37;1m";

// Cursor movements

	/**
	 * Store the current cursor position. Does not work with all ANSI terminals.
	 */
	public static final String SAVE_CURSOR_POS = "[s";

	/**
	 * Restore the current cursor position. Does not work with all ANSI terminals.
	 */
	public static final String RESTORE_CURSOR_POS = "[u";

	/**
	 * Remove characters from the current cursor position until the end of the line.
	 */
	public static final String ERASE_LINE_END = "[K";

	/**
	 * Clear the terminal.
	 */
	public static final String CLEAR_SCREEN = "[2J";

// Accessors

	/**
	 * Clear the terminal.
	 */
	public static void
	clear( PrintStream out )
	{
		out.print( CLEAR_SCREEN );
	}

	/**
	 * Same as using {@link #ERASE_LINE_END} with <code>out</code>.
	 * @param out The PrintStream to use for output.
	 */
	public static void
	killEndOfLine( PrintStream out )
	{
		out.print( ERASE_LINE_END );
	}

	/**
	 * Move the cursor backward of <code>cols</code> columns. This does not
	 * erase the line.
	 * @param out The PrintStream to use for output.
	 * @param cols The number of columns to move the cursor backward.
	 */
	public static void
	moveCursorBackward( PrintStream out, int cols )
	{
		out.print( "["+cols+"D" );
	}

	/**
	 * Move the cursor forward of <code>cols</code> columns. This does not
	 * erase the line.
	 * @param out The PrintStream to use for output.
	 * @param cols The number of columns to move the cursor forward.
	 */
	public static void
	moveCursorForward( PrintStream out, int cols )
	{
		out.print( "["+cols+"C" );
	}

	/**
	 * Go to <code>column</code> of <code>out</code> assuming a given
	 * <code>terminal_width</code>. The next output characters will start at
	 * `column'.
	 * @param out The PrintStream to use for output.
	 * @param column The column to jump to.
	 * @param terminal_width The width of the terminal.
	 */
	public static void
	cursorToColumn( PrintStream out, int column, int terminal_width )
	{
		out.print ( "[" );
		out.print ( terminal_width );
		out.print ( "C[$" );
		out.print ( column );
		out.print ( "D" );
	}
}
