package book.geom;

import book.InvalidValueException;

/**
 * Model of an orbiting camera.
 *
 * <p>This is a model of a camera, not a plain camera. It is not designed for
 * any kind of output system (it was tested with OpenGL and RenderMan). It can
 * use two conventions for coordinate systems, right-handed and left-handed (in
 * fact, this impacts only on the interpretation of angles actually).</p>
 *
 * <p>As a model, it stores informations about the camera:</p><ul>
 *	<li>A cartesian position of the focal point (called the "eye").</li>
 *	<li>A cartesian position of a looked-at point (called the "center" ).</li>
 *	<li>A up vector indicating where is the top, and allowing to bank the
 *	camera.</li>
 *	<li>A near clipping plane indicating the position of the camera lens with
 *	respect to the focal point (eye).</li>
 *	<li>A field of view angle (fov) indicating to aperture of the camera
 *	lens.</li>
 *	<li>An aspect ratio indicating the proportions of the lens rectangle.</li>
 * </ul>
 *
 * <p>The camera always track the center point. If you want to move the camera
 * without changing its orientation, move the center point of the same amount
 * as the camera.</p>
 *
 * @author Antoine Dutot
 * @since 200001008
 * @version 0.1
 */
public class Camera
{
// Types

	public static final boolean LEFT_HANDED  = false;
	public static final boolean RIGHT_HANDED = true;

// Attributes
	
	/**
	 * If true, the system is right handed. Right handed is the OpenGL
	 * convention (and the usual mathematical one) where the positive Z axis
	 * points toward the viewer, the positive X axis to the right and the
	 * positive Y axis to the top. The left handed coordinate system has the
	 * same X and Y axes but the positive Z axis points inside the screen.
	 *
	 * This setting affects several commands: #set_eye(), #set_eye_polar(),
	 * #set_center() and #get_view_matrix().
	 */
	protected boolean rh;

	/**
	 * Position of the camera focal point.
	 */
	protected Vector3 eye;
		
	/**
	 * Looked at point.
	 */
	protected Vector3 ctr;
		
	/**
	 * Vector indicating where is the top. 
	 */
	protected Vector3 upv;
	
	/**
	 * Field of view (apperture) of the camera.
	 */
	protected float fov;
		
	/**
	 * Near clipping plane (position of the camera lens).
	 */
	protected float near;
	
	/**
	 * Far clipping plane.
	 */
	protected float far;
	
	/**
	 * Pixel aspect ratio. Given the height we calculate the width by
	 * (height*aspect). And given the width we calculate the height by
	 * (width/aspect).
	 */
	protected float aspect;

	/**
	 * Cache of the radius to speed up access. The radius is the distance
	 * between #eye and #ctr. Thus, it must be recomputed each time one of
	 * these change.
	 */
	protected float radius;

// Constructors

	/**
	 * New camera with default eye, center and up vectors. The camera uses a
	 * right handed coordinate system by default.
	 */
	public
	Camera()
	{
		this( true );
	}

	/**
	 * New camera with default eye, center and up vectors. You can choose to
	 * use a right_handed coordinate system or not.
	 */
	public
	Camera( boolean right_handed )
	{
		eye = new Vector3();
		ctr = new Vector3();
		upv = new Vector3();
		
		reset( right_handed );
	}
	
	/**
	 * Reset the camera to its default values. The camera at [0,1,1], the
	 * center at [0,0,0], the up vector along Y, the fov to 60, the near plane
	 * to 1, the far plane to 10000, the frame aspect ratio to 1.0, and the
	 * system as right_handed.
	 */
	public void
	reset( boolean right_handed )
	{
		eye.set( 0, 1, 1 );
		ctr.set( 0, 0, 0 );
		upv.set( 0, 1, 0 );
		
		recalc_radius();

		fov    = 60;
		near   = 1;
		far    = 1000;
		aspect = 1.0f;
		rh     = right_handed;
	}

// Commands -- Setters

	public void
	set_up_vector( float x, float y, float z )
	{
		upv.set( x, y, z );
	}
	
	/**
	 * Move the eye to the absolute cartesian coordinates (x,y,z).
	 */
	public void
	set_eye( float x, float y, float z )
	{
	    eye.set( x, y, z );
		recalc_radius();
	}
	
	/**
	 * Move the eye to the absolute polar coordinates
	 * (radius,theta,phi). theta is expressed in the XZ plane
	 * (horizontal when the up vector is [0,1,0]). Angles are given following
	 * the trigonometric direction. theta and phi are in degrees.
	 *
	 * If radius is negative, 180.0 is added to phi and theta. If theta
	 * or phi are larger than 2pi, they are reset to their corresponding
	 * angle inside 2pi.
	 */
	public void
	set_eye_polar( float radius, float theta, float phi )
	{
		float pr;	// radius in the XY plane
		int m;		// number of complete "360"

		// A negative radius reverses the direction of view.
	
		if( radius < 0 )
		{
			theta += 180;
			phi   += 180;
			radius = -radius;
		}
		
		// Handle multiples of PI.
	
		if( theta >= 360.0 )
		{
			m      = (int) ( theta / 360 );
			theta -= ( m * 360 );
		}
		
		if( phi >= 360.0 )
		{
			m    = (int) ( phi / 360 );
			phi -= ( m * 360 );
		}

		// If we are in a left handed coordinate system, the
		// angles are reversed

		if( ! rh )
		{
			theta = - theta;
			phi   = - phi;
		}

		phi   = (float) ( ( Math.PI / 180.0f ) * phi );		// to radians
		theta = (float) ( ( Math.PI / 180.0f ) * theta );
		
		// Position with repsect to origin and Y axis

		pr          = ((float) Math.cos( phi )) * radius;		// radius in the XZ plane
		eye.data[1] = ((float) Math.sin( phi )) * radius;		// Y
		eye.data[0] = ((float) Math.sin( theta )) * pr;		// X
		eye.data[2] = ((float) Math.cos( theta )) * pr;		// Z
		
		// Translate to the real center

		eye.add( ctr );
		this.radius = radius;
	}
	
	/**
	 * Move the center to the absolute cartesian coordinates (x,y,z).
	 */
	public void
	set_center( float x, float y, float z )
	{
		ctr.set( x, y, z );
		recalc_radius();
	}

// Commands -- View settings.
	
	/**
	 * Change the near and far clipping plane positions. This defines the front
	 * and the back, respectively, of the viewing pyramid.
	 */
	public void
	set_clip_planes( float near, float far ) throws InvalidValueException
	{
		if( near <= 0 )   throw new InvalidValueException( "near must be > 0" );
		if( near >= far ) throw new InvalidValueException( "near must be < far" );
	
		this.near = near;
		this.far  = far;
	}
	
	/**
	 * Change the "field of view" angle. The field of view is, in degrees,
	 * the apperture of the camera.
	 */
	public void
	set_fov_angle( float angle ) throws InvalidValueException
	{
		if( angle < 0 || angle >= 180 ) throw new InvalidValueException( "angle must be in [0..180]" );
	
		fov = angle;
	}
	
	/**
	 * Change the aspect ratio width/height. Thus for instance, given the
	 * height of the near plane, we can compute its width by mutliplying this
	 * height by ratio.
	 */
	public void
	set_aspect( float ratio ) throws InvalidValueException
	{
		if( ratio <= 0 ) throw new InvalidValueException( "ratio must be > 0" );
	
		aspect = ratio;
	}
	
	/**
	 * Set the fov angle and calculate the near plane and aspect. This
	 * method allows you to precisely know the dimensions of the view volume.
	 * The near plane will be of width units horizontally and of height
	 * units vertically. If an object is at near from the camera and it have
	 * these dimensions, it will fit exactly in the view.
	 *
	 * The width and height give the size of the whole screen. The screen
	 * will go from -width/2 to width/2 horizontally and from -height/2
	 * to height/2 vertically with (0,0) as the origin.
	 *
	 * The near plane will be calculated automatically from the fov angle
	 * (in degrees) and the desired height. The aspect will be set
	 * automatically from width/height. The far plane will be at near +
	 * depth, that is, the total depth of the view volume will be depth.
	 */
	public void
	set_view_volume( float angle, float width, float height, float depth )
		throws InvalidValueException
	{
		if( angle < 0 || angle >= 180 ) throw new InvalidValueException( "angle must be in [0..180]" );
		if( depth <= 0 )                throw new InvalidValueException( "depth must be > 0" );
	
		// The formula for converting degrees to radians is
		// ((PI/180)*angle) (in fact (((2*PI)/360)*angle)), but here we
		// need only the half of that angle thus we take (PI/360).
		//
		// tan ( angle ) = h / n, thus
		// n = h / tan ( angle )

		fov    = angle;
		near   = (float) ( ( height / 2.0 ) / Math.tan( ( Math.PI / 360.0 ) * fov ) );
		far    = near + depth;
		aspect = width / height;
	}

// Accessors -- Realization of the view and attributes settings

	/**
	 * Does the camera uses a right handed coordinate system? For example
	 * OpenGL uses this system whild RenderMan uses a left handed coordinate
	 * system by default.
	 */
	public boolean
	is_right_handed()
	{
		return rh;
	}

	/**
	 * Position of the camera.
	 */
	public Vector3
	get_eye()
	{
		return eye;
	}
	
	/**
	 * Looked at point.
	 */
	public Vector3
	get_center()
	{
		return ctr;
	}
		
	/**
	 * Vector indicating where is the top. 
	 */
	public Vector3
	get_up_vector()
	{
		return upv;
	}
	
	/**
	 * Field of view (apperture) of the camera.
	 */
	public float
	get_fov()
	{
		return fov;
	}
	
	/**
	 * Near clipping plane (distance of the camera lens from the eye).
	 */
	public float
	get_near_plane()
	{
		return near;
	}
	
	/**
	 * Far clipping plane.
	 */
	public float
	get_far_plane()
	{
		return far;
	}
	
	/**
	 * Pixel aspect ratio. This is always the ratio width / height, thus,
	 * given the height we calculate the width by (height*aspect), and given
	 * the width we calculate the height by (width/aspect).
	 */
	public float
	get_aspect_ratio()
	{
		return aspect;
	}
	
	/**
	 * Distance from the eye and the center.
	 */
	public float
	get_radius()
	{
		return radius;
	}
	
	/**
	 * Get the viewing pyramid base dimensions.
	 *
	 * <p>The attributes of the camera are the field of view, the near and far
	 * clipping planes and the aspect ratio. The projection matrix is thus
	 * usually set according to these settings.</p>
	 *
	 * <p>This method returns the size of the base of the frustum pyramid
	 * projection that is the size of screen. This is calculated using the fov
	 * angle, the near plane distance and the aspect ratio of the camera.</p>
	 *
	 * <p>Thus, usually you set the aspect ratio to the one of your output
	 * viewport (often this is the value height/width), and then make use of
	 * either near or fov to change the camera settings. Changing near zoom or
	 * unzoom on objects, changing fov change the perspective (smaller fov,
	 * give impressions of very large view, larger fov give impressions of very
	 * small cameras. This is because as the fov get smaller the rays are
	 * closer to be parallel).</p>
	 *
	 * <p>If you change both the near plane and the fov accordingly
	 * (proportionally), then the view will not seem to change (provided you
	 * move objects in the scene accordingly to stay at the same distance from
	 * the near plane).</p>
	 *
	 * <p>For example, with OpenGL, the frustum is for a camera instance:
	 * <code>( -wh[0], wh[0], -wh[1], wh[1], camera.near, camera.far
	 * )</code>.</p>
	 */
	public void
	get_frustum_base( float[] wh )
	{
		// Deducing the frustum from the fov, near/far and aspect settings:
		//
		// We only know the angle of aperture of the camera and the distance
		// from the camera to the near plane. We are looking for the size of
		// the near rectangle:
		//
		// <code>
		//          -Z
		//  \........:......../  <-- far plane
		//   \       :       /
		//    \      :      /
		//     \__b__:...../ <------ near plane,
		//      \  |_|    /          b is 1/2 of the length of
		//       \   n   /           the near plane
		//        \__|  /
		//         \a| /
		//          \|/
		//           O <------------ camera
		// </code>
		// 
		// The a angle is 1/2 of the aperture of the camera wich is the total
		// opening angle. It forms a rectangular triangle with the near
		// clipping plane. We know the distance from the camera to this near
		// clipping plane. b is half the length of the near plane, and is the
		// value we are searching. n is the near setting, the distance of
		// the near plane from the camera.
		// 
		// Therefore, knowing that tan(a)/b = n, we easily deduce that
		// b = tan(a)*n.
		//
		// As the tan routine of the C library uses radians, we must
		// convert a from degrees to radians. we know that 360 degrees
		// make 2PI radians. Thus we can deduce that
		// ( 2PI/360 * angle_in_degrees = angle_in_radians ). However a
		// is only 1/2 of the fov angle, thus the formula here is
		// ( PI/360 * angle_in_degrees ) = angle_in_radians.
		//
		// See the file "Mesa_top_dir/src-glu/glu.c" wich gives this
		// routine.

		float w, h;

		h = (float) (Math.tan( ( Math.PI / 360 ) * fov ) * near);
		w = h * aspect;

		wh[0] = w;
		wh[1] = h;
	}
	
	/**
	 * Compute the viewing transformation m to fit the camera location.
	 *
	 * <p>The resulting transformation is stored in m. Thus, the viewing
	 * transformation should be multiplied by m. The previous contents of m are
	 * not used. (OpenGL calls this matrix Modelview).</p>
	 *
	 * <p>These two settings are set with respect to the eye and ctr points
	 * indicating respectively the position of the viewer and the point in
	 * focus (looked at), and according to the upv vector indicating where is
	 * the top.</p>
	 */
	public void
	get_view_matrix( Matrix4 m )
	{
		// Aligning the camera on the view axis:
		//
		// We know one vector wich determines the direction of view: z. We
		// know another vector that determines the up direction: y. They
		// form a plane, thus we can compute a third vector, x,
		// perpendicular to this plane with a cross product. Using the new
		// x and z wich form a plane perpendicular to the plane defined by
		// y and z, we can compute anew y so that it is perpendicular to x
		// and z, and thus find an orthogonal matrix.
		//
		// We have thus determined an orthogonal matrix wich is the
		// rotation matrix to apply in order to orient the camera.
		// It remains to translate this matrix so that the viewing eye is
		// at the origin, et Voila !
		//
		// This is described in "Computer Graphics Principles and Practice"
		// Chapter 5.7 (page 220 in my edition). The gluLookAt routine
		// also does this job, and inspired me. I used the Mesa
		// implementation. See the file "Mesa_top_dir/src-glu/glu.c".
		//
		// N.B.: Y must not be colinear to Z, if this is the case, we must
		// add a small epsilon.
	
		Matrix4 mm = new Matrix4();
		Vector3 x  = new Vector3();
		Vector3 y  = (Vector3) upv.clone();
		Vector3 z  = new Vector3(
			( eye.data[0] - ctr.data[0] ),
			( eye.data[1] - ctr.data[1] ),
			( eye.data[2] - ctr.data[2] ) );

		z.normalize();
		x.crossProduct( y, z );
		y.crossProduct( z, x );
		x.normalize();
		y.normalize();

		// Set the translation:

		if( rh )
		{
			mm.setLine( 0, 1, 0, 0, -eye.data[0] );
			mm.setLine( 1, 0, 1, 0, -eye.data[1] );
			mm.setLine( 2, 0, 0, 1, -eye.data[2] );
			mm.setLine( 3, 0, 0, 0, 1 );
		}
		else
		{
			mm.setLine( 0, 1, 0, 0, eye.data[0] );
			mm.setLine( 1, 0, 1, 0, eye.data[1] );
			mm.setLine( 2, 0, 0, 1, eye.data[2] );
			mm.setLine( 3, 0, 0, 0, 1 );
		}

		// Set the rotation:

		m.setLine( 0, x.data[0], x.data[1], x.data[2], 0 );
		m.setLine( 1, y.data[0], y.data[1], y.data[2], 0 );
		m.setLine( 2, z.data[0], z.data[1], z.data[2], 0 );
		m.setLine( 3, 0,         0,         0,         1 );

		// Compose the translation with the rotation:

		m.mult( mm );

		if( ! rh )
		{
		//	m.set( 0, 2, - mm.get( 0, 2 ) );
		//	m.set( 1, 2, - mm.get( 1, 2 ) );
		//	m.set( 2, 2, - mm.get( 2, 2 ) );
		//	m.set( 3, 2, - mm.get( 3, 2 ) );
		}
	}

// Utility

	/**
	 * Recompute the distance from the center to the eye. The result is stored
	 * in radius and can be obtained via get_radius.
	 */
	protected void
	recalc_radius()
	{
		Vector3 v;
		
		v = (Vector3) eye.clone();
		v.sub( ctr );
	
		radius = v.length();
	}

// Misc.

	public String
	toString()
	{
		StringBuffer buf;
		
		buf = new StringBuffer( "Camera_[eye=" );
		buf.append( eye.toString() );
		buf.append( ":ctr=" );
		buf.append( ctr.toString() );
		buf.append( ":fov=" );
		buf.append( fov );
		buf.append( ":near=" );
		buf.append( near );
		buf.append( ":far=" );
		buf.append( far );
		buf.append( ":aspect=" );
		buf.append( aspect );
		buf.append( "]" );

		return buf.toString();
	}
/*
	public void
	to_stream( PrintWriter out )
	{
		out.print( "Camera_[eye=" );
		out.print( eye.toString() );
		out.print( ":ctr=" );
		out.print( ctr.toString() );
		out.print( ":fov=" );
		out.print( fov );
		out.print( ":near=" );
		out.print( near );
		out.print( ":far=" );
		out.print( far );
		out.print( ":aspect=" );
		out.print( aspect );
		out.print( ']' );
	}
*/
}
