package book.geom;

/**
 * <h2>Transformation that allows to orbit arround a point.</h2>
 *
 * @author Antoine Dutot
 * @since 20040520
 */
public class OrbitTransform
	extends MatrixTransform
{
// Attributes

	private static final long serialVersionUID = 12124342452524L;
	
	/**
	 * Position of the camera focal point.
	 */
	protected Vector3 eye;
		
	/**
	 * Looked at point.
	 */
	protected Vector3 ctr;
		
	/**
	 * Vector indicating where is the top. 
	 */
	protected Vector3 upv;

	/**
	 * Cache of the radius to speed up access. The radius is the distance
	 * between #eye and #ctr. Thus, it must be recomputed each time one of
	 * these change.
	 */
	protected float radius;

// Constructors

	public
	OrbitTransform()
	{
		eye = new Vector3();
		ctr = new Vector3();
		upv = new Vector3( 0, 1, 0 );
	}

	public void
	reset()
	{
		eye.set( 0, 1, 1 );
		ctr.set( 0, 0, 0 );
		upv.set( 0, 1, 0 );
		
		recalcRadius();
	}

// Commands -- Setters

	public void
	setUpVector( float x, float y, float z )
	{
		upv.set( x, y, z );
	}
	
	/**
	 * Move the eye to the absolute cartesian coordinates (x,y,z).
	 */
	public void
	setEye( float x, float y, float z )
	{
	    eye.set( x, y, z );
		recalcRadius();
	}
	
	/**
	 * Move the eye to the absolute polar coordinates
	 * (radius,theta,phi). theta is expressed in the XZ plane
	 * (horizontal when the up vector is [0,1,0]). Angles are given following
	 * the trigonometric direction. theta and phi are in degrees.
	 *
	 * If radius is negative, 180.0 is added to phi and theta. If theta
	 * or phi are larger than 2pi, they are reset to their corresponding
	 * angle inside 2pi.
	 */
	public void
	setEyePolar( float radius, float theta, float phi )
	{
		float pr;	// radius in the XY plane
		int m;		// number of complete "360"

		// A negative radius reverses the direction of view.
	
		if( radius < 0 )
		{
			theta += 180;
			phi   += 180;
			radius = -radius;
		}
		
		// Handle multiples of PI.
	
		if( theta >= 360.0 )
		{
			m      = (int) ( theta / 360 );
			theta -= ( m * 360 );
		}
		
		if( phi >= 360.0 )
		{
			m    = (int) ( phi / 360 );
			phi -= ( m * 360 );
		}

		// If we are in a left handed coordinate system, the
		// angles are reversed

//		if( ! rh )
//		{
//			theta = - theta;
//			phi   = - phi;
//		}

		phi   = (float) ( ( Math.PI / 180.0 ) * phi );		// to radians
		theta = (float) ( ( Math.PI / 180.0 ) * theta );
		
		// Position with repsect to origin and Y axis

		pr          = (float) Math.cos( phi ) * radius;		// radius in the XZ plane
		eye.data[1] = (float) Math.sin( phi ) * radius;		// Y
		eye.data[0] = (float) Math.sin( theta ) * pr;		// X
		eye.data[2] = (float) Math.cos( theta ) * pr;		// Z
		
		// Translate to the real center

		eye.add( ctr );
		this.radius = radius;
	}
	
	/**
	 * Move the center to the absolute cartesian coordinates (x,y,z).
	 */
	public void
	setCenter( float x, float y, float z )
	{
		ctr.set( x, y, z );
		recalcRadius();
	}

// Accessors -- Realization of the view and attributes settings

	/**
	 * Position of the camera.
	 */
	public Vector3
	getEye()
	{
		return eye;
	}
	
	/**
	 * Looked at point.
	 */
	public Vector3
	getCenter()
	{
		return ctr;
	}
		
	/**
	 * Vector indicating where is the top. 
	 */
	public Vector3
	getUpVector()
	{
		return upv;
	}
	
	/**
	 * Distance from the eye and the center.
	 */
	public float
	getRadius()
	{
		return radius;
	}
	
	/**
	 * (Re)Compute the viewing transformation to fit the camera location.
	 */
	public void
	recompute()
	{
		// Aligning the camera on the view axis:
		//
		// We know one vector wich determines the direction of view: z. We
		// know another vector that determines the up direction: y. They
		// form a plane, thus we can compute a third vector, x,
		// perpendicular to this plane with a cross product. Using the new
		// x and z wich form a plane perpendicular to the plane defined by
		// y and z, we can compute anew y so that it is perpendicular to x
		// and z, and thus find an orthogonal matrix.
		//
		// We have thus determined an orthogonal matrix wich is the
		// rotation matrix to apply in order to orient the camera.
		// It remains to translate this matrix so that the viewing eye is
		// at the origin, et Voila !
		//
		// This is described in "Computer Graphics Principles and Practice"
		// Chapter 5.7 (page 220 in my edition). The gluLookAt routine
		// also does this job, and inspired me. I used the Mesa
		// implementation. See the file "Mesa_top_dir/src-glu/glu.c".
		//
		// N.B.: Y must not be colinear to Z, if this is the case, we must
		// add a small epsilon.
	
		Matrix4 mm = new Matrix4();
		Vector3 x  = new Vector3();
		Vector3 y  = (Vector3) upv.clone();
		Vector3 z  = new Vector3(
			( eye.data[0] - ctr.data[0] ),
			( eye.data[1] - ctr.data[1] ),
			( eye.data[2] - ctr.data[2] ) );

		z.normalize();
		x.crossProduct( y, z );
		y.crossProduct( z, x );
		x.normalize();
		y.normalize();

		// Set the translation:

		mm.setLine( 0, 1, 0, 0, -eye.data[0] );
		mm.setLine( 1, 0, 1, 0, -eye.data[1] );
		mm.setLine( 2, 0, 0, 1, -eye.data[2] );
		mm.setLine( 3, 0, 0, 0, 1 );

		// Set the rotation:

		setLine( 0, x.data[0], x.data[1], x.data[2], 0 );
		setLine( 1, y.data[0], y.data[1], y.data[2], 0 );
		setLine( 2, z.data[0], z.data[1], z.data[2], 0 );
		setLine( 3, 0,         0,         0,         1 );

		// Compose the translation with the rotation:

		mult( mm );
	}

// Utility

	/**
	 * Recompute the distance from the center to the eye. The result is stored
	 * in radius and can be obtained via getRadius.
	 */
	protected void
	recalcRadius()
	{
		Vector3 v;
		
		v = (Vector3) eye.clone();
		v.sub( ctr );
	
		radius = v.length();
	}
}
