package book.geom;

/**
 * <h2>Definition of a projection.</h2>
 *
 * <p>Everything in the Sprites library is defined in 3D, but will at the end
 * be projected on a 2D plane that will itself be mapped to the viewport.</p>
 *
 * <p>A projection is defined by a projection type (orthographic or
 * perspective) and two points defining a viewable area. The lowest point
 * define the lowest visible (x,y) coordinate on the camera plane as well as
 * the near clipping plane. The highest point defines the highest (x,y)
 * coordinate on the camera plane as well as the far clipping plane.</p>
 *
 * <p>It is also possible to directly set a matrix for the projection.</p>
 *
 * <p>A projection defines a new space, as well as a new unit system. All
 * sub-sprites will have their locator expressed with respect to the
 * projection.</p>
 *
 * <p>A projection optionaly defines a viewport, else it uses the default
 * viewport.</p>
 *
 * @author Antoine Dutot
 * @since 20040110
 */
public class Projection
	extends Box3
{
// Constants

	private static final long serialVersionUID = -3523502542993870608L;

	/**
	 * Kind of projection.
	 */
	public static final int PROJ_OTHER = 0;

	/**
	 * Projection matrix if not to be computed automatically.
	 */
	public Matrix4 proj = null;

	/**
	 * Kind of projection.
	 */
	public static final int PROJ_ORTHOGRAPHIC = 1;

	/**
	 * Kind of projection.
	 */
	public static final int PROJ_PERSPECTIVE = 2;

// Attributes

	/**
	 * Projection type (otho or persp or other).
	 */
	protected int projtype = PROJ_PERSPECTIVE;

	/**
	 * Optional viewport.
	 */
	protected Viewport vp = null;

// Constructors

	public
	Projection()
	{
		this( -0.5f, -0.5f, -0.5f, 0.5f, 0.5f, 0.5f );
	}

	public
	Projection( float low_x, float low_y, float low_z, float hi_x, float hi_y, float hi_z )
	{
		super( low_x, low_y, low_z, hi_x, hi_y, hi_z );
		proj = null;
		projtype = PROJ_PERSPECTIVE;
	}

	public
	Projection( float low_x, float low_y, float low_z, float hi_x, float hi_y, float hi_z, Viewport vp )
	{
		super( low_x, low_y, low_z, hi_x, hi_y, hi_z );
		proj = null;
		projtype = PROJ_PERSPECTIVE;
		this.vp = new Viewport( vp );
	}

	public
	Projection( Viewport vp )
	{
		this( -0.5f, -0.5f, -0.5f, 0.5f, 0.5f, 0.5f, vp );
	}

	public
	Projection( Projection other )
	{
		lo.copy( other.lo );
		hi.copy( other.hi );

		if( other.vp != null )
		     vp = new Viewport( other.vp );
		else vp = null;
	}

// Commands

	/**
	 * Is this projection orthographic?.
	 */
	public boolean
	isOrthographic()
	{
		return( projtype == PROJ_ORTHOGRAPHIC );
	}

	/**
	 * Is this projection perspective?.
	 */
	public boolean
	isPerspective()
	{
		return( projtype == PROJ_PERSPECTIVE );
	}

	/**
	 * True if the projection defines a viewport.
	 */
	public boolean
	hasViewport()
	{
		return( vp != null );
	}

	/**
	 * Viewport of this projection or null if the projection uses the default
	 * viewport.
	 */
	public Viewport
	getViewport()
	{
		return vp;
	}

// Commands

	/**
	 * Specify a perspective projection.
	 */
	public void
	setProjectionPerspective()
	{
		projtype = PROJ_PERSPECTIVE;
	}

	/**
	 * Specify an orhtographic projection.
	 */
	public void
	setProjectionOrthographic()
	{
		projtype = PROJ_ORTHOGRAPHIC;
	}

	/**
	 * Copy the viewport <code>vp</code>.
	 */
	public void
	setViewport( Viewport vp )
	{
		if( this.vp == null )
		     this.vp = new Viewport( vp );
		else this.vp.copy( vp );
	}

	public void
	setViewport( int x, int y, int width, int height )
	{
		if( vp != null )
		     vp.set( x, y, width, height );
		else vp  = new Viewport( x, y, width, height );
	}

	/**
	 * Make this projection use the default viewport.
	 */
	public void
	unsetViewport()
	{
		this.vp = null;
	}

	/**
	 * Set the projection matrix <code>M</code> and its <code>type</code>. The
	 * type can be {@link #PROJ_PERSPECTIVE}, {@link #PROJ_ORTHOGRAPHIC} or
	 * {@link #PROJ_OTHER}. Pass null for <code>M</code> and {@link
	 * #PROJ_PERSPECTIVE} or {@link #PROJ_ORTHOGRAPHIC} to let the sprite
	 * compute itself the projection matrix. Passing {@link #PROJ_OTHER}
	 * with a null matrix is an error and has undefined behavior.
	 */
	public void
	setProjection( Matrix4 M, int type )
	{
		if( M == null )
		{
			assert( type != PROJ_OTHER );
			proj = M;
			projtype = type;
		}
		else
		{
			proj = M;
			projtype = type;
		}
	}

	public String
	toString()
	{
		if( vp != null )
		     return "[proj="+projtype+" lo("+lo.x+","+lo.y+","+lo.z+") hi("+hi.x+","+hi.y+","+hi.z+") vp("+vp.getX()+","+vp.getY()+"/"+vp.getWidth()+"x"+vp.getHeight()+")]";
		else return "[proj="+projtype+" lo("+lo.x+","+lo.y+","+lo.z+") hi("+hi.x+","+hi.y+","+hi.z+")]";
	}
}
