package book.gfx;

import book.nio.*;

/**
 * <h2>Base for images.</h2>
 *
 * <p>This class only defines informations on the image as well as the image
 * data itself. There is no provision for image loading and displaying.</p>
 *
 * @author Antoine Dutot
 * @since 20031218
 */
public class Image
{
	/**
	 * Image width in pixels.
	 */
	protected int width = -1;

	/**
	 * Image height in pixels.
	 */
	protected int height = -1;

	/**
	 * Size of a scaneline in bytes (including padding).
	 */
	protected int pitch = -1;

	/**
	 * Components per pixel (eg. 4 for RGBA).
	 */
	protected int cpp = -1;

	/**
	 * Bits per pixel (only 8, 16 or 32).
	 */
	protected int bpp = 8;

	/**
	 * Inde of the alpha channel component starting from zero, -1 indicate no
	 * alpha (eg. 3 for RGBA).
	 */
	protected int alpha = -1;

	/**
	 * Image data.
	 */
	protected ByteData data = null;

// Constructors

	/**
	 * New allocated image.
	 * @see #allocate(int,int,int,int,int)
	 */
	public
	Image( int width, int height, int cpp, int bpp, int alpha )
	{
		allocate( width, height, cpp, bpp, alpha, false );
	}

	/**
	 * Empty non allocated image.
	 */
	protected
	Image()
	{
	}

// Accessors

	/**
	 * Width in pixels.
	 */
	public int
	getWidth()
	{
		return width;
	}

	/**
	 * Height in pixels.
	 */
	public int
	getHeight()
	{
		return height;
	}

	/**
	 * Does the image has an alpha component?.
	 * @see #getAlpha()
	 */
	public boolean
	hasAlpha()
	{
		return( alpha >= 0 && alpha < cpp );
	}

	/**
	 * Image alpha component, or -1 if none. The alpha component is between
	 * zero and {@link #getComponentsPerPixel()}.
	 * @see #hasAlpha()
	 */
	public int
	getAlpha()
	{
		return alpha;
	}

	/**
	 * Number of components per pixel (red, green, blue, alpha, luminance,
	 * etc.).
	 */
	public int
	getComponentsPerPixel()
	{
		return cpp;
	}

	/**
	 * Number of bits per pixel. This returns only 8, 16 or 32.
	 */
	public int
	getBitsPerPixel()
	{
		return bpp;
	}

	/**
	 * Image data under the form of a native byte data buffer. This can be null
	 * if this is a descriptor only.
	 */
	public ByteData
	getData()
	{
		return data;
	}

	/**
	 * True if this image is a descriptor only, that do not store data.
	 */
	public boolean
	noStorage()
	{
		return ( data == null );
	}

// Commands

	/**
	 * Allocate a new image, replacing a possible old one.
	 * @param width The width in pixels of the new image.
	 * @param height The height in pixels of the new image.
	 * @param cpp Components per pixels (eg. 4 for RGBA).
	 * @param bpp Bits per pixel (supported values are 24, 32, 64, 128 only).
	 * @param alpha Index of the alpha component starting from zero (eg 3 in RGBA), if
	 * there is not alpha channel, -1 should be used. The pitch is set
	 * automatically.
	 * @param noStorage If true no storage is allocated for the image, only
	 * the description of the image is updated. This allows image descriptors
	 * like OpenGL textures for example.
	 * @see #wrap(int,int,int,int,int,int,ByteData)
	 */
	public void
	allocate( int width, int height, int cpp, int bpp, int alpha, boolean noStorage )
	{
		assert( width > 0 );
		assert( height > 0 );
		assert( cpp > 0 );
		assert( bpp > 0 );

		if( bpp != 24 && bpp != 32 && bpp != 64 && bpp != 128 )
		{
			throw new IllegalArgumentException(
				"Image actually only support bits-per-pixel sizes of 24, 32, 64 or 128 bits" );
		}

		this.width  = width;
		this.height = height;
		this.cpp    = cpp;
		this.bpp    = bpp;
		this.alpha  = alpha;

		if( alpha < 0 || alpha >= cpp )
			alpha = -1;

		// Compute the size in bytes.

		if( bpp == 24 )
		{
			assert( cpp == 3 );
			pitch = width * ( bpp / 8 );	// Does not take padding into account.
		}
		else
		{
			pitch = width * cpp * ( bpp / 32 );
		}

		// Allocate memory.

		if( noStorage )
		     data = null;
		else data = ByteData.allocateDirect( pitch * height );
	}

	/**
	 * Like {@link #allocate(int,int,int,int,int) but do not create a new byte
	 * data buffer, instead use the one given. The <code>pitch</code> is the
	 * size of a scanline in bytes including padding if desired. Be careful, no
	 * copy is made, this means that this can lead to two or more images
	 * sharing the same data. Never pass null for the data, instead use the
	 * noStorage argument of the allocate() method.
	 */
	public void
	wrap( int width, int height, int pitch, int cpp, int bpp, int alpha, ByteData data )
	{
		assert( width > 0 );
		assert( height > 0 );
		assert( pitch > 0 );
		assert( cpp > 0 );
		assert( bpp > 0 );
		assert( data != null );

		if( bpp != 32 && bpp != 64 && bpp != 128 )
			throw new IllegalArgumentException(
				"Image actually only support bits-per-pixel sizes of 8, 16 or 32 bits" );

		if( pitch < ( width * cpp * ( bpp / 32 ) ) )
			throw new IllegalArgumentException(
				"Wrong pitch "+pitch+", does not correspond to image width ("+width+") and cpp("+cpp+")/bpp("+bpp+")" );

		this.width  = width;
		this.height = height;
		this.pitch  = pitch;
		this.cpp    = cpp;
		this.bpp    = bpp;
		this.alpha  = alpha;

		if( alpha < 0 || alpha >= cpp )
			alpha = -1;

		// Compute the size in bits.

		int byteSize = pitch * height;

		if( byteSize > data.getCount() )
			throw new IllegalArgumentException(
				"The image should contain "+byteSize+" bytes but the data buffer is only "+data.getCount()+" bytes large..." );

		// Allocate memory.

		this.data = data;
	}
}
