package book.gfx;

/**
 * A RGBA color specification independant of any output medium.
 *
 * The four components Red, Green, Blue and Alpha are specified as floats
 * (real numbers with the largest precision in Java) on the range [0..1]. For
 * alpha, 0 is completly transparent an 1 is completly opaque.
 *
 * However, negative values or values larger than 1.0 are also allowsed, but
 * are generally used with the #add() method. When passing a Rgba instance to
 * another object you can use the #clamp() method to ensure that the components
 * effectively lie in the interval [0..1].
 *
 * @author Antoine Dutot
 * @since 20001220
 */
public class Rgba
	implements java.io.Serializable
{
// Attributes

	private static final long serialVersionUID = 1699066216318123573L;

	/**
	 * Red component.
	 */
	public float red;

	/**
	 * Green component.
	 */
	public float green;

	/**
	 * Blue component.
	 */
	public float blue;

	/**
	 * Alpha component.
	 */
	public float alpha;

// Constructors

	/**
	 * New opaque black RGBA color.
	 */
	public
	Rgba()
	{
		set( 0, 0, 0, 1 );
	}

	/**
	 * New opaque (red,green,blue) color.
	 */
	public
	Rgba( float red, float green, float blue )
	{
		set( red, green, blue, 1 );
	}

	/**
	 * New (red,green,blue,alpha) color.
	 */
	public
	Rgba( float red, float green, float blue, float alpha )
	{
		set( red, green, blue, alpha );
	}

	/**
	 * New copy of <code>other</code>.
	 */
	public
	Rgba( Rgba other )
	{
		set( other.red, other.green, other.blue, other.alpha );
	}

// Accessors

	public boolean
	equals( Object other )
	{
		if( other instanceof Rgba )
		{
			Rgba o = (Rgba) other;
			
			if( o.red == red && o.green == green && o.blue == blue && o.alpha == alpha )
				return true;
		}
	
		return false;
	}
	
	/**
	 * The red component.
	 */
	public float
	getRed()
	{
		return red;
	}

	/**
	 * The green component.
	 */
	public float
	getGreen()
	{
		return green;
	}

	/**
	 * The blue component.
	 */
	public float
	getBlue()
	{
		return blue;
	}

	/**
	 * The alpha component.
	 */
	public float
	getAlpha()
	{
		return alpha;
	}

// Commands

	/**
	 * Copy all components of <code>other</code> in this.
	 */
	public void
	copy( Rgba other )
	{
		red   = other.red;
		green = other.green;
		blue  = other.blue;
		alpha = other.alpha;
	}

	/**
	 * Set the (red,green,blue) color without modifying the alpha.
	 */
	public void
	set( float red, float green, float blue )
	{
		this.red   = red;
		this.green = green;
		this.blue  = blue;
	 //	this.alpha = 1.0;
	}

	/**
	 * Set the (red,green,blue,alpha) color.
	 */
	public void
	set( float red, float green, float blue, float alpha )
	{
		this.red   = red;
		this.green = green;
		this.blue  = blue;
		this.alpha = alpha;
	}

	/**
	 * Set only the red component.
	 */
	public void
	setRed( float red )
	{
		this.red = red;
	}

	/**
	 * Set only the green component.
	 */
	public void
	setGreen( float green )
	{
		this.green = green;
	}

	/**
	 * Set only the blue component.
	 */
	public void
	setBlue( float blue )
	{
		this.blue = blue;
	}

	/**
	 * Set only the alpha component.
	 */
	public void
	setAlpha( float alpha )
	{
		this.alpha = alpha;
	}

	/**
	 * Ensure that all components range in the interval [0..1].
	 */
	public void
	clamp()
	{
		if     ( red   > 1 ) red   = 1;
		else if( red   < 0 ) red   = 0;
		if     ( green > 1 ) green = 1;
		else if( green < 0 ) green = 0;
		if     ( blue  > 1 ) blue  = 1;
		else if( blue  < 0 ) blue  = 0;
		if     ( alpha > 1 ) alpha = 1;
		else if( alpha < 0 ) alpha = 0;
	}

	public void
	add( float red, float green, float blue, float alpha )
	{
		this.red   += red;
		this.green += green;
		this.blue  += blue;
		this.alpha += alpha;
	}

	public void
	add( Rgba other )
	{
		add( other.red, other.green, other.blue, other.alpha );
	}

	public String
	toString()
	{
		StringBuffer buf = new StringBuffer();
		buf.append( "[" );
		buf.append( Double.toString( red ) );
		buf.append( "," );
		buf.append( Double.toString( green ) );
		buf.append( "," );
		buf.append( Double.toString( blue ) );
		buf.append( "," );
		buf.append( Double.toString( alpha ) );
		buf.append( "]" );

		return buf.toString();
	}
}

// vim:ts=4:ai:
