package book.nio;

import java.util.*; import java.nio.*;

/**
 * <h2>Float array.</h2>
 *
 * @author Antoine Dutot
 * @since 20031217
 */
public abstract class FloatData
{
	/**
	 * Allocate a new buffer according to the current underlying machine.
	 * This method detects if the program is running under a JVM or was
	 * compiled using GCJ and creates the faster implementation possible.
	 */
	public static FloatData
	allocateDirect( int count )
	{
		Properties p = System.getProperties();
		String v = (String) p.get( "java.vendor" );

		if( count <= 0 )
			throw new IllegalArgumentException( "cannot allocate a FloatData with size <= 0" );

		if( v.startsWith( "Free Software" ) )
			return new FloatDataCni( count );

		return new FloatDataJvm( count );
	}

	/**
	 * Allocate a new buffer by wrapping existing data. This method works only
	 * when data is a java array of native types (byte, short, int, long,
	 * float, double) or a NIO direct buffer (FloatBuffer, ShortBuffer,
	 * IntBuffer, LongBuffer, FloatBuffer, DoubleBuffer).
	 */
	public static FloatData
	wrap( Object data )
	{
		if( data instanceof float[] )
		{
			return new FloatDataCni( (float[]) data );
		}
		else if( data instanceof ByteBuffer )
		{
			return new FloatDataJvm( (ByteBuffer) data );
		}
		else
		{
			throw new IllegalArgumentException(
				"Can only create a FloatData from an array of a FloatBuffer" );
		}
	}


	/**
	 * Used only internally, returns the buffer as seen by the native
	 * application code.
	 */
	public abstract Object
	getNative();

	/**
	 * Number of cell in this buffer.
	 */
	public abstract int
	getCount();

	/**
	 * Value of the <code>i</code>-th cell of the buffer.
	 */
	public abstract float
	get( int i );

	/**
	 * Set the <code>i</code>-th cell to the given <code>value</code>.
	 */
	public abstract void
	set( int i, float value );

	/**
	 * Copy the range <code>src_pos</code> to <code>src_pos+count</code> of
	 * <code>values</code> into this buffer placing data from
	 * <code>dest_pos</code> to <code>dest_pos+count</code>.
	 */
	public abstract void
	set( int dest_pos, int src_pos, int count, float values[] );
}
