package book.nio;

import java.util.*; import java.nio.*;

/**
 * <h2>Int array.</h2>
 *
 * @author Antoine Dutot
 * @since 20031217
 */
public abstract class IntData
{
	/**
	 * Allocate a new buffer according to the current underlying machine.
	 * This method detects if the program is running under a JVM or was
	 * compiled using GCJ and creates the faster implementation possible.
	 */
	public static IntData
	allocateDirect( int count )
	{
		Properties p = System.getProperties();
		String v = (String) p.get( "java.vendor" );

		if( count <= 0 )
			throw new IllegalArgumentException( "cannot allocate a IntData with size <= 0" );

		if( v.startsWith( "Free Software" ) )
			return new IntDataCni( count );

		return new IntDataJvm( count );
	}

	/**
	 * Allocate a new buffer by wrapping existing data. This method works only
	 * when data is a java array of native types (byte, short, int, long,
	 * float, double) or a NIO direct buffer (IntBuffer, ShortBuffer,
	 * IntBuffer, LongBuffer, FloatBuffer, DoubleBuffer).
	 */
	public static IntData
	wrap( Object data )
	{
		if( data instanceof int[] )
		{
			return new IntDataCni( (int[]) data );
		}
		else if( data instanceof ByteBuffer )
		{
			return new IntDataJvm( (ByteBuffer) data );
		}
		else
		{
			throw new IllegalArgumentException(
				"Can only create a IntData from an array of a IntBuffer" );
		}
	}


	/**
	 * Used only internally, returns the buffer as seen by the native
	 * application code.
	 */
	public abstract Object
	getNative();

	/**
	 * Number of cell in this buffer.
	 */
	public abstract int
	getCount();

	/**
	 * Value of the <code>i</code>-th cell of the buffer.
	 */
	public abstract int
	get( int i );

	/**
	 * Set the <code>i</code>-th cell to the given <code>value</code>.
	 */
	public abstract void
	set( int i, int value );

	/**
	 * Copy the range <code>src_pos</code> to <code>src_pos+count</code> of
	 * <code>values</code> into this buffer placing data from
	 * <code>dest_pos</code> to <code>dest_pos+count</code>.
	 */
	public abstract void
	set( int dest_pos, int src_pos, int count, int values[] );
}
