package book.set;

/**
 * Doubly linked node.
 *
 * A Bilinkable is the base of most doubly linked nodes. It only defines
 * references on the previous and next nodes in the bidirectionnal list. These
 * references can be null.
 *
 * An important property of this class is that all copy operations never copy
 * the #prev and #next attributes(that is, here, the copy makes nothing, in
 * heirs however other attributes may be copied in the usual way).
 *
 * @author Antoine Dutot
 * @since 19991022
 * @version 0.1
 */
public class Bilinkable
{
// Attributes
	
	/**
	 * Predecessor or null.
	 */
	public Bilinkable prev;

	/**
	 * Successor or null.
	 */
	public Bilinkable next;

// Constructors

	/**
	 * New node with no predecessor nor successor.
	 */
	public
	Bilinkable()
	{
		prev = null;
		next = null;
	}
	
	/**
	 * New node pointing on predecessor and on successor.
	 */
	public
	Bilinkable( Bilinkable predecessor, Bilinkable successor )
	{
		prev = predecessor;
		next = successor;
	}

// Accessors

	/**
	 * Predecessor or null.
	 */
	public Bilinkable
	get_previous()
	{
		return prev;
	}

	/**
	 * Successor or null.
	 */
	public Bilinkable
	get_next()
	{
		return next;
	}

// Commands
	
	/**
	 * Change the predecessor without updating it.
	 */
	public void
	set_previous( Bilinkable predecessor )
	{
		prev = predecessor;
	}

	/**
	 * Change the successor without updating it.
	 */
	public void
	set_next( Bilinkable successor )
	{
		next = successor;
	}
	
	/**
	 * Insert successor between this and the actual successor.
	 *
	 * This updates successor so that it points on the actual successor.
	 * Complexity: O(1).
	 */
	public void
	append( Bilinkable successor )
	{
		successor.next = this.next;
		successor.prev = this;
		this.next.prev = successor;
		this.next = successor;
	}
	
	/**
	 * Insert predecessor between this and the actual predecessor.
	 *
	 * This updates predecessor so that it points on the actual predecessor.
	 * Complexity: O(1).
	 */
	public void
	prepend( Bilinkable predecessor )
	{
		predecessor.next = this;
		predecessor.prev = this.prev;
		this.prev.next = predecessor;
		this.prev = predecessor;
	}
	
	/**
	 * Remove the actual successor and update the links.
	 *
	 * The current successor is lost, you must have backed it up, it is
	 * replaced by the node comming just after it.
	 * Complexity: O(1).
	 */
	public void
	remove_after()
	{
		next = next.next; 	// no need to test if next exists
		next.prev = this;	// in `BilinkedList', it can be `ring'
	}
	
	/**
	 * Remove the current predecessor and update the links.
	 *
	 * The current predecessor is lost, you must have backed it up, it is
	 * replaced by the node comming just before it.
	 * Complexity: O(1).
	 */
	public void
	prune_before()
	{
		prev = prev.prev; 
		prev.next = this;
	}
	
	
	/**
	 * Swap this with other.
	 */
	public void
	swap( Bilinkable other )
	{
		Bilinkable node;
		
		node = next;
		next = other.next;
		other.next = node;

		node = prev;
		prev = other.prev;
		other.prev = node;

		// TODO BUG?
		next.prev = this;
		prev.next = this;
		other.next.prev = other;
		other.prev.next = other;
	}
}

// vim:ts=4:ai:
