package book.set;

/**
 * Doubly linked list.
 * 
 * @author Antoine Dutot
 * @since 19980707
 * @version 0.1
 */
public class BilinkedList
{
	/**
	 * First and last node marking node.
	 * This node is not part of the ring, but is the only access point.
	 */
	protected Bilinkable ring;

// Constructors
	
	/**
	 * New empty bidirectionnal list.
	 */
	public
	BilinkedList()
	{
		ring = new Bilinkable();
		ring.next = ring;	// et cadedis!
		ring.prev = ring;	// On peut monter ainsi indefiniment.
	}

	/**
	 * New list of n copies of value.
	 */
	public
	BilinkedList( Object value, int n )
	{
		ring = new Bilinkable();
		ring.next = ring;
		ring.prev = ring;
		put_n( ring, n, value );
	}

	/**
	 * New copy of range [beg..end[.
	 */
	public
	BilinkedList( Iterator beg, Iterator end )
	{
		ring = new Bilinkable();
		ring.next = ring;
		ring.prev = ring;
		put_range_copy( ring, beg, end );
	}

	/**
	 * Make a copy of other.
	 */
	public
	BilinkedList( BilinkedList other )
	{
		ring = new Bilinkable();
		ring.next = ring;
		ring.prev = ring;
		put_range_copy( ring, other.ring.next, other.ring );
	}

// Accessors

	/**
	 * Is list empty?
	 */
	public boolean
	is_empty()
	{
		return( ring.next == ring );
	}

	/**
	 * Does iterator is dereferencable ?
	 */
	public boolean
	valid_iterator( Iterator iterator )
	{
		return( iterator != ring );
	}

	/**
	 * Number of items.
	 *
	 * Complexity: O(n), be careful! BilinkedList does not own a
	 * count attribute.
	 */
	public int
	get_count()
	{
		Bilinkable node;
		int n;
		
		node = ring.next;
		n = 0;

		while( node != ring )
		{
			n++;
			node = node.next;
		}

		return n;
	}

	/**
	 * First item.
	 *
	 * Complexity: O(1).
	 */
	public Object
	get_front()
	{
		return((ObjectBilinkable)ring.next).value;
	}
	
	/**
	 * Last item.
	 *
	 * Complexity: O(1).
	 */
	public Object
	get_back()
	{
		return((ObjectBilinkable)ring.prev).value;
	}
	
	/**
	 * Iterator on first item. end if empty list.
	 *
	 * Complexity: O(1).
	 */
	public BilinkedListIterator
	get_begin()
	{
		return new BilinkedListIterator( ring.next );
	}
	
	/**
	 * Iterator on past-the-last item. Not dereferencable.
	 *
	 * Complexity: O(1).
	 */
	public BilinkedListIterator
	get_end()
	{
		return new BilinkedListIterator( ring );
	}
	
	/**
	 * Set iterator to point on first item or get_end if empty list.
	 *
	 * Complexity: O(1).
	 */
	public void
	get_begin( BilinkedListIterator iterator )
	{
		iterator.set_node( ring.next );
	}
	
	/**
	 * Set iterator to point on past-the-last item.
	 *
	 * Complexity: O(1).
	 */
	public void
	get_end( BilinkedListIterator iterator )
	{
		iterator.set_node( ring );
	}

// Insertion

	/**
	 * Insert value before all items.
	 *
	 * Complexity: O(1).
	 */
	public void
	put_front( Object value )
	{ 
		ring.append( new ObjectBilinkable( value ) );
	}

	/**
	 * Insert value after all items.
	 *
	 * Complexity: O(1).
	 */
	public void
	put_back( Object value )
	{
		ring.prepend( new ObjectBilinkable( value ) );
	}

	/**
	 * Alias for #put_back().
	 */
	public void
	add( Object value )
	{
		ring.prepend( new ObjectBilinkable( value ) );
	}

	/**
	 * Insert value before pos.
	 *
	 * Complexity: O(1).
	 */
	public void
	put( BilinkedListIterator pos, Object value )
	{
		pos.node.prepend( new ObjectBilinkable( value ) );
	}
	
	/**
	 * Insert range [beg..end[ before pos.
	 *
	 * Complexity: O(n).
	 */
	public void
	put_range( BilinkedListIterator pos, Iterator beg, Iterator end )
	{
		put_range_copy( pos.node, beg, end );
	}
		
	/**
	 * Insert n copies of value before pos.
	 *
	 * Complexity: O(n).
	 */
	public void
	put_n  ( BilinkedListIterator pos, int n, Object value )
	{
		put_n( pos.node, n, value );
	}

// Commands

	/**
	 * Swap other and current contents.
	 *
	 * Complexity: O(1).
	 */
	public void
	swap( BilinkedList other )
	{
		Bilinkable node;
		
		node = ring;
		ring = other.ring;
		other.ring = node;
	}

	/**
	 * Move the entire list other before pos.
	 *
	 * Complexity: O(1) (yes its fast!).
	 */
	public void
	splice( BilinkedListIterator pos, BilinkedList other )
	{
		splice( pos.node, other.ring.next, other.ring );
	}
	
	/**
	 * Move item pointed to by i before pos.
	 *
	 * Complexity: O(1).
	 */
	public void
	splice( BilinkedListIterator pos, BilinkedListIterator i )
	{
		splice( pos.node, i.node, i.node.next );
	}
	
	/**
	 * Move range [beg..end[ before pos.
	 *
	 * Complexity: O(1).
	 */
	public void
	splice( BilinkedListIterator pos, BilinkedListIterator beg, BilinkedListIterator end )
	{
		splice( pos.node, beg.node, end.node );
	}

// Removing

	/**
	 * Remove first item.
	 *
	 * Complexity: O(1).
	 */
	public void
	prune_front()
	{
		prune( ring.next );
	}

	/**
	 * Remove last item.
	 *
	 * Complexity: O(1).
	 */
	public void
	prune_back()
	{
		prune( ring.prev );
	}
	
	/**
	 * Remove item pointed to by pos.
	 *
	 * Complexity: O(1).
	 */
	public void
	prune( BilinkedListIterator pos )
	{
		prune( pos.node );
	}
	
	/**
	 * Remove range [beg..end[.
	 *
	 * Complexity: O(n).
	 */
	public void
	prune_range( BilinkedListIterator beg, BilinkedListIterator end )
	{
		prune_range( beg.node, end.node );
	}

	/**
	 * Remove all items.
	 */
	public void
	clear()
	{
		prune_range( ring.next, ring );
		// the Ring falls in Mount Doom...
	}

// Commands -- Misc.

	/**
	 * Send representation to out.
	 */
	public void
	to_stream( java.io.PrintStream out )
	{
		Bilinkable beg;
		Bilinkable end;
		
		beg = ring.next;
		end = ring;

		if( is_empty() )
		{
			out.print( "empty" );
		}
		else 
		{
			while( beg != end && beg.next != end )
			{
				out.print( ((ObjectBilinkable)beg).get_value().toString() );
				beg = beg.next;
			}
			
			out.print( ((ObjectBilinkable)beg).get_value().toString() );
		}
	}

// Commands -- Utility

	/**
	 * Insert a copy of range [beg..end[ before pos.
	 */
	protected void
	put_range_copy( Bilinkable pos, Bilinkable beg, Bilinkable end )
	{
		Bilinkable prv;
		
		prv = pos.prev;	// we know pos.prev exits

		while( beg != end )
		{
			prv.next = new ObjectBilinkable(
				((ObjectBilinkable)beg).get_value(), prv, null );
			prv = prv.next;
			beg = beg.next;
		}

		prv.next = pos;
		pos.prev = prv;
	}

	/**
	 * Insert a copy of range [beg..end[ before pos.
	 */
	protected void
	put_range_copy( Bilinkable pos, Iterator beg, Iterator end )
	{
		Bilinkable prv;
		
		prv = pos.prev;

		while( ! beg.equals( end ) )
		{
			prv.next = new ObjectBilinkable( beg.get_value(), prv, null );
			prv = prv.next;
			beg.next();
		}

		prv.next = pos;
		pos.prev = prv;
	}

	/**
	 * Insert n copies of value before pos.
	 */
	protected void
	put_n( Bilinkable pos, int n, Object value )
	{
		Bilinkable prv;
		
		prv = pos.prev;

		for( int i = 0; i < n; ++i )
		{
			prv.next = new ObjectBilinkable( value, prv, null );
			prv = prv.next;
		}

		prv.next = pos;
		pos.prev = prv;
	}

	/**
	 * Remove pos.
	 */
	protected void
	prune( Bilinkable pos )
	{
		pos.prev.remove_after();
	}

	/**
	 * Remove range [beg..end[.
	 */
	protected void
	prune_range( Bilinkable beg, Bilinkable end )
	{
		Bilinkable next;
		Bilinkable prev;

		next = beg;
		prev = beg.prev;

		while( beg != end )
		{
			next = beg.next;
			beg = next;
		}

		prev.next = next;
		next.prev = prev;
	}

	/**
	 * Remove range [beg..end[ from its position and insert it before pos.
	 *
	 * This method is potentially dangerous when the source and destination
	 * lists are the same. In this case, be carefull that the insertion
	 * point be not in the moved range(this would seem to work, but the
	 * structure of the list would become an unpredictable porridge ;-).
	 */
	protected void
	splice( Bilinkable pos, Bilinkable beg, Bilinkable end )
	{
		// add beg before pos:
		pos.prev.next = beg;
		
		// remove [beg..end[ from the old list
		beg.prev.next = end;
		end.prev = beg.prev;
		
		// reset the new chunk:
		beg.prev = pos.prev;
		pos.prev = end.prev;
		end.prev.next = pos;
	}
}

// vim:ts=4:ai:
