package book.set;

import book.InvalidIndexException;

/**
 * 32 boolean flags.
 *
 * A Bits is simply an easy way to set and check 32 boolean flags. There are
 * two kinds of methods in this class, the first one take a bit as argument.
 * These are #getb(), #setb() and #chob(). The other kind take as argument the
 * index of the bit to change. These are #get_bit() and #set_bit().
 *
 * There is also methods to change and access the complete bit set:
 * #set_value() and #get_value().
 *
 * The method names of the first kind (#getb(), #setb(), #chob()) are cryptic
 * but I like them, they are simple and still "Bean-oriented" (get/set).
 *
 * @author Antoine Dutot
 * @since 20000612
 * @version 0.1
 */
public class Bits
{
// Attributes
	
	/**
	 * 32 bits.
	 */
	protected int flag;
	
// Creation

	/**
	 * New empty bit set.
	 */
	public
	Bits()
	{
	}
	
	/**
	 * New bit set initialized to value.
	 */
	public
	Bits( int value )
	{
		flag = value;
	}

// Predicates
	
	/**
	 * No bit set ?
	 */
	public boolean
	is_empty()
	{
		return( flag == 0 );
	}

// Accessors -- contents
	
	/**
	 * Complete bit set.
	 */
	public int
	get_value()
	{
		return flag;
	}
	
	/**
	 * Are all the true bits of mask set ?
	 */
	public boolean
	getb( int mask )
	{
		return( ( flag & mask ) != 0 );
	}

	/**
	 * Are all the true bits of mask set ?
	 */
	public static boolean
	getb( int flag, int mask )
	{
		return( ( flag & mask ) != 0 );
	}

// Commands

	/**
	 * Change the complete bit set.
	 */
	public void
	set_value( int value )
	{
		flag = value;
	}

	/**
	 * Clear the complete bit set so that no bit is on.
	 */
	public void
	clear()
	{
		flag = 0;
	}
	
	/**
	 * Set to false all the true bits of mask.
	 *
	 * Other bits are unchanged.
	 */
	public void
	clrb( int mask )
	{
		flag &= ~mask;
	}
	
	/**
	 * Set to true all the true bits of mask.
	 *
	 * Other bits are unchanged.
	 */
	public void
	setb( int mask )
	{
		flag |= mask;
	}
	
	/**
	 * Set the i-th bit if on is true, else unset it.
	 *
	 * Other bits are unchanged.
	 */
	public void
	chob( boolean on, int mask )
	{
		if( on ) setb( mask );
		else      clrb( mask );
	}

// Accessors -- Get one bit only.
	
	/**
	 * i-th bit value.
	 */
	public boolean
	get_bit( int i ) throws InvalidIndexException
	{
		if( i < 0 || i > 31 )
		{
			throw new InvalidIndexException( i );
		}
		
		return( ( flag & ( 1 << i ) ) != 0 );
	}

// Commands -- Change one bit only.
	
	/**
	 * Set the i-th bit to on.
	 */
	public void
	set_bit( int i, boolean on ) throws InvalidIndexException
	{
		if( i < 0 || i > 31 )
		{
			throw new InvalidIndexException( i );
		}

		if( on ) flag |= ( 1<< i );
		else      flag &= ~( 1 << i );
	}
}

// vim:ts=4:ai:
