package book.set;

import book.InvalidStateException;

/**
 * Simple double ended queue.
 *
 * A double ended queue, a Deque, is both a fifo and a lifo stack at once.
 * That is, you can push elements at head or at end in constant time(not
 * considering the automatic memory managment). You can also pop elements at
 * head or at back in constant time.
 *
 * A deque is pronounced "deck".
 *
 * Implementation:
 * This is a really basic implementation of a deque. It will work well and is
 * robust but is probably not the most refined way of doing it.
 *
 * <code>
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	|A|_|D|E|Q|U|E| | | | | | |T|H|I|S|_|I|S|_|		1
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	 ^             ^           ^               ^
 * 	map           pte         beg             cap
 * 
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	|T|H|I|S|_|I|S|_|A|_|D|E|Q|U|E| | | | | | |		2
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	 ^                             ^           ^
 * 	map                           pte         cap
 * 	beg
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	| | | | | | |T|H|I|S|_|I|S|_|A|_|D|E|Q|U|E|		3
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	 ^           ^                             ^
 * 	map         beg                           cap
 * 	pte
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	| | | | | | | | | | | | | | | | | | | | | |		4
 * 	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * 	 ^                                         ^
 * 	map                                       cap
 * 	pte
 * 	beg
 * </code>
 *
 * In the diagram above, and in the rest of this file we will say that in the
 * case 1 the map is "segmented" wheras in cases 2 and 3 it is "packed". The
 * expression #pte stands for "past-the-end", #cap means "capacity" and #beg is
 * the abbreviation of "begin". The segmented case is caracterized by the fact
 * that #beg is strictly greater than #pte, this is an invariant of an
 * hypotetical state #is_segmented().
 *
 * As soon as #beg == #pte after an insertion(at front or at back), the array
 * is reallocated twice its size. Thus, #beg == #pte happens only when
 * #is_empty() is true(case 4).
 *
 * Also note that begining pointers(#beg and #map) are on the first cell wheras
 * ending pointers(#pte and #cap) are after the last used cell. Thus #pte is
 * really "past-the-end" of the range of used cells and #cap is really
 * past-the-last usable cell.
 *	
 * Invariants:
 *	[1] #pte cannot be on #cap, when #pte arrive on #cap it is automatically
 *	    moved at #map (this can be seen in case 3);
 *
 *	[2] #beg == #pte only when the deque #is_empty(). If after an insert #beg
 *	    == #pte, then the deque is automatically enlarged. This also means that
 *	    there always exists at least one free cell in a non empty deque.
 *
 * TODO: a trim() option to remove unused capacity
 *
 * @author Antoine Dutot
 * @since 19991025
 * @version 0.1
 */
public class Deque
{
// Attributes

	/**
	 * array of G.
	 */
	protected Object[] map;
	
	/**
	 * Past the last cell of #map.
	 */
	protected int cap;
	
	/**
	 * Pointer on first element.
	 */
	protected int beg;
	
	/**
	 * Past-The-End element of the used range.
	 */
	protected int pte;

// Constructors

	/**
	 * New deque with the default initial capacity.
	 */
	public
	Deque()
	{
		this( 8 );
	}
	
	/**
	 * New empty deque with an specified #initial_capacity().
	 */
	public
	Deque( int initial_capacity )
	{
		if( initial_capacity < 2 ) initial_capacity = 2;
	
		map = new Object [ initial_capacity ];
		cap = initial_capacity;
		beg = pte = 0;
	}

// Accessors

	public boolean
	invariant()
	{
		return( map != null );
	}

	/**
	 * No elements ?
	 */
	public boolean
	is_empty()
	{
		return( beg == pte );
	}

// Accessors

	/**
	 * How many elements ?
	 */
	public int
	get_count()
	{
		if( beg <= pte )
		{
			// packed
			return( pte - beg );
		}
		else
		{
			// segmented
			return( pte + ( cap - beg ) );
		}
	}

	/**
	 * How large can be count before an automatic reallocation ?
	 *
	 * At the time the capacity-th element will be inserted the deque will be
	 * automatically enlarged.
	 */
	public int
	get_capacity()
	{
		return cap;
	}

	/**
	 * First element.
	 */
	public Object
	get_front() throws InvalidStateException
	{
		if( is_empty() )
		{
			throw new InvalidStateException(
				"deque is empty" );
		}
		
		return map [ beg ];
	}
	
	/**
	 * Last element.
	 */
	public Object
	get_back() throws InvalidStateException
	{
		if( is_empty() )
		{
			throw new InvalidStateException(
				"deque is empty" );
		}
		
		if( pte != 0 )
		{
			return map [ pte - 1 ];
		}
		else
		{
			// pte at start, cannot decrement
			return map [ cap - 1 ];
		}
	}

	/**
	 * i-th element.
	 */
	public Object
	get( int i )
	{
	//	if( ( pte < beg )
	//	&&   ( ( beg + i ) >= cap ) )
		if( ( beg + i ) >= cap )
		{
			return map [ i - ( cap - beg ) ];
		}
		else
		{
			return map [ beg + i ];
		}
	}

	/**
	 * New iterator on the first element.
	 */
	public DequeIterator
	get_begin()
	{
		return new DequeIterator( this, beg );
	}

	/**
	 * Iterator past-the-last element.
	 */
	public DequeIterator
	get_end()
	{
		return new DequeIterator( this, pte );
	}

	/**
	 * Set iterator to point on the first element of this.
	 */
	public void
	get_begin( DequeIterator iterator )
	{
		iterator.deck = this;
		iterator.cursor = beg;
	}

	/**
	 * Set iterator to point on past-the-last element of this.
	 */
	public void
	get_end( DequeIterator iterator )
	{
		iterator.deck = this;
		iterator.cursor = pte;
	}

// Commands

	/**
	 * Ensure the deque has new_capacity.
	 *
	 * Use this method when you know capacity will not be enough for the
	 * number of elements you plan to insert later. This allows to exactly
	 * predict the time used by each operation since no automatic grow
	 * will take place and most operations are in O(1).
	 *
	 * If new_capacity is less or equal to #count() nothing is done. If it is
	 * less or equal to #capacity() but stictly greater than #count() the
	 * capacity is effectively decreased. This allows to gain memory after
	 * having pushed then poped a very large ammount of elements(if you know
	 * you will not use anew the deque for so many elements).
	 */
	public void
	reserve( int new_capacity )
	{
		if( new_capacity <= get_count() )
		{
			return;
		}
	
		Object[] nmap = new Object[new_capacity];	// new_map
		int i = 0;									// new_pte

		if( beg == pte )	// segmented case
		{
			for( int p = beg; p != cap; ++p, ++i )
			{
				nmap[i] = map[p];
			}

			for( int p = 0; p < pte; ++p, ++i )
			{
				nmap[i] = map[p];
			}
			
		}
		else				// packed case
		{
			for( int p = beg; p < pte; ++p, ++i)
			{
				nmap[i] = map[p];
			}
		}
		
		map = nmap;
		cap = new_capacity;
		beg = 0;
		pte = beg + i;
	}

	/**
	 * Push value in front of sequence.
	 */
	public void
	put_front( Object value )
	{
		//
		// When we enter here we are ensured by invariant [2] that an empty
		// cell exists.
		//
		if( beg == 0 )
		{
			beg = cap - 1;
		}
		else
		{
			--beg;
		}

		map[beg] = value;

		if( beg == pte )
		{
			grow();
		}
	}

	/**
	 * Push value at back of sequence.
	 */
	public void
	put_back( Object value )
	{
		//
		// When we enter here we are ensured by invariant [2] that an empty
		// cell exists. we are also ensured that #pte do not point out of
		// the map by invariant [1].
		//
	 //	inv check( pte >= map and pte < cap )

		map[pte] = value;
		
		++ pte;

		if( pte == cap )
		{
			pte = 0;
		}

		if( pte == beg )
		{
			grow();
		}
	}
	
	/**
	 * Pop first element.
	 */
	public void
	prune_front()
	{
	//	inv pre( not is_empty() )
		if( is_empty() )
		{
			throw new InvalidStateException(
				"cannot prune front of empty deque" );
		}

		++ beg;
		
		if( beg == cap )
		{
			beg = 0;
		}
	}
	
	/**
	 * Pop last element.
	 */
	public void
	prune_back()
	{
		if( is_empty() )
		{
			throw new InvalidStateException(
				"cannot prune front of empty deque" );
		}
	
		-- pte;
	
		if( pte < 0 )
		{
			pte = cap - 1;
		}
	}

	/**
	 * Make the deque empty.
	 */
	public void
	clear()
	{
		beg = pte = 0;
	}

// Commands -- Misc.

	/**
	 * Representation to out.
	 */
	public void
	to_stream( java.io.PrintStream out )
	{
		if( is_empty() )
		{
			out.println( "empty" );
		}
		else
		{
			if( beg > pte )	// segmented case
			{
				for( int p = beg; p < cap; ++p )
				{
					out.print( map[p] );
					out.print( ' ' );
				}

				for( int p = 0; p < pte; ++p )
				{
					out.print( map[p] );
					out.print( ' ' );
				}
			}
			else				// packed case
			{
				for( int p = beg; p < pte; ++p )
				{
					out.print( map[p] );
					out.print( ' ' );
				}
			}
		}
	}

// Commands -- Utility
	
	/**
	 * Enlarge the map by two.
	 *
	 * If the map was segmented, it is repacked.
	 */
	protected void
	grow()
	{
	 //	inv pre( beg == pte )
		
		//
		// ( #beg == #pte ) is a temporary anormal state: It is like
		// #is_empty() but in fact it is full. This is the only place invariant
		// [2] is not fullfilled.
		//
		// Thus, #grow() only adds pre and post conditions to #reserve()
		// because #grow() must be called at specific times when #count()
		// reaches #capacity().
		//
		reserve( cap << 1 );

		//
		// back to normal state, invariant 2 is now ok.
		//
	 //	post( beg != pte )
	 //	post( invariant() )
	}
	
	/**
	 * Indices of internal sentinels #beg, #pte and #cap.
	 */
	protected void
	debug()
	{
	 	System.err.println( "-----------------------" );
	 	System.err.println( "beg=" + beg );
	 	System.err.println( "pte=" + pte );
	 	System.err.println( "cap=" + cap + "=[0.." + ( cap - 1 ) + ']' );
	 	System.err.println( "-----------------------" );
	}
}

// vim:ts=4:ai:
