package book.set;

import book.Pair;

/**
 * Hash table.
 *
 * @author Antoine Dutot
 * @since 19980701
 * @version 0.1
 */
public class Hash
{
// Attributes
	
	/**
	 * Total number of items in the table.
	 */
	protected int nitems;
	
	/**
	 * Pages of items. Each page may contain zero or more items, as linked
	 * lists of nodes.
	 */
	protected HashNode[] pages;

	/**
	 * A prototype of the node to use. The default is to use
	 * ObjectHashNode instances, but this can be set by the user.
	 */
	protected HashNode proto;

// Attributes -- Primes table
	
	protected static final int MAX_PRIME = 28;
	protected static final int PRIMES [] = 
	{
		7,          23,
		53,         97,         193,       389,       769,
		1543,       3079,       6151,      12289,     24593,
		49157,      98317,      196613,    393241,    786433,
		1572869,    3145739,    6291469,   12582917,  25165843,
		50331653,   100663319,  201326611, 402653189, 805306457, 
		1610612741//, 3221225473, 4294967291
	};

// Constructors

	/**
	 * New hash table. The hash table is made to handle keys and values of
	 * type Object. Keys are considered equal using #Object.equals() and hash
	 * codes are found by using #Object.hashCode().
	 */
	public
	Hash()
	{
		this( PRIMES[0], new ObjectHashNode() );
	}

	/**
	 * Like the default constructor but with an initial number of pages. The
	 * number of pages(buckets) influences the performance of the hash.
	 */
	public
	Hash( int initial_page_count )
	{
		this( initial_page_count, new ObjectHashNode() );
	}
	
	/**
	 * New hash table with a specific prototype of HashNode. The kind of
	 * node to use is given by the prototype. This allows to define how the
	 * hash code is computed, the way keys and data are stored(if they are
	 * distinct or not), etc. For example you can now define that the hash code
	 * be the same for char[] and Strings so that object stored with a
	 * String key can be retrieved with a char array. You can have nodes
	 * where the key is the stored value.
	 */
	public
	Hash( HashNode prototype )
	{
		this( PRIMES[0], prototype );
	}
	
	/**
	 * Like #Hash(HashNode) but with an initial number of pages.
	 */
	public
	Hash( int initial_page_count, HashNode prototype )
	{
		int n;
		
		if( initial_page_count > PRIMES[MAX_PRIME - 1] )
		{
			initial_page_count = PRIMES[MAX_PRIME - 1];
		}

		n = PRIMES [ closest_prime( initial_page_count ) ];
		
		this.nitems = 0;
		this.proto = prototype;
		this.pages = new HashNode [n];
		
		init_pages();
	}


// Accessors
	
	public boolean
	invariant()
	{
		return( pages != null && pages.length >= 7 );
	}

	/**
	 * No items ?
	 */
	public boolean
	is_empty()
	{
		return( nitems == 0 );
	}

// Accessors

	/**
	 * Total number of items in the table. See also #get_page_count().
	 */
	public int
	get_count()
	{
		return nitems;
	}

	/**
	 * Number of pages(buckets) in the table. See also #get_count().
	 */
	public int
	get_page_count()
	{
		return pages.length;
	}

	/**
	 * Should always be less than 90% else the table is overloaded. See also
	 * #set_count().
	 */
	public int
	get_load_average()
	{
		return( ( nitems * 100 ) / pages.length );
	}

// Commands -- Adding

	/**
	 * Insert a new node addressed by key containing value. Complexity:
	 * O(1). See also #Hash.add(HashNode).
	 */
	public void
	add( Object key, Object value )
	{
		HashNode node;

		try
		{
			node = proto.new_instance( key, value );
		}
		catch( CloneNotSupportedException e )
		{
			throw new RuntimeException( e.getMessage() );
		}
		
		add( node );
	}

	/**
	 * Add a premade node in the hash. Complexity: 0(1). Here the node is
	 * already constructed. See also #Hash.add(Object,Object).
	 */
	public void
	add( HashNode node )
	{
		int page;
		
		page = address( node.get_key() );
		node.set_next( pages[page] );
		pages[page] = node;
		nitems += 1;
	}

	/**
	 * Insert a new node grouping it with other nodes of identiqual key. The
	 * new node is inserted just after the first node that has the same key, or
	 * at the begining of the page if no node already has the same key.
	 *
	 * Complexity: ~ O(n/m) with n = #count() and m = #page_count(). Note the
	 * difference of complexity with a simple #add() wich is in O(1). Here,
	 * we need to search in the page before inserting.
	 *
	 * See also #Hash.add_multi(HashNode) and #Hash.add(Object,Object)
	 * and [Hash.add(HashNode).
	 */
	public void
	add_multi( Object key, Object value )
	{
		HashNode node;

		try
		{
			node = proto.new_instance( key, value );
		}
		catch( CloneNotSupportedException e )
		{
			throw new RuntimeException( e.getMessage() );
		}
		
		add_multi( node );
	}
	
	/**
	 * Insert a new_node grouping it with other nodes of identiqual key.
	 * See also #Hash.add_multi(Object,Object) and #Hash.add(Object,Object)
	 * and #Hash.add(HashNode).
	 */
	public void
	add_multi( HashNode new_node )
	{
		int page;
		HashNode node;

		page = address( new_node.get_key() );
		node = pages[page];

		while( node != null )
		{
			if( node.key_equals( new_node.get_key() ) )
			{
				node.append( new_node );
				break;
			}
				
			node = (HashNode) node.next;
		}

		if( node == null )
		{
			new_node.set_next( pages[page] );
			pages[page] = new_node;
		}

		nitems += 1;
	}


// Accessors -- Iterators

	/**
	 * New iterator on the first node. Complexity: O(n) with n = #page_count().
	 */
	public HashIterator
	get_begin()
	{
		return new HashIterator( this, get_first_node() );
	}

	/**
	 * New iterator point past-the-last node. Complexity: O(1).
	 */
	public HashIterator
	get_end()
	{
		return new HashIterator( this, null );
	}

	/**
	 * Make iterator point on the first node.
	 */
	public void
	get_begin( HashIterator iterator )
	{
		iterator.set_node( this, get_first_node() );
	}

	/**
	 * Make iterator point past-the-last node.
	 */
	public void
	get_end( HashIterator iterator )
	{
		iterator.set_node( this, null );
	}

	/**
	 * Object associated to a key(like finding a node an getting its value).
	 * For specific kinds of nodes this may not work. Returns null if not
	 * found.
	 */
	public Object
	find_value( Object key )
	{
		HashNode node;

		node = find_node( key );

		if( node != null )
		{
			return node.get_value();
		}

		return null;
	}

	/**
	 * First iterator on a node matching key or #get_end() if not found.
	 * Complexity: ~ O(n/m) with n = #count() and m = #page_count().	
	 */
	public HashIterator
	find( Object key )
	{
		return new HashIterator( this, find_node( key ) );
	}

	/**
	 * Like #Hash.find(key) but iterator is set to the result.
	 */
	public void
	find( Object key, HashIterator iterator )
	{
		iterator.set_node( this, find_node( key ) );
	}

/*	public void
 *	find_multi( Object key, Iterator begin, Iterator end )
 *	{
 *		find_multiples_nodes( key,  );
 *		begin.set_node( this, paire
 *	}
 */
// Accessors -- Nodes
	
	/**
	 * First node matching key or null if not found. Complexity: ~ O(n/m)
	 * with n = #count() and m = #page_count().	
	 */
	public HashNode
	find_node( Object key )
	{
		int page;
		HashNode node;

		page = address( key );
		node = pages[page];

		while( node != null )
		{
			if( node.key_equals( key ) )
			{
				return node;
			}
				
			node = (HashNode) node.next;
		}

		return null;
	}
	
	/**
	 * Store all occurences of key in result or empty range if not found.
	 * result will contain instances of HashNode. The empty range is
	 * represented by the pair(null,null).
	 */
	@SuppressWarnings("unchecked")
	public void
	find_node_multi( Object key, Pair result )
	{	
		int n;
		int page;
		HashNode node;
			
		n = pages.length;
		page = address( key );
		node = pages[page];
			
		result.setFirst( null );
		result.setSecond( null );

		// Search first occurence

		while( node != null )
		{
			if( node.key_equals( key ) )
			{
				result.setFirst( node );
				break;
			}
				
			node = (HashNode) node.next;
		}

		// Search past-the-end occurence

		while( node != null )
		{
			if( ! node.key_equals( key ) )
			{
				result.setSecond( node );
				return;
			}
			
			node = (HashNode) node.next;
		}

		// If the past the end iterator is not in current bucket

		while( node == null && ++page < n )
		{
			node = pages[page];
		}

		result.setSecond( node );
	}
	
	/**
	 * Reference on the first node or null if #is_empty(). Complexity: O(n)
	 * with n = #page_count().
	 */
	public HashNode
	get_first_node()
	{
		int n;

		if( is_empty() )
		{
			return null;
		}

		n = pages.length;
	
		for( int page = 0; page < n; ++page )
		{
			if( pages[page] != null )
			{
				return pages[page];
			}
		}
						// we should never arrive here,
		return null;	// the table is empty but nitems > 0
	}

	/**
	 * Successor of node in the table or null if node is the last one.
	 * Complexity: at worst O(n) with n = #page_count() but most often O(1).
	 */
	public HashNode
	get_next_node( HashNode node )
	{
		int n;
		int page;

		if( node.next != null )
		{
			return(HashNode) node.next;
		}
		
		n = pages.length;
		page = address( node.get_key() ) + 1;
			
		while( page < n )
		{
			if( pages[page] != null )
			{
				return pages[page];
			}

			page += 1;
		}
		
		return null;
	}

// Commands -- Removing
	
	/**
	 * Remove the first node matching key. The corresponding node(and its
	 * value) is not returned(side effect). To ensure the node was effectively
	 * in the hash table, use #Hash.remove(Object,Iterator).
	 *
	 * Complexity: at worst O(n/m) with n = #count() and m = #page_count(). Note
	 * that it is not an error to search an non existent node, but a linear
	 * search in O(n/m) still occurs.
	 */
	public void
	remove( Object key )
	{
		prune( key );
	}

	/**
	 * Remove the first node matching key and place it in iterator.
	 * iterator is equal to #Hash.get_end() if no node match key.
	 */
	public void
	remove( Object key, HashIterator iterator )
	{
		HashNode node;
		
		node = prune( key );
		iterator.set_node( this, node );
	}
	
	/**
	 * Clears the table. Complexity: O(n) with n = #count().
	 */
	public void
	clear()
	{
		int n;

		n = pages.length;

		for( int i = 0; i < n; ++i )
		{
			prune_page( i );
		}
	}


// Commands -- Resizing, swapping
	
	/**
	 * Exchange the contents of other with this. Complexity: O(1).
	 */
	public void
	swap( Hash other )
	{
		int it;
		HashNode[] ht;
		
		ht = pages;
		pages = other.pages;
		other.pages = ht;

		it = nitems;
		nitems = other.nitems;
		other.nitems = it;
	}

	/**
	 * Reset the number of page(buckets). If #preserve_groups() is true, this
	 * method will handle nodes grouped by #add_multi(). This is useful only if
	 * you used this method to create groups instead of using #add() each time.
	 * Be carefull that using this option is slower(as #add_multi() is slower
	 * than #add()).
	 */
	public void
	set_page_count( int new_count, boolean preserve_groups )
	{
		HashNode[] old_pages, ht;
		int n;
		Linkable node;

		n = PRIMES[closest_prime( new_count )];

		if( n == pages.length )
		{
			return;
		}
		
		ht = pages;
		pages = new HashNode [ n ];
		old_pages = ht;
		init_pages();
		
		nitems = 0;
		n = old_pages.length;
		
		if( preserve_groups )
		{
			for( int i = 0; i < n; i++ )
			{
				while( old_pages[i] != null )
				{
					node = old_pages[i];
					old_pages[i] = (HashNode) node.next;
					add_multi( (HashNode) node );
				}
			}
		}
		else
		{
			for( int i = 0; i < n; i++ )
			{
				while( old_pages [ i ] != null )
				{
					node = old_pages[i];
					old_pages[i] = (HashNode) node.next;
					add( (HashNode) node );
				}
			}
		}
	}

// Commands -- Misc.

	/**
	 * Send the hash table representation to out.
	 */
	public void
	to_stream( java.io.PrintStream out )
	{
		HashNode node;
		int n;

		n = pages.length;
		
		for( int page = 0; page < n; ++page )
		{
			node = pages[page];

			out.print( " page=" );
			out.print( page );
			out.print( ": " );
				
			while( node != null && node.next != null )
			{
				out.print( " [" );
				out.print( node.get_hash_code( node.get_key() ) );
				out.print( ']' );
				
				node = (HashNode) node.next;
			}
				
			if( node != null )
			{
				out.print( " [" );
				out.print( node.get_hash_code( node.get_key() ) );
				out.print( ']' );
			}
				
			out.print( '\n' );
		}
	}

// Accessors -- Utility
	
	/**
	 * Page index for key. Complexity: O(1).
	 */
	protected int
	address( Object key )
	{	
		int h;

		h = proto.get_hash_code( key );

		if( h < 0 )
		{
			return( -h % pages.length );
		}

		return( h % pages.length );
	}
	
	/**
	 * First prime number greater or equal to number. Complexity: O(log(n))
	 * in the primes number array.
	 */
	protected int
	closest_prime( int number )
	{
		int b;
		int e;
		int mid;

		// This is a dichotomy(a.k.a. binary search)
		b = 0;
		e = MAX_PRIME;
		mid = ( e - b );

		while( mid > 0 )
		{
			mid = ( e - b ) >> 1;
			
			if( PRIMES[b+mid] < number )
			{
				b = b + ( mid + 1 );
			}
			else
			{
				e = e - ( mid );
			}
		}

		return b;
	}
	
// Commands -- Utility

	/**
	 * Set all pages to null. Only when the pages are empty but not
	 * initialized.
	 */
	protected void
	init_pages()
	{
		int n;

		n = pages.length;
	
		for( int i = 0; i < n; ++i )
		{
			pages[i] = null;
		}
	}

	/**
	 * Remove all items in a page. Complexity O(n) in the page.
	 */
	protected void
	prune_page( int page )
	{
		Linkable node;
		
		node = pages[page];

		while( node != null )
		{
			node = node.next;

			--nitems;
		}
	
		pages[page] = null;
	}

	/**
	 * Remove the first node matching key and return it or null. This method
	 * is protected since it has a side effect(remove the node) but still
	 * returns a value.
	 */
	protected HashNode
	prune( Object key )
	{
		// Here we must start from the begining of the list to search the
		// previous element in order to rebuild the list. This is due to
		// the fact the list is singly chained:

		int page;
		HashNode node, prev;
		
		page = address( key );
		node = pages[page];
		prev = null;

		while( node != null )
		{
			if( node.key_equals( key ) )
			{
				if( prev != null )
				{
					prev.next = node.next;
				}
				else
				{
					pages[page] = (HashNode) node.next;
				}
				
				nitems -= 1;
				
				return node;
			}
			
			prev = node;
			node = (HashNode) node.next;
		}

		return null;
	}
}

// vim:ts=4:ai:
