package book.set;

import book.set.Array;
import book.set.IntArray;

/**
 * Index array.
 *
 * <p>An index array allows to keep a set of items with each stored item
 * maintaining its index constant even when new items are inserted or removed
 * in the middle.</p>
 *
 * <p>One cannot choose the index where the element will be stored. The index
 * where the element was stored is get using the {@link #lastIndex()}
 * method.</p>
 *
 * <p>As elements are inserted or removed, not all index between 0 and a valid
 * index contain something. Use the {@link #validIndex(int)}.</p>
 *
 * @author Antoine Dutot
 * @since 20020628
 */
public class IndexArray
{
// Attributes

	/**
	 * Array of items.
	 */
	protected Array set = new Array();

	/**
	 * Reusable indices.
	 */
	protected IntArray freeIdx = new IntArray();

	/**
	 * Last index used.
	 */
	protected int lx = -1;

// Constructors

	/**
	 * New empty index array.
	 */
	public
	IndexArray()
	{
	}

// Accessors

	/**
	 * Last used indeX.
	 */
	public int
	lastIndex()
	{
		return lx;
	}

	/**
	 * Number of used cells.
	 */
	public int
	count()
	{
		return( set.size() - freeIdx.size() );
	}

	/**
	 * Does index points on an item in this set?.
	 */
	public boolean
	validIndex( int index )
	{
		return( index >= 0 && index < set.size() && set.item(index) != null );
	}

	/**
	 * i-th item.
	 * Synonym of {@link #get(int)}.
	 */
	public Object
	item( int i )
	{
		return set.item( i );
	}

	/**
	 * i-th item.
	 * Synonym of {@link #item(int)}.
	 */
	public Object
	get( int i )
	{
		return set.item( i );
	}

// Commands

	/**
	 * Add an item. The index at which the item is inserted can be get using
	 * {@link #lastIndex()}.
	 */
	public void
	add( Object item )
	{
		if( freeIdx.size() > 0 )
		{
			set.set( freeIdx.back(), item );
			lx = freeIdx.back();
			freeIdx.pruneBack();
		}
		else
		{
			lx = set.size();
			set.add( item );
		}
	}

	/**
	 * Append <code>other</code> to this. All indices in <code>other</code>
	 * remain the same with an offset added. Empty cells in the middle are
	 * therefore preserved. This property allows to keep indices in place. When
	 * this method returns the {@link #lastIndex()} is changed to contain the
	 * offset.
	 */
	public void
	append( IndexArray other )
	{
		int n = set.size();
		int m = other.set.size();

		for( int i=0; i<m; ++i )
			set.add( other.set.get( i ) );

		m = other.freeIdx.size();

		for( int i=0; i<m; ++i )
			freeIdx.add( other.freeIdx.get( i ) + n );

		lx = n;
	}

	/**
	 * Remove the i-th item.
	 * Synonym of {@link #prune(int)}.
	 */
	public void
	remove( int i )
	{
		prune( i );
	}

	/**
	 * Remove the i-th item.
	 * Synonym of {@link #remove(int)}.
	 */
	public void
	prune( int i )
	{
		lx = -1;

		if( i > 0 && i < set.size() && set.item( i ) != null )
		{
			set.set( i, null );
			freeIdx.add( i );
			lx = i;
		}	
	}

	/**
	 * Clear all references.
	 */
	public void
	clear()
	{
		set.setCount( 0 );
		freeIdx.setCount( 0 );
	}
}
