package book.set;

/**
 * Singly linked list.
 *
 * @author Antoine Dutot.
 * @since 19980727
 * @version 0.1
 */
public class LinkedList
{
// Attributes

	/**
	 * Pointer on the first item of the list.
	 *
	 * This is only a Linkable storing a value and at the same time do not
	 * consider the head as a special case. However this implies lots of
	 * downcasts in the code below.
	 */
	protected Linkable head;

// Constructors
	
	/**
	 * New empty list.
	 */
	public
	LinkedList()
	{
		head = new Linkable();
		head.set_next( null );
	}

	/**
	 * New list of n references to value.
	 */
	public
	LinkedList( Object value, int n )
	{
		head = new Linkable();
		head.set_next( null );
		
		append_n( head, n, value );
	}

	/**
	 * Make a copy of range [beg..end[.
	 */
	public
	LinkedList( Iterator beg, Iterator end )
	{
		head = new Linkable();
		head.set_next( null );

		append_range_copy( head, beg, end );
	}

// Predicates

	/**
	 * No items ?
	 */
	public boolean
	is_empty()
	{
		return( head.next == null );
	}

	/**
	 * Does iterator is dereferencable?
	 */
	public boolean
	valid_iterator( LinkedListIterator iterator )
	{
		if( iterator != null )
		{
			return( ! iterator.is_end() );
		}
		
		return false;
	} 

// Accessors

	/**
	 * Number of items.
	 *
	 * Complexity: O(n), be careful! LinkedList does not own a count
	 * property.
	 */
	public int
	get_count()
	{
		int n;
		
		n = 0;
		
		for( Linkable x = head.next;  x != null;  x = x.next )
		{
			++n;
		}
		
		return n;
	}

	/**
	 * First item.
	 */
	public Object
	get_front()
	{
		if( is_empty() ) return null;

		return((ObjectLinkable)head.next).value;
	}
	
	/**
	 * Last item.
	 *
	 * Complexity: O(n).
	 */
	public Object
	get_back()
	{
		if( is_empty() ) return null;
		
		return((ObjectLinkable)last_node()).value;
	}

	/**
	 * Store in prev the predecessor of pos.
	 *
	 * No object is created. Complexity: O(n).
	 */
	public void
	get_previous( LinkedListIterator pos, LinkedListIterator prev )
	{ 
		if( pos.node == head.next ) prev.set_node( pos.node );

		prev.set_node( (ObjectLinkable) prev_node( pos.node ) );
	} 

	/**
	 * Iterator on first item. idem to end if the list is empty.
	 *
	 * Complexity: O(1).
	 */
	public LinkedListIterator
	get_begin()
	{
		return new LinkedListIterator( (ObjectLinkable) head.next );
	}

	/**
	 * Like begin without arguments, but store the result in i.
	 *
	 * No object is created. Complexity: O(1).
	 */
	public void
	get_begin( LinkedListIterator i )
	{
		i.set_node( (ObjectLinkable) head.next );
	}

	/**
	 * Iterator on past-the-last item. Not dereferencable.
	 *
	 * Complexity: O(1).
	 */
	public LinkedListIterator
	get_end()
	{
		return new LinkedListIterator( (ObjectLinkable) null );
	}

	/**
	 * Like end without arguments, but store the result in i.
	 *
	 * No object is created.
	 * Complexity: O(1).
	 */
	public void
	get_end( LinkedListIterator i )
	{
		i.set_node( null );
	}

// Commands

	/**
	 * Swap other and current contents.
	 *
	 * Complexity: O(1).
	 */
	public void
	swap( LinkedList other )
	{
		Linkable node;

		node = head;
		head = other.head;
		other.head = node;
	} 

	/**
	 * Insert value before all items.
	 *
	 * Complexity: O(1).
	 */
	public void
	put_front( Object value )
	{
		head.set_next( new ObjectLinkable( value, head.next ) );
	}

	/**
	 * Insert value after all items.
	 *
	 * Complexity: O(n).
	 */
	public void
	put_back( Object value )
	{
		last_node().append( new ObjectLinkable( value, null ) );
	}

	/**
	 * Remove first iterm.
	 *
	 * Complexity: O(1).
	 */
	public void
	prune_front()
	{
		Linkable n;
		
		n = head.next;
		head.set_next( n.next );
	}

	/**
	 * Remove last item.
	 *
	 * Complexity: O(n). Its complexity discourages its use ...
	 */
	public void
	prune_back()
	{
		Linkable tail;
		Linkable prev;
		
		tail = head.next;
		prev = head;

		while( tail.next != null )
		{
			prev = tail;
			tail = tail.next;
		}
		
		prev.set_next( null );
	}

// Insertion

	/**
	 * Insert value before pos.
	 *
	 * Complexity: O(n) in forward list see #put_after().
	 *
	 * Due to internal representation, in fact, it is not required that pos be
	 * valid, it can be end and work even if the list is empty. This is
	 * specific to this list however and should not be relied upon.
	 */
	public void
	put( LinkedListIterator pos, Object value )
	{
		prev_node(pos.node).append( new ObjectLinkable( value, null ) );
	}

	/**
	 * Insert n copies of value before pos.
	 *
	 * Complexity: O(2n) in forward list see #put_n_after() and #put().
	 */
	public void
	put_n( LinkedListIterator pos, int n, Object value )
	{
		append_n( prev_node( pos.node ), n, value );
	}

	/**
	 * Insert a copy of range [beg..end[ before pos.
	 *
	 * Complexity: O(2n) in forward list see #put_range_after() and #put().
	 */
	public void
	put_range( LinkedListIterator pos, Iterator beg, Iterator end )
	{
		append_range_copy( prev_node( pos.node ), beg, end );
	}

	/**
	 * Insert value after pos.
	 *
	 * Complexity: O(1).
	 */
	public void
	put_after( LinkedListIterator pos, Object value )
	{
		pos.node.set_next( new ObjectLinkable( value, pos.node.next ) );
	}

	/**
	 * Insert n copies of value after pos.
	 *
	 * Complexity: O(n).
	 */
	public void
	put_n_after( LinkedListIterator pos, int n, Object value )
	{
		append_n( pos.node, n, value );
	}

	/**
	 * Insert a copy of the range [beg..end[ after pos.
	 *
	 * Complexity: O(n).
	 */
	public void
	put_range_after( LinkedListIterator pos, Iterator beg, Iterator end )
	{
		append_range_copy( pos.node, beg, end );
	}

// Removing

	/**
	 * Remove value pointed to by pos.
	 *
	 * Complexity: at worst O(n) in forward list, see #prune_after().
	 */
	public void
	prune( LinkedListIterator pos )
	{
		prev_node(pos.node).remove_after();
	}

	/**
	 * Remove range [beg..end[.
	 *
	 * Complexity: at worst O(2n) in forward list, see #prune_range_after().
	 */
	public void
	prune_range( LinkedListIterator beg, LinkedListIterator end )
	{
		prune_range_after( prev_node( beg.node ), end.node );
	} 

	/**
	 * Remove pos successor.
	 *
	 * Complexity: O(1).
	 */
	public void
	prune_after( LinkedListIterator pos )
	{
		pos.node.remove_after();
	}

	/**
	 * Remove range ]beg..end[.
	 *
	 * Complexity: O(n).
	 */
	public void
	prune_range_after( LinkedListIterator beg, LinkedListIterator end )
	{
		prune_range_after( beg.node, end.node );
	}

	/**
	 * Remove all items.
	 *
	 * Note: their values are not deleted if they are pointers.
	 */
	public void
	clear()
	{
		prune_range_after( head, null );
	}

// Accessors

	/**
	 * Representation.
	 */
	public String
	toString()
	{
		Linkable cur;
		StringBuffer out;
		
		if( is_empty() )
		{
			return "empty";
		}
		
		cur = head.next;
		out = new StringBuffer();

		while( cur != null && cur.next != null )
		{
			out.append( ((ObjectLinkable)cur).value.toString() );
			out.append( ", " );

			cur = cur.next;
		}

		if( cur != null )
		{
			out.append( ((ObjectLinkable)cur).value.toString() );
		}

		return out.toString();
	}

	/**
	 * send representation to out.
	 */
	public void
	to_stream( java.io.PrintStream out )
	{
		Linkable cur;
		
		if( is_empty() )
		{
			out.print( "empty" );
			return;
		}
		
		cur = head.next;

		while( cur != null && cur.next != null )
		{
			out.print( ((ObjectLinkable)cur).value.toString() );
			out.print( ", " );
			cur = cur.next;
		}

		if( cur != null )
		{
			out.print( ((ObjectLinkable)cur).value.toString() );
		}
	}

// Commands -- Utility

	/**
	 * Append a copy of range [beg..end[ after to.
	 *
	 * Complexity: O(n).
	 */
	protected void
	append_range_copy( Linkable to, Iterator beg, Iterator end )
	{
		for(; ! beg.equals( end ); beg.next() )
		{
			to.set_next( new ObjectLinkable( beg.get_value(), to.next ) );
			to = to.next;
		}
	}

	/**
	 * Append a copy of range [beg..end[ after to.
	 *
	 * Complexity: O(n).
	 */
	protected void
	append_list_copy( Linkable to, Linkable beg, Linkable end )
	{
		while( beg != end )
		{
			to.set_next( new ObjectLinkable(
				((ObjectLinkable)beg).value, to.next ) );
			
			to = to.next;
			beg = beg.next;
		}
	}
	
	/**
	 * Append n copies of value after to:
	 *
	 * Complexity: O(n).
	 */
	protected void
	append_n( Linkable to, int n, Object value )
	{
		Linkable after_last;
		
		after_last = to.next;

		for( int i = 0; i < n; ++i )
		{
			to.set_next( new ObjectLinkable( value, null ) );
			to = to.next;
		}

		to.next = after_last;
	}
	
	/**
	 * Removes the range ]pos..last[, the successor of pos becomes last.
	 *
	 * Complexity: O(1).
	 * Note that pos and last are not removed.
	 */
	protected void
	prune_range_after( Linkable pos, Linkable last )
	{
		pos.set_next( last );
	}
	
	/**
	 * Predecessor of pos (can be #head so it is never null).
	 *
	 * Complexity: O(n).
	 */
	protected Linkable
	prev_node( Linkable pos )
	{
		Linkable cur;
		
		cur = head;

		while( cur != null && cur.next != pos )
		{
			cur = cur.next;
		}
		
		return cur;
	}
	
	/**
	 * Pointer on last item or #head if list is empty.
	 *
	 * Complexity: O(n).
	 */
	protected Linkable
	last_node()
	{
		Linkable cur;
		
		cur = head;

		while( cur.next != null )
		{
			cur = cur.next;
		}
		
		return  cur;
	}
}

// vim:ts=4:ai:
