package book.set;

/**
 * Simple tree node
 *
 * @author Antoine Dutot
 * @since 20010620
 * @version 0.1
 */
public class TreeNode
{
// Attributes

	/**
	 * Parent.
	 */
	protected TreeNode up;

	/**
	 * Children nodes.
	 */
	protected Array children;

// Constructors

	/**
	 * New "root and leaf" node.
	 */
	public
	TreeNode()
	{
		up = null;
		children = new Array();
	}
	
	/**
	 * Each child is unparented from this node and this node in turn is
	 * detached and removed from its parent.
	 */
	public void
	detach()
	{
		int n;
		
		//
		// Detach children.
		//
		if( ! is_leaf() )
		{
			n = children.size();
		
			for( int i = 0; i < n; ++i )
			{
				((TreeNode)children.item(i)).unset_parent();
			}
			
			children.setCount( 0 );
		}

		//
		// Detach from parent.
		//
		if( ! is_root() )
		{
			up.prune_child( this );
		}
	}

// Predicates

	/**
	 * Is i the index of a child of this node ?
	 */
	public boolean
	valid_index( int i )
	{
		return( i >= 0 && i < child_count() );
	}
	
	/**
	 * No parent node ?
	 */
	public boolean
	is_root()
	{
		return( up == null );
	}

	/**
	 * No descendant items ?
	 */
	public boolean
	is_leaf()
	{
		return children.isEmpty();
	}

// Accessors

	/**
	 * Parent node or null.
	 *
	 * null indicate the node is a root node.
	 */
	public TreeNode
	parent()
	{
		return up;
	}
	
	/**
	 * i-th child.
	 */
	public TreeNode
	child( int i )
	{
		return(TreeNode) children.item( i );
	}

	/**
	 * How many children nodes ?
	 *
	 * Zero indicates this node is a leaf node.
	 */
	public int
	child_count()
	{
		return children.size();
	}

// Commands

	/**
	 * Attach to a parent node.
	 *
	 * If there was no previous parent, this node is no more a root node.
	 * If there was a previous parent, the child is *not* removed from it.
	 *
	 * The child does not insert itself in the new parent. This is the
	 * responsability of the parent (it is done in #put_child()). This is
	 * the reason why this method is protected.
	 */
	protected void
	set_parent( TreeNode parent )
	{
		up = parent;
	}
	
	/**
	 * Detach from the parent node.
	 *
	 * This node becomes a root node.
	 *
	 * The child does not remove itself from the old parent. This is the
	 * responsability of the parent (it is done in #prune_child(),
	 * #prune_child_at() and #wipe_out()). This is the reason why this method
	 * is protected.
	 */
	protected void
	unset_parent()
	{
		up = null;
	}

// Commands
	
	/**
	 * Add a reference to child at the end of the children list.
	 *
	 * The child is reparented (its #set_parent() method is called) under this
	 * node. If this node had no child yet, it is no more a leaf node. 
	 */
	public void
	put_child( TreeNode child )
	{
		children.putBack( child );
		child.set_parent( this );
	}
	
	/**
	 * Synonym of {@link #put_child(TreeNode)}.
	 */
	public void
	add( TreeNode child )
	{
		children.putBack( child );
		child.set_parent( this );
	}
	
	/**
	 * Remove child from the list of children node.
	 *
	 * This is a no-op if child is not in the list but a linear search still
	 * occurs: the complexity is in O(n).
	 *
	 * The child is unparented(its unset_parent method is called). If child was
	 * the last child, this node becomes a leaf node.
	 *
	 * See #prune_child_at() and #wipe_out().
	 */
	public void
	prune_child( TreeNode child )
	{
		int n;
		
		n = children.size();
	
		for( int i = 0; i < n; ++i )
		{
			if( children.item( i ) == child )
			{
				prune_child_at( i );

				break;
			}
		}
	}
	
	/**
	 * Remove the i-th child reference.
	 *
	 * Complexity O(1).
	 *
	 * The child is unparented(its unset_parent() method is called). If
	 * i was the last child(index 0), this node becomes a leaf node.
	 *
	 * See #prune_child() and #wipe_out().
	 */
	public void
	prune_child_at( int i )
	{
		((TreeNode)children.item(i)).unset_parent();
		children.prune( i );
	}

	/**
	 * Remove all children.
	 *
	 * The childen are unparented(their #unset_parent() method is called). If
	 * this node as the owned_children flag set, the children are also deleted.
	 *
	 * See #prune_child_at() and #prune_child().
	 */
	public void
	clear()
	{
		int n;

		n = children.size();
	
		for( int i = 0; i < n; ++i )
		{
			((TreeNode)children.item(i)).unset_parent();
		}

		children.setCount( 0 );
	}
}

// vim:ts=4:ai:
