package book.swing;

import book.*;

import java.io.*;
import java.util.*;
import java.awt.*;

/**
 * Font helper and storage.
 *
 * <p>This class contains a set of font instances usable with all swing
 * components. It automatically create and remember fonts by a key name. This
 * context is also able to instanciate and override fonts using files
 * (currently only TrueType) stored in a jar located in the classpath. This
 * feature use a simple map method that bind to a particular font family an URL
 * pointing inside a jar file (a mere path relative to a jar root).</p>
 *
 * @author Antoine Dutot
 * @since 20030418
 */
public class FontContext
{
// Attributes

	/**
	 * Created fonts. This contains pairs (key,value) where the key is a name
	 * and the value a {@link java.awt.Font}. The name is not the name of the
	 * font but a name like "icon-font" or "plain-text", etc.
	 */
	protected HashMap<String,Font> fonts = new HashMap<String,Font>();

	/**
	 * Map of families to font file names. For fonts that cannot be found in
	 * the JRE path, this map provides pairs (key,value) where the key is a
	 * font family name and the value a font instance at 11 points (created
	 * using <tt>Font.createFont()</tt>). This is portable for fonts embedded
	 * in jar files.
	 * @see #map_family(String,String)
	 */
	protected HashMap<String,Font> map = new HashMap<String,Font>();

// Constructors

	/**
	 * Create some default fonts.
	 *
	 * <p>The fonts created all pertain to <code>family</code> in three sizes
	 * <code>normal</code>, <code>small</code> and <code>big</code>. They are
	 * stored under the names:</p><ul>
	 * 	<li>plain</li>
	 * 	<li>plain-small</li>
	 * 	<li>plain-big</li>
	 * 	<li>italic</li>
	 * 	<li>italic-small</li>
	 * 	<li>italic-big</li>
	 * 	<li>bold</li>
	 * 	<li>bold-small</li>
	 * 	<li>bold-big</li>
	 * </ul>
	 * <p>Default fonts are initialized only if the family is found. Font
	 * getters are smart enough to provide usable fonts if requested ones are
	 * not available.</p>
	 */
	public
	FontContext( String family, int normal, int small, int big )
	{
		init_default_fonts( family, normal, small, big );
	}

	/**
	 * Like {@link #FontContext(String,int,int,int)} but also provide a
	 * fallback URL for the font family. If the family cannot be found, this
	 * mapping will be inserted as if {@link #map_family(String,String)} had
	 * been used using <code>family</code> as the key.
	 * @throws IOException if the font <code>URL<code> cannot be open or read.
	 * @throws NotFoundException if the font <code>URL</code> cannot be
	 * located.
	 * @throws FontFormatException if the <code>URL</code> does not point on a
	 * valid font.
	 */
	public
	FontContext( String family, int normal, int small, int big, String URL )
		throws IOException, FontFormatException, NotFoundException
	{
		try
		{
			map_family( family, URL );
		}
		catch( SingletonException e )
		{
			// Should neve happen, since the map is empty.
			throw new RuntimeException( e.getMessage() );
		}

		init_default_fonts( family, normal, small, big );
	}

	/**
	 * Copy of <code>other</code>. All fonts are copied, not only referenced.
	 */
	public
	FontContext( FontContext other )
	{
		Iterator k = other.fonts.keySet().iterator();
		Font     f;
		String   s;

		while( k.hasNext() )
		{
			s = (String) k.next();
			f = other.get_font( s, (Font)null );

			if( f != null )
				add_font( s, f.getFamily(), f.getStyle(), f.getSize() );
		}
	}

	/**
	 * Try to initialise some default fonts. The fonts are initialized only if
	 * <code>family</code> is found.
	 * @see #FontContext(String,int,int,int)
	 */
	protected void
	init_default_fonts( String family, int normal, int small, int big )
	{
		if( family_exists( family ) )
		{
			add_font( "plain",       family, Font.PLAIN, normal );
			add_font( "plain-small", family, Font.PLAIN, small );
			add_font( "plain-big",   family, Font.PLAIN, big );

			add_font( "italic",       family, Font.ITALIC, normal );
			add_font( "italic-small", family, Font.ITALIC, small );
			add_font( "italic-big",   family, Font.ITALIC, big );

			add_font( "bold",       family, Font.BOLD, normal );
			add_font( "bold-small", family, Font.BOLD, small );
			add_font( "bold-big",   family, Font.BOLD, big );
		}
	}

// Accessors

	/**
	 * True if the font <code>family</code> exits or is mapped. Mapped fonts
	 * are always looked up first. This is an heavy test use it moderately :-).
	 */
	public boolean
	family_exists( String family )
	{
		GraphicsEnvironment genv;
		String              fenv[];
		String              f;

		if( map.get( family ) != null )
			return true;

		genv   = GraphicsEnvironment.getLocalGraphicsEnvironment();
		fenv   = genv.getAvailableFontFamilyNames();
		family = family.toLowerCase();

		for( int i=1; i<fenv.length; i++ )
		{
			f = fenv[i].toLowerCase();

			if( f.equals( family ) )
				return true;
		}

		return false;
	}

	/**
	 * Get a predefined font by its <code>key</code>, or create a new one if
	 * <code>key</code> does not name any font. The {@link
	 * #get_font(String,Font)} and {@link #get_font(String,Graphics)} methods
	 * should always be preferred to this one.
	 */
	public Font
	get_font( String key )
	{
		Font f = fonts.get( key );

		if( f == null )
			f = new Font( "lucida", Font.PLAIN, 12 );

		return f;
	}

	/**
	 * Get a predefined font by its <code>key</code>, or the actual
	 * <code>graphics</code> font if the requested font is not available.
	 */
	public Font
	get_font( String key, Graphics graphics )
	{
		Font f = fonts.get( key );

		if( f == null )
			return graphics.getFont();

		return f;
	}

	/**
	 * Get a predefined font by its <code>key</code>, or the
	 * <code>fallback</code> font if the requested font is not available.
	 */
	public Font
	get_font( String key, Font fallback )
	{
		Font f = fonts.get( key );

		if( f == null )
			return fallback;

		return f;
	}

// Commands

	/**
	 * Map a font <code>family</code> name to <code>URL</code> (most often the
	 * <code>URL</code> is part of a jar file). Each time {@link
	 * #add_font(String,String,int,int)} will be used with <code>family</code>
	 * as argument the mapped font will be used instead. Also {@link
	 * #family_exists(String)} will always return true for mapped families.
	 * @throws IOException if the <code>URL</code> cannot be open or read.
	 * @throws NotFoundException if the <code>URL</code> cannot be located.
	 * @throws SingletonException if the <code>family</code> is already mapped.
	 * @throws FontFormatException if the <code>URL</code> does not point on a
	 * valid font.
	 */
	public void
	map_family( String family, String URL )
		throws IOException, FontFormatException, NotFoundException, SingletonException
	{
		FileInputStream in;
		java.net.URL    url = ClassLoader.getSystemResource( URL );
		Font            f;
		Object          o;

		if( url != null )
		{
			in = new FileInputStream( url.getPath() );
			f  = Font.createFont( Font.TRUETYPE_FONT, in );
			o  = map.put( family, f );

			in.close();

			if( o != null )
			{
				map.put( family, (Font) o );
				throw new SingletonException(
					"family '"+family+"' is already mapped" );
			}
		}
		else
		{
			throw new NotFoundException(
				"cannot find URL '"+URL+"'" );
		}
		
	}

	/**
	 * Add or replace a font whose key will be <code>name</code>. The font will
	 * pertain to <code>family</code>, having <code>style</code> and
	 * <code>size</code>. It is not checked that the font <code>family</code>
	 * exits.
	 * @throws SingletonException if a mapping with <code>name</code> already
	 * exists.
	 */
	public void
	add_font( String name, String family, int style, int size )
		throws SingletonException
	{
		Font   f = map.get( family );
		Object o;

		if( f != null )
		     o = fonts.put( name, f.deriveFont( style, size ) );
		else o = fonts.put( name, new Font( family, style, size ) );

		if( o != null )
		{
			fonts.put( name, (Font) o );
			throw new SingletonException(
				"a font is already attached to name '"+name+"'" );
		}
	}
}
