package book.swing;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;

/**
 * Simple 2D plotter for curves.
 *
 * @author Antoine Dutot
 * @since 20030519
 */
public class Plotter2d
	extends JPanel
	implements ComponentListener
{
// Constants

	private static final long serialVersionUID = 1041306906512549155L;
	protected static final Color COLOR_BG     = new Color( 240, 240, 240 );
	protected static final Color COLOR_BORDER = new Color(  90,  90,  90 );
	protected static final Color COLOR_AXIS   = new Color( 150, 150, 150 );
	protected static final Color COLOR_TITLE  = new Color(  90,  90,  90 );

// Attributes -- Painting

	/**
	 * Rendering context.
	 */
	protected SwingContext ctx;

	/**
	 * Font.
	 */
	protected Font font;

	/**
	 * Graph bounds.
	 */
	protected double xmin = -1, xmax = 1, ymin = -1, ymax = 1;

	/**
	 * Translation of the graph along X and Y in pixels space.
	 */
	protected int offset_x, offset_y;

	/**
	 * Bounds of the component where the graph is painted in pixel space.
	 */
	protected int area_width = 10, area_height = 10;

	/**
	 * Scaling factors to switch from graph coordinates to pixel coordinates.
	 */
	protected double scalex = 1.0, scaley = 1.0;

	/**
	 * Set when the component where drawing occurs changed and scaling factors
	 * need to be recomputed. This flag is set in {@link
	 * #componentResized(ComponentEvent)} and removed in {@link
	 * #paint(Graphics2D)}. When this flag is true, the {@link
	 * #compute_transform()} method must be called.
	 */
	protected boolean scale_changed = true;

	/**
	 * Graph title.
	 */
	protected String title;

	/**
	 * Show tick labels?.
	 */
	protected boolean show_labels = true;

	/**
	 * Keep the axis orthonormal?.
	 */
	protected boolean keep_ratio = false;

	/**
	 * List of curves.
	 */
	protected ArrayList<PlotterCurve> curves = new ArrayList<PlotterCurve>();

	/**
	 * Show only <code>window</code> points?. If negative zero all points are
	 * shown.
	 */
	protected int window = 0;

	/**
	 * Legend for the window size (avoid to reset it at every step).
	 */
	protected String winstr = "";

// Constructors

	public
	Plotter2d( SwingContext context, String title )
	{
		ctx = context;
		this.title = title;
		setOpaque( true );
		addComponentListener( this );
		font = context.get_font( "plain-small", getFont() );
	}

// Accessors

	public SwingContext
	get_swing_context()
	{
		return ctx;
	}

	public int
	get_window()
	{
		return window;
	}

// Accessors -- Coordinates conversion.

	/**
	 * Coordinate x from graph space into pixel space.
	 */
	public int
	xx( double x )
	{
		return( (int) ( ( x - xmin ) * scalex ) ) + offset_x;
	}

	/**
	 * Inverse of {@link #xx(double)}.
	 */
	public double
	xx_inv( int x )
	{
		return( ( ( x - offset_x ) / scalex ) + xmin );
	}

	/**
	 * Coordinate y from graph space into pixel space. This is not the same
	 * command than {@link #xx(double)} since we may have to invert the Y axis
	 * (drawing space origin is at bottom-left, pixel space origin at
	 * top-left).
	 */
	public int
	yy( double y )
	{
		return( area_height - ( (int) ( ( y - ymin ) * scaley ) ) ) - offset_y;
	}

	/**
	 * Inverse of {@link #yy(double)}.
	 */
	public double
	yy_inv( int y )
	{
		return( ( ( ( area_height - y ) - offset_y ) / scaley ) + ymin );
	}

	/**
	 * Lenght along X from graph space into pixel space.
	 */
	public int
	ww( double w )
	{
		return (int) ( w * scalex );
	}

	/**
	 * Inverse of {@link #ww(double)}.
	 */
	public double
	ww_inv( int w )
	{
		return w / scalex;
	}

	/**
	 * Lenght along Y from graph space into pixel space.
	 */
	public int
	hh( double h )
	{
		return (int) ( h * scaley );
	}

	/**
	 * Inverse of {@link #hh(double)}.
	 */
	public double
	hh_inv( int h )
	{
		return h / scaley;
	}

// Commands

	public void
	set_bounds( double xmin, double xmax, double ymin, double ymax )
	{
		this.xmin = xmin;
		this.xmax = xmax;
		this.ymin = ymin;
		this.ymax = ymax;
		scale_changed = true;
	}

	/**
	 * Show only the <code>width</code> more recent points. Setting
	 * <code>width</code> to zero or less shows all points.
	 */
	public void
	set_window( int width )
	{
		window = width;

		if( window > 0 )
			winstr = Integer.toString( window ) + " points";
	}

	public void
	add_curve( PlotterCurve curve )
	{
		curves.add( curve );
		curve.set_plotter( this );
	}

	public void
	paintComponent( Graphics g )
	{
		paint( (Graphics2D) g );
	}

	protected void
	paint( Graphics2D g )
	{
		if( scale_changed )
		{
			compute_transform();
			scale_changed = false;
		}

		g.setColor( COLOR_BG );
		g.fillRect( 0, 0, area_width+1, area_height+1 );
		g.setColor( COLOR_BORDER );
		g.drawRect( 0, 0, area_width, area_height );
		paint_axis( g );
		paint_title( g );
		paint_legend( g );

		int n = curves.size();

		for( int i=0; i<n; ++i )
		{
			paint_curve( g, curves.get( i ) );
		}
	}

	protected void
	paint_axis( Graphics2D g )
	{
		g.setColor( COLOR_AXIS );
		g.drawLine( xx(xmin), yy(0), xx(xmax), yy(0) );
		g.drawLine( xx(0), yy(ymin), xx(0), yy(ymax) );
	}

	protected void
	paint_title( Graphics2D g )
	{
	//	g.setFont( font );
	//	g.setColor( COLOR_TITLE );
	//	g.drawString( title, 5, 15 );
	}

	protected void
	paint_legend( Graphics2D g )
	{
		int n = curves.size();
		int y = font.getSize()+3;

		g.setFont( font );

		for( int i=0; i<n; ++i )
		{
			PlotterCurve c = curves.get( i );

			if( c.get_name() != null & c.get_name().length() > 0 )
			{
				g.setColor( c.get_color() );
				g.drawString( c.get_name(), 5, y );
				y += font.getSize()+3;
			}
		}

		if( window > 0 )
		{
			g.setColor( COLOR_TITLE );
			g.drawString( winstr, 5, y );
		}
	}

	protected void
	paint_curve( Graphics2D g, PlotterCurve curve )
	{
		curve.paint( g );
	}

	/**
	 * Compute scaling factors and translations. As the tree coordinates are
	 * not expressed in the same space as the area where it is drawn (one is in
	 * arbitrary coordinates specified by {@link #bounds}, the other is in
	 * pixels specified by {@link #area_width} and {@link #area_height}).
	 * Furthermore, the tree as an aspect ratio that may not fit the component
	 * aspect ratio.
	 */
	public void
	compute_transform()
	{
		double bw, bh, pw, ph;
		
		pw   = area_width  = getWidth()  - ( 1 + offset_x * 2 );	// area width in pixel coordinates
		ph   = area_height = getHeight() - ( 1 + offset_y * 2 );	// area height in pixel coordinates
		bw   = xmax - xmin;											// graph width in graph coordinates
		bh   = ymax - ymin;											// graph height in graph coordinates

		// Now, compute the scale factors.

		if( keep_ratio )
		{
			if( pw <= 0 || ph <= 0 )
			{
				// Not known yet...

				scalex = scaley = 1;
			}
			else
			{
				if( ( bw / bh ) > ( pw / ph ) )
				     scalex = scaley = pw / bw;
				else scalex = scaley = ph / bh;
			}
		}
		else
		{
			// Fit the available space.

			scalex = pw / bw;
			scaley = ph / bh;
		}
	}

// Events

	public void
	componentHidden( ComponentEvent e )
	{
	}

	public void
	componentMoved( ComponentEvent e )
	{
	}

	public void
	componentResized( ComponentEvent e )
	{
		area_width    = getWidth();
		area_height   = getHeight();
		scale_changed = true;
	}

	public void
	componentShown( ComponentEvent e )
	{
	}
}
