package book.swing;

import book.*;

import java.io.*;
import java.awt.*;
import javax.swing.*;

/**
 * Utilities to use Swing.
 *
 * @author Antoine Dutot
 * @since 20030417
 */
public class SwingContext
{
// Attributes -- Fonts

	/**
	 * Fonts.
	 */
	public FontContext fonts;

// Constructors

	/**
	 * New swing context. This setups default fonts using the "arial" family at
	 * 11, 9 and 20 point.
	 */
	public
	SwingContext()
	{
		this( "arial", 11, 9, 20 );
	}

	/**
	 * Make a copy of <code>other</code>.
	 */
	public
	SwingContext( SwingContext other )
	{
		fonts = new FontContext( other.fonts );
	}

	/**
	 * New swing context. This setups default fonts from <code>family</code>
	 * using sizes <code>normal</code>, <code>small</code> and
	 * <code>big</code>.
	 */
	public
	SwingContext( String family, int normal, int small, int big )
	{
		fonts = new FontContext( family, normal, small, big );
	}

	/**
	 * New swing context. This setups default fonts from <code>family</code>
	 * using sizes <code>normal</code>, <code>small</code> and
	 * <code>big</code> mapping <code>family</code> to <code>URL</code>
	 * in the {@link book.swing.FontContext}.
	 * @see book.swing.FontContext#FontContext(String,int,int,int,String)
	 */
	public
	SwingContext( String family, int normal, int small, int big, String URL )
		throws IOException, FontFormatException, NotFoundException
	{
		fonts = new FontContext( family, normal, small, big, URL );
	}

// Accessors

	/**
	 * Font context.
	 */
	public FontContext
	get_font_context()
	{
		return fonts;
	}

	/**
	 * Get a predefined font by its <code>key</code>, or create a new one if
	 * <code>key</code> does not name any font. The {@link
	 * #get_font(String,Font)} and {@link #get_font(String,Graphics)} methods
	 * should always be preferred to this one.
	 */
	public Font
	get_font( String key )
	{
		return fonts.get_font( key );
	}

	/**
	 * Get a predefined font by its <code>key</code>, or the actual
	 * <code>graphics</code> font if the requested font is not available.
	 */
	public Font
	get_font( String key, Graphics graphics )
	{
		return fonts.get_font( key, graphics );
	}

	/**
	 * Get a predefined font by its <code>key</code>, or the
	 * <code>fallback</code> font if the requested font is not available.
	 */
	public Font
	get_font( String key, Font fallback )
	{
		return fonts.get_font( key, fallback );
	}

	/**
	 * Try to load an icon stored at <code>URL</code> in the jars of the
	 * classpath. Returns null if failed.
	 */
	public static ImageIcon
	load_icon( String URL )
	{
		java.net.URL url = ClassLoader.getSystemResource( URL );

		if( url != null )
			return new ImageIcon( url );

		return null;
	}

	/**
	 * Try to load an icon stored at <code>URL</code> in the jars of the
	 * classpath using the default toolkit. Returns null if failed.
	 */
	public static Image
	load_image( String URL )
	{
		java.net.URL url = ClassLoader.getSystemResource( URL );

		if( url != null )
			return Toolkit.getDefaultToolkit().createImage( url );

		return null;
	}

// Commands

	/**
	 * Try to change the look&amp;feel to something smarter. A good one is
	 * "net.sourceforge.mlf.metouia.MetouiaLookAndFeel" it is used by default
	 * is <code>laf</code> is null.
	 */
	public static void
	change_laf( String laf )
	{
		if( laf == null )
			laf = "net.sourceforge.mlf.metouia.MetouiaLookAndFeel";

		if( laf != null )
		{
			try
			{
				// The good one: "net.sourceforge.mlf.metouia.MetouiaLookAndFeel".
				UIManager.setLookAndFeel( laf );
			}
			catch( ClassNotFoundException e )          { System.err.println( "cannot change the look and feel: class not found: " + e.getMessage() ); }
			catch( InstantiationException e )          { System.err.println( "cannot change the look and feel: instantiation error: " + e.getMessage() ); }
			catch( IllegalAccessException e )          { System.err.println( "cannot change the look and feel: illegal access: " + e.getMessage() ); }
			catch( UnsupportedLookAndFeelException e ) { System.err.println( "cannot change the look and feel: unsupported LaF: " + e.getMessage() ); }
		}
		else
		{
			System.err.println( "no look and feel specified: you are using the insane metal LAF, very bad." );
		}
	}
}
