package book.swing;

import java.awt.*;

/**
 * Text.
 *
 * <p>Utility class to automatically scale, center, and display text. This
 * class contains a string and allows to display it centered or aligned inside
 * a given area. It handles font resizing.</p>
 *
 * <p>It is first created using a text string and a {@link java.awt.Font}. Then
 * at any time, given a {@link java.awt.Graphics2D}, it can display the text
 * anywhere, centered arround a point, or left or right alligned.</p>
 *
 * <p>The class also gives the possibility to zoom the text by specifying a
 * width in pixels where it should fit.</p>
 *
 * @author Antoine Dutot
 * @since 20030418
 */
public class Text
{
// Attributes

	/**
	 * Text value.
	 */
	protected String text;

	/**
	 * Text font.
	 */
	protected Font font;

	/**
	 * Font metrics.
	 */
	protected FontMetrics metrics;

	/**
	 * Width in pixels of the text.
	 */
	protected int text_width;

	/**
	 * Height int pixels of the text.
	 */
	protected int text_height;

	/**
	 * Distance from left-most border to center.
	 */
	protected int center_offx;

	/**
	 * Distance from baseline to center.
	 */
	protected int center_offy;

	/**
	 * True if the font was modified and need to be recomputed.
	 */
	protected boolean font_modified = true;

	/**
	 * True if the text was modified and need to be recomputed.
	 */
	protected boolean text_modified = true;

// Constructors

	public
	Text( String text, Font font )
	{
		set_font( font );
		set_text( text );
	}

// Accessors

	/**
	 * Font.
	 */
	public Font
	get_font()
	{
		return font;
	}

	/**
	 * Text value.
	 */
	public String
	get_text()
	{
		return text;
	}

// Commands

	/**
	 * Draw the text horizontally centered arroung <code>x</code>. The text is
	 * vertically centered arroung <code>y</code>, therefore <code>y</code> is
	 * not the baseline.
	 * @see #draw_left_aligned(int,int,Graphics2D)
	 * @see #draw_right_aligned(int,int,Graphics2D)
	 */
	public void
	draw_centered( int x, int y, Graphics2D g )
	{
		g.setFont( font );
		recompute_lengths( g );
		g.drawString( text, x-center_offx, y+center_offy );
	}

	/**
	 * Draw the text with its left border along <code>x</code>. The text is
	 * vertically centered arroung <code>y</code>, therefore <code>y</code> is
	 * not the baseline.
	 * @see #draw_centered(int,int,Graphics2D)
	 * @see #draw_right_aligned(int,int,Graphics2D)
	 */
	public void
	draw_left_aligned( int x, int y, Graphics2D g )
	{
		g.setFont( font );
		recompute_lengths( g );
		g.drawString( text, x, y+center_offy );
	}

	/**
	 * Draw the text with its right border along <code>x</code>. The text is
	 * vertically centered arroung <code>y</code>, therefore <code>y</code> is
	 * not the baseline.
	 * @see #draw_centered(int,int,Graphics2D)
	 * @see #draw_left_aligned(int,int,Graphics2D)
	 */
	public void
	draw_right_aligned( int x, int y, Graphics2D g )
	{
		g.setFont( font );
		recompute_lengths( g );
		g.drawString( text, x-text_width, y+center_offy );
	}

// Commands

	/**
	 * Set the <code>font</code>.
	 */
	public void
	set_font( Font font )
	{
		this.font = font;
		font_modified = true;
	}

	/**
	 * Change the <code>text</code> value.
	 */
	public void
	set_text( String text )
	{
		this.text = text;
		text_modified = true;
	}

	/**
	 * Change the font <code>size</code>.
	 * @see #fit_width(int,Graphics2D)
	 */
	public void
	set_font_size( int size )
	{
		if( size != font.getSize() )
		{
			font = font.deriveFont( (float) size );
			font_modified = true;
		}
	}

	/**
	 * Zoom the font so that the text width fit in <code>width</code>. The text
	 * is not zoom under 4 points.
	 * @see #set_font_size(int)
	 */
	public void
	fit_width( int width, Graphics2D g )
	{
		recompute_lengths( g );

		float ratio = ((float)width) / ((float)text_width);
		float size  = font.getSize() * ratio;
		int   isize = (int) size;

		if( isize > 4 )
			set_font_size( isize );
	}

	/**
	 * Recompute all lengths using {@link #metrics}. This method is to be
	 * called only before painting. The class remembers when settings changed.
	 */
	protected void
	recompute_lengths( Graphics2D g )
	{
		if( font_modified )
		{
			metrics = g.getFontMetrics();
		}

		if( text_modified || font_modified )
		{
			text_width  = metrics.stringWidth( text );
			text_height = ( metrics.getAscent() + metrics.getDescent() );
			center_offx = ( text_width / 2 );
			center_offy = ( text_height / 2 ) - metrics.getDescent();
		}

		font_modified = false;
		text_modified = false;
	}
}
