package embeddings.dictionary;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.jena.rdf.model.Model;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.RDFNode;
import org.apache.jena.rdf.model.Statement;
import org.apache.jena.rdf.model.StmtIterator;

import datasets.Dataset;
import util.Constants;

public class DictionaryHelper {

	private Dictionary dictionary;

	public DictionaryHelper() {
		this.dictionary = new Dictionary();
	}

	/**
	 * Reads the dictionary from a given file path
	 * 
	 * @param dataFolderPath
	 */
	public Dictionary readDictionary(String filePath, Dataset dataset) {
		Map<Integer, String> id2relations = dataset.readMap(filePath);
		Map<String, Integer> rel2ID = id2relations.entrySet().stream()
				.collect(Collectors.toMap(Map.Entry::getValue, Map.Entry::getKey));

		return new Dictionary(rel2ID, id2relations);

	}

	/**
	 * Reads a single model from path and creates the corresponding dictionary
	 * 
	 * @param filePath
	 * @return
	 */
	public Dictionary createDictionary(String filePath) {
		Model model = ModelFactory.createDefaultModel();
		model.read(filePath + "/train.txt", "TTL");
		return createDictionary(model);
	}

	/**
	 * Creates the relations dictionary for a given model
	 * 
	 * @param model
	 * @return
	 */
	public Dictionary createDictionary(Model model) {
		StmtIterator stmtIterator = model.listStatements();
		while (stmtIterator.hasNext()) {
			Statement curStmt = stmtIterator.next();

			String predicate = curStmt.getPredicate().toString();
			RDFNode object = curStmt.getObject();

			if (object.isResource()) {
				dictionary.addRelation(predicate);
			}
		}
		return dictionary;
	}

	/**
	 * Writes the dictionary to file
	 * 
	 * @param dictionary
	 * @param dataFolderPath
	 */
	public void saveDict2File(String dataFolderPath) {
		printMap(dictionary.getId2Relations(), dataFolderPath + Constants.REL_DICT_FILE);
	}

	/**
	 * Writes a map to a file: each key value pair in a new line separated by \t
	 * 
	 * @param dictionary
	 * @param savedFile
	 */
	public void printMap(Map<?, ?> dictionary, String savedFile) {
		StringBuilder builder = new StringBuilder();
		dictionary.forEach((k, v) -> {
			builder.append(k).append("\t").append(v.toString()).append("\n");
		});

		try (BufferedWriter bw = new BufferedWriter(new FileWriter(savedFile))) {
			bw.append(builder);
			bw.flush();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Reads an entire folder contents to a single model
	 * 
	 * @param dataFolderPath
	 * @return
	 */
	public Model readDatasetFolder(String dataFolderPath) {
		Model model = ModelFactory.createDefaultModel();
		File folder = new File(dataFolderPath);
		if (folder != null && folder.isDirectory() && folder.listFiles().length > 0) {
			for (String fileName : folder.list()) {
				model.read(dataFolderPath + "/" + fileName);
			}
		}
		return model;
	}

	public Dictionary getDictionary() {
		return dictionary;
	}

}
