package fact_check.path;

import java.io.File;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.stream.Collectors;

import org.apache.jena.rdf.model.ResourceFactory;
import org.apache.jena.sparql.lang.sparql_11.ParseException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import embeddings.EmbeddingModel;
import embeddings.dictionary.Dictionary;
import fact_check.query.QueryExecutioner;
import fact_check.score.NPMICalculator;
import fact_check.score.OccurrencesCounter;
import graph_search.Matrix;
import graph_search.algorithms.PropertySearch;
import graph_search.algorithms.SearchAlgorithm;
import util.LogUtils;
import util.PrintToFileUtils;

public class PathCreator {
	private static final Logger LOGGER = LoggerFactory.getLogger(PathCreator.class);

	private EmbeddingModel emodel;

	private Dictionary dictionary;

	private Matrix matrix;

	private int k;
	
	private QueryExecutioner sparqlExec;

	public PathCreator(Dictionary dictionary, EmbeddingModel emodel, Matrix matrix, int k, QueryExecutioner sparqlExec) {
		this.dictionary = dictionary;
		this.emodel = emodel;
		this.matrix = matrix;
		this.k = k;
		this.sparqlExec = sparqlExec;
	}

	/**
	 * Calculates the meta-paths for a single edge
	 * 
	 * @param edge
	 * @return the paths found
	 */
	public Set<Property> getMetaPaths(String edge, int l, boolean isLoopsAllowed) {
		Map<String, Integer> rel2ID = dictionary.getRelations2ID();

		// get corresponding ids for embeddings
		int edgeID = rel2ID.getOrDefault(edge, -1);

		// if any isn't found, return. It's an error
		if (edgeID < 0) {
			throw new IllegalArgumentException("Could not find the given predicate's embedding");
		}

		// search for property combos
		SearchAlgorithm propertyCombos = new PropertySearch(matrix, emodel);
		Set<Property> propertyPaths = propertyCombos.findPaths(edgeID, k, l, isLoopsAllowed);

		return propertyPaths;
	}

	/**
	 * Calculates the meta-paths for multiple edges
	 * 
	 * @param edges
	 * @return the edge to paths found map
	 */
	public Map<String, Set<Property>> getMultipleMetaPaths(Set<String> edges, int l, boolean isLoopsAllowed) {
		ConcurrentMap<String, Set<Property>> metaPaths = new ConcurrentHashMap<String, Set<Property>>();
		StringBuffer buffer = new StringBuffer();
		edges.parallelStream().forEach(edge -> {
			Set<Property> propertyPaths = getMetaPaths(edge, l, isLoopsAllowed);
			LOGGER.info("Processing meta-path: " + edge);
			// calculate pnpmi for each meta-path
			org.apache.jena.rdf.model.Property edgeProp = ResourceFactory.createProperty(edge);
			
			OccurrencesCounter c = new OccurrencesCounter(edgeProp, sparqlExec, false);
			for (Property path : propertyPaths) {
				if (c.getSubjectTypes().isEmpty() || c.getObjectTypes().isEmpty()) {
					path.setPathNPMI(0);
				}
				NPMICalculator task = new NPMICalculator(path, c, dictionary.getId2Relations());
				try {
					task.calculatePMIScore();
				} catch (ParseException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			
			metaPaths.put(edge, propertyPaths);
			LOGGER.info(metaPaths.size()+" - Processed meta-path: " + edge);
			buffer.append("\nPredicate:").append("\t").append(edge);
			addPrintableMetaPaths(buffer, propertyPaths);
		});
		print(buffer.toString());
		return metaPaths;
	}

	public EmbeddingModel getEmodel() {
		return emodel;
	}

	public void setEmodel(EmbeddingModel emodel) {
		this.emodel = emodel;
	}

	public Dictionary getDictionary() {
		return dictionary;
	}

	public void setDictionary(Dictionary dictionary) {
		this.dictionary = dictionary;
	}

	public void addPrintableMetaPaths(StringBuffer builder, Set<Property> propertyPaths) {
		List<Property> result = propertyPaths.stream()
				.sorted(Comparator.comparingDouble(Property::getPathLength).thenComparing(Property::getPathCost))
				.collect(Collectors.toList());

		builder.append("\nLength").append("\t").append("A* score").append("\t").append("Path(s)");
		for (Property path : result) {
			builder.append("\n").append(path.getPathLength()).append("\t").append(path.getPathCost()).append("\t")
					.append(PropertyHelper.translate2DirectedIRI(path, dictionary.getId2Relations()));
		}
	}

	public void print(String string) {
		int index = 1;
		String fileName = index + "_metapaths_" + k + ".txt";
		File file = new File(fileName);
		while (file.exists()) {
			String newFileName = ++index + "_metapaths_" + k + ".txt";
			file = new File(newFileName);
		}
		LogUtils.printTextToLog(file.getAbsolutePath());
		PrintToFileUtils.printStringToFile(string, file);
	}

}
