/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import org.miv.graphstream.graph.Edge;
import org.miv.graphstream.graph.Element;
import org.miv.graphstream.graph.Graph;
import org.miv.graphstream.graph.GraphListener;
import org.miv.graphstream.graph.Node;
import org.miv.graphstream.graph.implementations.DefaultGraph;
import org.miv.graphstream.io.DynamicGraphWriterHelper;
import org.miv.graphstream.io.GraphParseException;
import org.miv.graphstream.io.GraphReaderListenerHelper;
import org.miv.graphstream.io.GraphReaderWeb;
import org.miv.graphstream.io.GraphWriterHelper;
import org.miv.graphstream.ui.GraphViewer;
import org.miv.graphstream.ui.GraphViewerRemote;
import org.miv.graphstream.ui.swing.SwingGraphViewer;
import org.miv.util.Environment;
import org.miv.util.NotFoundException;

/**
 * Explore the web starting from a given page, but tries restrict to known blogs domains.
 * 
 * @author Antoine Dutot
 */
public class BlogFerret extends JFrame implements ActionListener, GraphListener
{
    private static final long serialVersionUID = 1L;

    protected Graph graph;
    
    protected GraphViewer viewer;
    
    protected GraphViewerRemote remote;

    protected DynamicGraphWriterHelper dynOut;
    
    protected JPanel commands;
    
    protected JButton save;
    
    protected JButton exit;
    
    protected JPanel info;
    
    protected JLabel dictSize;
    
    protected int nLeafs = 0;
    
	public static void main( String args[] )
	{
		new BlogFerret( args );
	}
	
	public BlogFerret( String args[] )
	{
		Environment env = Environment.getGlobalEnvironment();
		
		env.setParameter( "display", "true" );
		env.readCommandLine( args );
		
		String  start   = env.getParameter( "start" );
        boolean display = env.getBooleanParameter( "display" );
		
		if( start == null || start.length() == 0 )
			start = askForStartURL();
		
		if( start == null || start.length() == 0 )
		{
			JOptionPane.showMessageDialog( this,  "No start URL given, cannot process.", "Error", JOptionPane.ERROR_MESSAGE );
			System.exit( 1 );
		}
		
		graph    = new DefaultGraph( start );
		viewer   = display ? new SwingGraphViewer( graph, true ) : null;
		remote   = display ? viewer.newViewerRemote() : null;
		commands = new JPanel( new BorderLayout() );
		dictSize = new JLabel( "0" );
		save     = new JButton( "Save" );
		info     = new JPanel();
		exit     = new JButton( "Exit" );
		
		if( display )
		{
			graph.setAttribute( "stylesheet", STYLE_SHEET );
			add( (JComponent)viewer.getComponent(), BorderLayout.CENTER );
		}
		
		graph.addGraphListener( this );
		save.addActionListener( this );
		exit.addActionListener( this );
		commands.add( save, BorderLayout.WEST );
		commands.add( exit, BorderLayout.EAST );
		commands.add( info, BorderLayout.CENTER );
		info.add( new JLabel( "Pages : " ) );
		info.add( dictSize );
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		add( commands, BorderLayout.SOUTH );
		setSize( 400, 400 );
		setVisible( true );
		
		String  output    = env.getParameter( "output" );
        int     depth     = (int)env.getNumberParameter( "depth" );
		boolean debug     = env.getBooleanParameter( "debug" );
		boolean onlySites = env.getBooleanParameter( "onlySites" );
		String  start2    = env.getParameter( "start2" );

		if( depth <= 0 )
			depth = 3;

        try
        {
    		if( output != null && output.length() > 0 )
    		{
    			dynOut = new DynamicGraphWriterHelper();
    			dynOut.begin( graph, output );
    		}
    		
    		if( start2 != null && start2.length() <= 0 )
    			start2 = null;
	
			parse( start, start2, depth, onlySites, debug );
	
			if( dynOut != null )
				dynOut.end();
        }
        catch( Exception e )
        {
			JOptionPane.showMessageDialog( this, e.getMessage(), "Error while fetching web informations.", JOptionPane.ERROR_MESSAGE );
			e.printStackTrace();
			System.exit( 1 );
        }
	}
	
	protected String askForStartURL()
	{
		return (String) JOptionPane.showInputDialog( this, "Give a start URL in large blogs domains (blogspot, wordpress, etc.)", "Please give a start URL", JOptionPane.QUESTION_MESSAGE );
	}
	
	protected void parse( String url, String url2, int depth, boolean onlySites, boolean debug )
		throws NotFoundException, GraphParseException, IOException
	{
		GraphReaderWeb in = new GraphReaderWeb( depth, onlySites, debug );
		GraphReaderListenerHelper helper = new GraphReaderListenerHelper( graph );

		for( String bl: BLACK_LIST )
			in.addToBlackList( bl );
		
		in.setOnlySites( onlySites );
		in.restrictToSites( SITES_RESTRICT );
		in.setRemoveQueries( true );
		in.setLabelAttributes( true );
		in.setWeightAttribute( "weight" );
		in.addGraphReaderListener( helper );
		in.begin( url );
		
 		if( url2 != null )
			in.addSite( url2 );
		
		while( in.nextEvents() )
		{
			dictSize.setText( String.format( "%d (%dn %dl %de)", in.getVisitedPageCount(),
					graph.getNodeCount(), nLeafs, graph.getEdgeCount() ) );
			
			if( dynOut != null )
				dynOut.flush();
		}
	
		in.end();
		dictSize.setText( String.format( "%d", in.getVisitedPageCount() ) );
		
		JOptionPane.showMessageDialog( this, "Finished !", "Information", JOptionPane.INFORMATION_MESSAGE );
		System.out.printf( "Finished !!" );
	}

	public void actionPerformed( ActionEvent e )
    {
		if( e.getSource() == exit )
		{
			System.exit( 0 );
		}
		else if( e.getSource() == save )
		{
			JFileChooser fc = new JFileChooser();
			
			if( fc.showSaveDialog( this ) == JFileChooser.APPROVE_OPTION )
			{
				String fileName = fc.getSelectedFile().getAbsolutePath();
				
				saveGraph( fileName );
			}
		}
    }
	
	protected void saveGraph( String fileName )
	{
		try
		{
			GraphWriterHelper out = new GraphWriterHelper( graph );
			out.write( fileName );
			JOptionPane.showMessageDialog( this,
					String.format( "Graph saved in '%s'", fileName ),
					"Save graph", JOptionPane.INFORMATION_MESSAGE );
		}
		catch( Exception e )
		{
			JOptionPane.showMessageDialog( this,
					String.format( "Error : %s", e.getMessage() ),
					"Error while saving graph", JOptionPane.ERROR_MESSAGE );
		}
	}
	
	protected static String[] SITES_RESTRICT = {
			"blogspot.com",
			"typepad.com",
			"blogs.com",
			"wordpress.com",
			"livejournal.com",
			"blogs.cnn.com"
		};
	
	protected static String[] BLACK_LIST = {
		"http://wordpress.com.*",
		"http://feeds.wordpress.com.*",
		"http://s.wordpress.com.*",
		"http://digg.com.*"
	};
	
	protected static final String STYLE_SHEET =
		"graph { width:4px; }" +
		"edge { color:#505050; arrow-width: 3px; arrow-length:6px; }" +
		"node { text-align:aside; }" +
		"edge.leaf { color:#D0D0D0; }" +
		"node.leaf { color:#C0C0C0; }";

	public void afterEdgeAdd( Graph graph, Edge edge )
    {
		Node n0 = edge.getNode0();
		Node n1 = edge.getNode1();
		
		if( n0.getDegree() == 1 
		 || n1.getDegree() == 1 )
		{
			edge.setAttribute( "class", "leaf" );
		}
		else if( n0.getDegree() == 2
		      || n1.getDegree() == 2 )
		{
			if( n0.hasAttribute( "class" ) ) nLeafs--;
			if( n1.hasAttribute( "class" ) ) nLeafs--;
			edge.removeAttribute( "class" );
			n0.removeAttribute( "class" );
			n1.removeAttribute( "class" );
		}
    }

	public void afterNodeAdd( Graph graph, Node node )
    {
		nLeafs++;
		node.setAttribute( "class", "leaf" );
    }

	public void attributeChanged( Element element, String attribute, Object oldValue,
            Object newValue )
    {
		if( element instanceof Edge )
		{
			if( attribute.equals( "weight" ) )
			{
				int w = ((Number)newValue).intValue();
				element.setAttribute( "width", Math.log( w ) );
			}
		}
    }

	public void beforeEdgeRemove( Graph graph, Edge edge )
    {
    }

	public void beforeGraphClear( Graph graph )
    {
    }

	public void beforeNodeRemove( Graph graph, Node node )
    {
    }

	public void stepBegins(Graph graph, double time)
	{
	}
}