/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool;

import java.io.*;
import java.util.Map;

import org.miv.graphstream.io.*;
import org.miv.graphstream.graph.implementations.DefaultGraph;

/**
 * Use the GraphReaderWeb to read URLs recursively and output the corresponding
 * graph to a file.
 *
 * @author Antoine Dutot
 * @author Yoann Pign
 * @since 2007
 */
public class Ferret implements GraphReaderListener
{
	public static void main( String args[] )
	{
		try
		{
			new Ferret( args );
		}
		catch( Exception e )
		{
			e.printStackTrace();
		}
	}
	
	public Ferret( String args[] )
		throws IOException, GraphParseException
	{
		int     maxDepth  = 3;
		String  outputDGS = null;
		boolean showGraph = false;
		boolean showLog   = false;
		boolean onlySites = false;
		String  startURL  = null;
		
		for( int i=0; i<args.length; ++i )
		{
			if( args[i].startsWith( "-" ) )
			{
				if( args[i].equals( "-d" ) )
				{
					i++;
					
					if( i >= args.length )
						fatalError( "-d option needs an argument" );
					
					try
					{
						maxDepth = Integer.parseInt( args[i] );
					}
					catch( NumberFormatException e )
					{
						fatalError( "-d option needs an integer argument (%s)", e.getMessage() );
					}
				}
				else if( args[i].equals( "-s" ) )
				{
					showGraph = true;
				}
				else if( args[i].equals( "-x" ) )
				{
					showLog = true;
				}
				else if( args[i].equals( "-w" ) )
				{
					onlySites = true;
				}
				else if( args[i].equals( "-o" ) )
				{
					i++;
					
					if( i >= args.length )
						fatalError( "-o option needs an argument" );
					
					outputDGS = args[i];
				}
				else if( args[i].equals( "-h" ) )
				{
					System.err.printf( "%nUsage: %s [options] <URL>%n", Ferret.class.getName() );
					System.err.printf( "   The <URL> is mandatory and tells from wich web site to start.%n" );
					System.err.printf( "   Options are:%n" );
					System.err.printf( "     -s ................ open a window and show the graph as it is read.%n" );
					System.err.printf( "     -o <filename>...... output the graph to the given filename.%n" );
					System.err.printf( "     -d <depth> ........ maximum depth to explore links from the main URL (depth is an integer).%n" );
					System.err.printf( "     -w ................ only add site names to the graph, not all pages.%n" );
					System.err.printf( "     -x ................ show log informations (page misses, parse errors, etc.)%n" );
					System.err.printf( "%n" );
					
					System.exit( 1 );
				}
				else
				{
					fatalError( "unknow option %s", args[i] );
				}
			}
			else 
			{
				if( startURL == null )
				{
					startURL = args[i];
				}
				else
				{
					fatalError( "error: need exactly one URL to start exploring" );
				}
			}
		}
		
		if( showGraph == false && outputDGS == null )
			showLog = true;
		
		explore( startURL, maxDepth, showGraph, showLog, onlySites, outputDGS );
	}
	
	protected static void fatalError( String msg, Object...data )
	{
		System.err.printf( msg, data );
		System.err.printf( "%n" );
		System.exit( 1 );
	}
	
	protected PrintStream outputDGS = null;
	
	protected void explore( String startURL, int maxDepth, boolean showGraph, boolean showLog, boolean onlySites, String outputDGS )
		throws IOException, GraphParseException
	{
		DefaultGraph graph = null;
		GraphReaderWeb reader = new GraphReaderWeb( maxDepth, false, false );

		if( outputDGS != null )
		{
			if( ! outputDGS.endsWith( ".dgs" ) )
				outputDGS = String.format( "%s.dgs", outputDGS );
			
			this.outputDGS = new PrintStream( outputDGS );
			
			this.outputDGS.printf( "DGS003%n" );
			this.outputDGS.printf( "Web 0 0%n%n" );
			
			reader.addGraphReaderListener( this );
			
			// We do not use the GraphRecorder, since it would force us to use
			// a graph (an option that is enabled only if display is requested),
			// and would consume lots memory for nothing.
		}

		if( showGraph )
		{
			graph = new DefaultGraph( startURL );

			graph.display();
			reader.addGraphReaderListener( new GraphReaderListenerHelper( graph ) );
		}
		
		reader.setShowLog( showLog );
		reader.setOnlySites( onlySites );
		
		reader.begin( startURL );
		while( reader.nextEvents() ) {};
		reader.end();
		
		if( outputDGS != null )
		{
			this.outputDGS.flush();
			this.outputDGS.close();
		}
		
		System.out.printf( "FINISHED" );
		System.out.flush();
	}

	public void edgeAdded( String id, String from, String to, boolean directed, Map<String, Object> attributes ) throws GraphParseException
	{
		if( outputDGS != null )
			outputDGS.printf( "ae \"%s\" \"%s\" %s \"%s\"%n",
				id.replace( '"', ' ' ),
				from.replace( '"', ' ' ),
				directed ? ">" : "",
				to.replace( '"', ' ' ) );
	}

	public void edgeChanged( String id, Map<String, Object> attributes ) throws GraphParseException
	{
	}

	public void edgeRemoved( String id ) throws GraphParseException
	{
	}

	public void graphChanged( Map<String, Object> attributes ) throws GraphParseException
	{
	}

	public void nodeAdded( String id, Map<String, Object> attributes ) throws GraphParseException
	{
		System.err.printf( "> [%s]%n", id );
		
		if( outputDGS !=  null )
			outputDGS.printf( "an \"%s\"%n", id.replace( '"', ' ' ) );
	}

	public void nodeChanged( String id, Map<String, Object> attributes ) throws GraphParseException
	{
	}

	public void nodeRemoved( String id ) throws GraphParseException
	{
	}

	public void stepBegins( double time ) throws GraphParseException
	{
	}

	public void unknownEventDetected( String unknown ) throws GraphParseException
	{
	}
}