/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.miv.graphstream.algorithm.ConnectedComponents;
import org.miv.graphstream.graph.Edge;
import org.miv.graphstream.graph.Graph;
import org.miv.graphstream.graph.Node;
import org.miv.graphstream.graph.implementations.DefaultGraph;
import org.miv.graphstream.io.GraphWriter;
import org.miv.graphstream.io.GraphWriterFactory;
import org.miv.graphstream.ui.GraphViewer;
import org.miv.graphstream.ui.GraphViewerRemote;
import org.miv.graphstream.ui.swing.SwingGraphViewer;

/**
 * Show the various connected components and allow to save a chosen one to a file. 
 * @author Antoine Dutot
 */
public class SaveConnectedComponents extends JFrame implements ListSelectionListener, ActionListener
{
    private static final long serialVersionUID = 1L;

    ConnectedComponents cc;
    
    Graph graph;
    
    JPanel components;
    
    DefaultListModel listModel;
    
    JList list;
    
    JButton saveComponent;
    
    GraphViewer viewer;

    GraphViewerRemote remote;

    JFileChooser fileChooser;
    
	public static void main( String args[] )
	{
		new SaveConnectedComponents( args );
	}
	
	public SaveConnectedComponents( String args[] )
	{
		String fileName = getFileToRead( args );
		
		graph         = new DefaultGraph( fileName );
		listModel     = new DefaultListModel();
		list          = new JList( listModel );
		saveComponent = new JButton( "Save Component" );
		components    = new JPanel( new BorderLayout() );
		cc            = new ConnectedComponents( graph );
		viewer        = new SwingGraphViewer( graph, true );
		remote        = viewer.newViewerRemote();
		
		graph.setAttribute( "stylesheet", styleSheet );
		remote.setQuality( 1 );
		
		cc.setCountAttribute( "id" );
		saveComponent.setEnabled( false );
		components.add( new JLabel( "Components:" ), BorderLayout.NORTH );
		components.add( list, BorderLayout.CENTER );
		components.add( saveComponent, BorderLayout.SOUTH );
		saveComponent.addActionListener( this );
		list.addListSelectionListener( this );
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		add( components, BorderLayout.EAST );
		add( (JComponent)viewer.getComponent(), BorderLayout.CENTER );
		setVisible( true );
		setSize( 800, 600 );
		
		try
		{
			graph.read( fileName );
			cc.compute();
			int c = cc.getConnectedComponentsCount();
			cc.compute();
			
			for( int i=0; i<c; i++ )
				listModel.addElement( i );
		}
		catch( Exception e )
		{
			e.printStackTrace();
			System.exit( 1 );
		}
	}
	
	protected String getFileToRead( String args[] )
	{
		if( args.length <= 0 )
		{
			fileChooser = new JFileChooser();
			
			if( fileChooser.showOpenDialog( this ) == JFileChooser.APPROVE_OPTION )
			{
				File file = fileChooser.getSelectedFile();
				return file.getAbsolutePath();
			}
			else
			{
				System.err.printf( "Please choose a file or pass it as argument." ) ;
				System.exit( 1 );
			}
		}
		
		return args[0];
	}
	
	public void valueChanged( ListSelectionEvent e )
    {
		if( e.getValueIsAdjusting() == false )
		{
			if( list.getSelectedIndex() == -1 )
			{
				saveComponent.setEnabled( false );
				clearColors();
			}
			else
			{
				clearColors();
				saveComponent.setEnabled( true );
				setColor( list.getSelectedIndex() );
			}
		}
    }

	public void actionPerformed( ActionEvent e )
    {
		if( fileChooser == null )
			fileChooser = new JFileChooser();
		
		if( fileChooser.showSaveDialog( this ) == JFileChooser.APPROVE_OPTION )
		{
			String fileName = fileChooser.getSelectedFile().getAbsolutePath();
			
			GraphWriter out = GraphWriterFactory.writerFor( fileName );
			
			if( out != null )
			{
				int c = list.getSelectedIndex();
				
				try
				{
					out.begin( fileName, fileName );
					
					Iterator<? extends Node> nodes = graph.getNodeIterator();
	
					while( nodes.hasNext() )
					{
						Node node = nodes.next();
						
						if( node.getNumber( "id" ) == c )
							out.addNode( node.getId(), node.getAttributeMap() );
					}
					
					Iterator<? extends Edge> edges = graph.getEdgeIterator();
					
					while( edges.hasNext() )
					{
						Edge edge = edges.next();
						
						if( edge.getNode0().getNumber( "id" ) == c )
						{
							out.addEdge( edge.getId(),
									edge.getNode0().getId(),
									edge.getNode1().getId(),
									edge.isDirected(),
									edge.getAttributeMap() );
						}
					}
					
					out.end();
				}
				catch( IOException ee )
				{
					ee.printStackTrace();
					System.exit( 1 );
				}
			}
		}
    }
	
	public void clearColors()
	{
		Iterator<? extends Node> nodes = graph.getNodeIterator();

		while( nodes.hasNext() )
		{
			Node node = nodes.next();
			node.removeAttribute( "color" );
		}
	}
	
	public void setColor( int c )
	{
		Iterator<? extends Node> nodes = graph.getNodeIterator();

		while( nodes.hasNext() )
		{
			Node node = nodes.next();
			
			int id = (int) node.getNumber( "id" );

			if( id == c )
			     node.setAttribute( "color", Color.RED );
			else node.removeAttribute( "color" );
		}
	}

	protected static final String styleSheet =
		"node { width: 6px; text-align: aside; }" +
		"node.selected { width: 8px; color:red; }";
}
