/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool.graphed;

import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.miv.graphstream.graph.Edge;
import org.miv.graphstream.graph.Element;
import org.miv.graphstream.graph.Graph;
import org.miv.graphstream.graph.GraphListener;
import org.miv.graphstream.graph.Node;
import org.miv.graphstream.ui.GraphViewerListener;
import org.miv.graphstream.ui.Sprite;

/**
 * The listener for all events and the provider for actions.
 * 
 * @author Antoine Dutot
 */
public class Actions implements WindowListener, GraphViewerListener, GraphListener, ListSelectionListener
{
// Attributes
	
	protected Context ctx;

// Attributes -- Modes
	
    protected Action modeAddNodes, modeAddEdges;
    
    protected ModeSelectNodesAction  modeSelectNodes;
    
    protected ModeSelectEdgesAction  modeSelectEdges;

// Attributes -- Commands
    
    protected AddNodeAction addNodeAction;
    
    protected AddEdgeAction addEdgeAction;
    
    protected DelNodeAction delNodeAction;
    
    protected DelEdgeAction delEdgeAction;
    
// Attributes -- Files
    
    protected Action openAction, saveAction;

// Constructors
	
	/**
	 * Set of actions.
	 */
	public Actions( Context context )
	{
		ctx = context;
		ctx.graph.addGraphListener( this );
		
		buildActions();
	}
	
	protected void buildActions()
	{
		openAction = new OpenAction( ctx );
		saveAction = new SaveAction( ctx );
		
		modeAddNodes    = new ModeAddNodesAction( ctx );
		modeAddEdges    = new ModeAddEdgesAction( ctx );
		modeSelectNodes = new ModeSelectNodesAction( ctx );
		modeSelectEdges = new ModeSelectEdgesAction( ctx );
		
		addNodeAction = new AddNodeAction( ctx );
		addEdgeAction = new AddEdgeAction( ctx );
		delNodeAction = new DelNodeAction( ctx );
		delEdgeAction = new DelEdgeAction( ctx );
	}

// WindowListener
	
	public void windowActivated( WindowEvent e )
    {
    }

	public void windowClosed( WindowEvent e )
    {
    }

	public void windowClosing( WindowEvent e )
    {
		System.exit( 0 );
    }

	public void windowDeactivated( WindowEvent e )
    {
    }

	public void windowDeiconified( WindowEvent e )
    {
    }

	public void windowIconified( WindowEvent e )
    {
    }

	public void windowOpened( WindowEvent e )
    {
    }
	
// GraphViewerListener
	
	public void backgroundClicked( float x, float y, int button, boolean clicked )
    {
		if( button == 1 && clicked == false )
		{
			if( ctx.isModeAddNodes() )
			{
				addNodeAction.fire( x, y );
			}
			else if( ctx.isModeSelNodes() )
			{
				ctx.selection.unselectAll();
			}
			else if( ctx.isModeSelEdges() )
			{
				ctx.selection.unselectAll();
			}
		}
    }

	public void nodeMoved( String id, float x, float y, float z )
    {
    }

	public void nodeSelected( String id, boolean selected )
    {
		if( ! selected )	// Here the selected means the mouse release.
		{
			if( ! ctx.isModeAddEdges() && ! ctx.isModeAddNodes() )
				modeSelectNodes.fire();

			ctx.selection.addNodeToListSelection( id );
			
			if( ctx.isModeAddEdges() )
				addEdgeAction.fire();
		}
    }

	public void spriteMoved( String id, float x, float y, float z )
    {
    }

	public void spriteSelected( String id, boolean selected )
    {
		if( ! selected )
		{
			modeSelectEdges.fire();
			ctx.selection.addEdgeToListSelection( id );
		}
    }


// GraphListener

	public void afterNodeAdd( Graph graph, Node node )
	{
	}

	public void afterEdgeAdd( Graph graph, Edge edge )
	{
	}

	public void beforeNodeRemove( Graph graph, Node node )
	{
	}

	public void beforeEdgeRemove( Graph graph, Edge edge )
	{
	}

	public void attributeChanged( Element element, String attribute, Object oldValue, Object newValue )
	{
	}

	public void beforeGraphClear( Graph graph )
	{
	}
	
	protected void updateStatus()
	{
	}

// ListSelectionListener
	
	public void valueChanged( ListSelectionEvent e )
	{
		if( e.getSource() == ctx.ui.lists.nodeList )
		{
			if( e.getValueIsAdjusting() == false )
			{
				if( ! ctx.isModeAddEdges() && ! ctx.isModeAddNodes() )
					ctx.actions.modeSelectNodes.fire();

				int beg = e.getFirstIndex();
				int end = e.getLastIndex();
				
				for( int i=beg; i<=end; i++ )
				{
					String id = ctx.selection.getNode( i );
					ctx.selection.highlightNode( id, ctx.selection.isNodeSelected( i ) );
				}
			}
		}
		else if( e.getSource() == ctx.ui.lists.edgeList )
		{
			if( e.getValueIsAdjusting() == false )
			{
				ctx.actions.modeSelectEdges.fire();

				int beg = e.getFirstIndex();
				int end = e.getLastIndex();
				
				for( int i=beg; i<=end; i++ )
				{
					String id = ctx.selection.getEdge( i );
					ctx.selection.highlightEdge( id, ctx.selection.isEdgeSelected( i ) );
				}
			}			
		}
	}

// Nested classes

	abstract class GraphEdAction extends AbstractAction
	{
	    private static final long serialVersionUID = 1L;

		protected Context ctx;
		
		public GraphEdAction( Context context, String name )
		{
			super( name );
			ctx = context;
		}
	}
	
	class ModeAddNodesAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

	    public ModeAddNodesAction( Context ctx )
	    {
	    	super( ctx, "Add Nodes" );
	    }
	    
		public void actionPerformed( ActionEvent e )
	    {
			ctx.setAddNodesMode();
	    }
	}

	class ModeAddEdgesAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

	    public ModeAddEdgesAction( Context ctx )
	    {
	    	super( ctx, "Add Edges" );
	    }
	    
		public void actionPerformed( ActionEvent e )
	    {
			ctx.setAddEdgesMode();
	    }
	}

	class ModeSelectNodesAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public ModeSelectNodesAction( Context ctx )
		{
			super( ctx, "Sel. Nodes" );
		}

		public void actionPerformed( ActionEvent e )
		{
			fire();
		}
		
		public void fire()
		{
			ctx.setSelNodesMode();			
		}
	}

	class ModeSelectEdgesAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public ModeSelectEdgesAction( Context ctx )
		{
			super( ctx, "Sel. Edges" );
		}

		public void actionPerformed( ActionEvent e )
		{
			fire();
		}
		
		public void fire()
		{
			ctx.setSelEdgesMode();			
		}
	}

	class OpenAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public OpenAction( Context ctx )
		{
			super( ctx, "Open Graph" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			
		}
	}

	class SaveAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public SaveAction( Context ctx )
		{
			super( ctx, "Save Graph" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			
		}
	}

	class AddNodeAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public AddNodeAction( Context ctx )
		{
			super( ctx, "Add Node" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			fire( (float)Math.random(), (float)Math.random() );
		}
		
		public void fire( float x, float y )
		{
			ctx.selection.unselectAll();
			
			String id = ctx.generateNodeId();
			
			Node node = ctx.graph.addNode( id );
			node.addAttribute( "xy", x, y );			
		}
	}

	class AddEdgeAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public AddEdgeAction( Context ctx )
		{
			super( ctx, "Add Edge" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			fire();
		}
		
		public void fire()
		{
			if( ctx.selection.getSelectedNodeCount() > 1 )
			{
				String newId = ctx.generateEdgeId();
				
				ctx.graph.addEdge( newId, ctx.selection.getSelectedNode( 0 ), ctx.selection.getSelectedNode( 1 ) );
				ctx.ui.viewer.step();	// XXX TODO this call is needed since the GraphViewerBase
				// uses a GraphListenerProxy even if the graph and viewer are in the same thread.
				// ?? Add a parameter to GraphViewer() to tell it that the two are in the same
				// thread ???
				Sprite s = ctx.ui.viewer.addSprite( newId );
				s.attachToEdge( newId );
				s.position( 0.5f );
				ctx.selection.unselectAll();
			}			
		}
	}

	class DelNodeAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public DelNodeAction( Context ctx )
		{
			super( ctx, "Del. Node" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			
		}
	}

	class DelEdgeAction extends GraphEdAction
	{
	    private static final long serialVersionUID = 1L;

		public DelEdgeAction( Context ctx )
		{
			super( ctx, "Del. Edge" );
		}
		
		public void actionPerformed( ActionEvent e )
		{
			
		}
	}

	public void stepBegins(Graph graph, double time)
	{
		// TODO Auto-generated method stub
		
	}
}