/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool.graphed;

import java.util.ArrayList;

import org.miv.graphstream.graph.Graph;
import org.miv.graphstream.graph.implementations.MultiGraph;
import org.miv.util.Environment;

/**
 * Values and settings shared across the whole application.
 * 
 * @author Antoine Dutot
 */
public class Context
{
// Attributes

	/**
	 * The four main working modes.
	 */
    public static enum Mode { ADD_NODES, ADD_EDGES, SEL_NODES, SEL_EDGES };
    
    /**
     * The current mode.
     */
    protected Mode mode = Mode.SEL_NODES;

    /**
     * The graph.
     */
    public Graph graph;

    /**
     * The actions of the application.
     */
    public Actions actions;
    
    /**
     * The environment (global parameters).
     */
    public Environment env;

	/**
	 * The user interface.
	 */
	public UI ui;
    
	/**
	 * The nodeSelection.
	 */
    public Selection selection;

    /**
     * Set of listeners.
     */
    public ArrayList<GraphEdListener> listeners = new ArrayList<GraphEdListener>();
    
// Constructors
    
    /**
     * New context.
     */
    public Context( String args[] )
    {
		env       = Environment.getGlobalEnvironment();
		graph     = new MultiGraph();
		actions   = new Actions( this );
		ui        = new UI( this );
		selection = new Selection( this, ui.lists.nodeList, ui.lists.edgeList );
		
		env.readCommandLine( args );
		graph.setAttribute( "stylesheet", styleSheet );
//		env.setParameter( "SwingGraphRenderer.interpolateBounds", "1" );
    }
    
// Access
    
    /**
     * The current mode.
     * @return a mode.
     */
    public Mode getMode()
    {
    	return mode;
    }
    
    public boolean isModeAddNodes()
    {
    	return( mode == Mode.ADD_NODES );
    }
    
    public boolean isModeAddEdges()
    {
    	return( mode == Mode.ADD_EDGES );
    }
    
    public boolean isModeSelNodes()
    {
    	return( mode == Mode.SEL_NODES );
    }
    
    public boolean isModeSelEdges()
    {
    	return( mode == Mode.SEL_EDGES );
    }
    
// Commands

    /**
     * Change the current mode.
     */
    public void setMode( Mode mode )
    {
    	this.mode = mode;

    	switch( mode )
    	{
    		case ADD_NODES:
    			ui.mANButton.setSelected( true );
    			break;
    		case ADD_EDGES:
    			ui.mAEButton.setSelected( true );
    			break;
    		case SEL_NODES:
    			ui.mSNButton.setSelected( true );
    			break;
    		case SEL_EDGES:
    			ui.mSEButton.setSelected( true );
    			break;
    	}
    	
    	for( GraphEdListener listener: listeners )
    		listener.modeChanged();
    }
    
    public void setAddNodesMode()
    {
    	setMode( Mode.ADD_NODES );
    }
    
    public void setAddEdgesMode()
    {
    	setMode( Mode.ADD_EDGES );
    }
    
    public void setSelNodesMode()
    {
    	setMode( Mode.SEL_NODES );
    }
    
    public void setSelEdgesMode()
    {
    	setMode( Mode.SEL_EDGES );
    }
    
	/**
	 * Generate a node identifier according to the node name model.
	 * @return A string that can be used as node identifier.
	 */
	protected String generateNodeId()
	{
		return String.format( "n%d", graph.getNodeCount() );
	}
	
	/**
	 * Generate a edge identifier according to the edge name model.
	 * @return A string that can be used as edge identifier.
	 */
	protected String generateEdgeId()
	{
		return String.format( "e%d", graph.getEdgeCount() );
	}
	
	protected void addListener( GraphEdListener listener )
	{
		listeners.add( listener );
	}
	
	protected void removeListener( GraphEdListener listener )
	{
		int index = listeners.indexOf( listener );
		
		if( index >= 0 )
			listeners.remove( index );
	}
    
// Constants

    /**
     * The default style sheet.
     */
	protected static String styleSheet =
		"graph {" +
		"	padding: 30px;" +
		"}" +
		"node {" +
		"	width: 10px;" +
		"	color: #202020;" +
		"	border-color: black;" +
		"	border-width: 1px;" +
		"}" +
		"node:selected {" +
		"	color: red;" +
		"	border-color: yellow;" +
		"	border-width: 1px;" +
		"}" +
		"edge {" +
		"	color: #404040;" +
		"	border-width: 0px;" +
		"}" +
		"edge:selected {" +
		"	color: red;" +
		"	border-color: yellow;" +
		"	border-width: 1px;" +
		"	width: 3px;" +
		"}" +
		"sprite {" +
		"	border-width: 1px;" +
		"	border-color: #00000080;" +
		"	color: #30303080;" +
		"	width: 8px;" +
		"}";
}