/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.tool.graphed;

import java.util.ArrayList;

import javax.swing.DefaultListModel;
import javax.swing.JList;

import org.miv.graphstream.graph.Element;

/**
 * Handle the currently selected node and edge lists.
 * 
 * <p>
 * The nodeSelection gives the set of nodes and if in "edge nodeSelection mode"
 * the corresponding list of edges that are currently worked on.
 * </p>
 * 
 * <p>
 * The nodeSelection always work on nodes. If GraphEd is in "edge nodeSelection mode",
 * the nodes also determine the selected edge set.
 * </p>
 * 
 * <p>
 * The two JList that show which
 * nodes and edges are selected are also referenced in this class.  
 * </p>
 * 
 * @author Antoine Dutot
 */
public class Selection
{
// Attributes -- GUI
	
	/**
	 * Show the selected nodes.
	 */
    protected JList nodeList;
    
    /**
     * Show the selected edges.
     */
    protected JList edgeList;
    
    protected DefaultListModel nodeListModel;

    protected DefaultListModel edgeListModel;

// Attributes -- Selection

    /**
     * The context
     */
    protected Context ctx;
    
	/**
	 * The set of actually selected nodes.
	 */
	public ArrayList<String> nodeSelection = new ArrayList<String>();

// Constructors
    
	/**
	 * New nodeSelection.
	 * @param context The global context.
	 * @param nodeList The JList that shows the node nodeSelection.
	 * @param edgeList The Jlist that shows the edge nodeSelection.
	 */
    public Selection( Context context, JList nodeList, JList edgeList )
    {
    	this.ctx      = context;
		this.nodeList = nodeList;
		this.edgeList = edgeList;
		
		nodeListModel = (DefaultListModel) nodeList.getModel();
		edgeListModel = (DefaultListModel) edgeList.getModel();
    }
    
// Access
    
    /**
     * Number of nodes in the nodeSelection. 
     */
    public int getSelectedNodeCount()
    {
    	return nodeSelection.size();
    }
    
    /**
     * The i-th element of the nodeSelection (a node id).
     * @param i The element index.
     * @return The identifier of a node.
     */
    public String getSelectedNode( int i )
    {
    	return nodeSelection.get( i );
    }
    
    /**
     * Get the i-th node of the node list.
     * @param i The node index in the list.
     * @return A node identifier.
     */
    public String getNode( int i )
    {
    	return (String) nodeListModel.getElementAt( i );
    }
    
    /**
     * Get the i-th edge of the node list.
     * @param i The edge index in the list.
     * @return A node identifier.
     */
    public String getEdge( int i )
    {
    	return (String) edgeListModel.getElementAt( i );
    }
    
    /**
     * Is the i-th node selected?.
     * @param i The node index in the node list.
     * @return True if the corresponding node is selected.
     */
    public boolean isNodeSelected( int i )
    {
    	return nodeList.isSelectedIndex( i );
    }
    
    /**
     * Is the i-th edge selected?.
     * @param i The edge index in the edge list.
     * @return True if the corresponding edge is selected.
     */
    public boolean isEdgeSelected( int i )
    {
    	return edgeList.isSelectedIndex( i );    	
    }

// Commands
    
	public void addNodeToListSelection( String id )
	{
		if( nodeSelection.indexOf( id ) < 0 )	// XXX Problem is makes it impossible to create loop edges.
			nodeSelection.add( id );

		addToListSelection( nodeList, nodeListModel, id );
	}
	   
	public void addEdgeToListSelection( String id )
	{
		addToListSelection( edgeList, edgeListModel, id );
	}

	public void removeNodeFromListSelection( String id )
	{
		int index = nodeSelection.indexOf( id );
		
		if( index >= 0 )
			nodeSelection.remove( index );
		
		removeFromListSelection( nodeList, nodeListModel, id );
	}
	
	public void removeEdgeFromListSelection( String id )
	{
		removeFromListSelection( edgeList, edgeListModel, id );		
	}
	
	protected void addToListSelection( JList list, DefaultListModel model, String id )
	{
		int oldSelection[] = list.getSelectedIndices();
		int newSelection[] = new int[oldSelection.length+1];
		   
		System.arraycopy( oldSelection, 0, newSelection, 0, oldSelection.length );
		   
		newSelection[oldSelection.length] = model.indexOf( id );
		   
		list.setSelectedIndices( newSelection );
	}
	
	protected void removeFromListSelection( JList list, DefaultListModel model, String id )
	{
		int oldSelection[] = list.getSelectedIndices();
		
		if( oldSelection != null && oldSelection.length > 0 )
		{
			int newSelection[] = new int[oldSelection.length-1];
			int j = 0;
			
			for( int i: oldSelection )
			{
				if( ! model.getElementAt( i ).equals( id ) )
				{
					newSelection[j] = i;
					j++;
				}	
			}
			
			list.setSelectedIndices( newSelection );
		}
	}
	   
	public void highlightNode( String id, boolean selected )
	{
		highlightElement( ctx.graph.getNode( id ), selected );
	}
	
	public void highlightEdge( String id, boolean selected )
	{
		highlightElement( ctx.graph.getEdge( id ), selected );	   
	}
	
	protected void highlightElement( Element e, boolean selected )
	{
		if( e != null )
		{
			if( selected )
			     e.addAttribute( "ui.selected" );
			else e.removeAttribute( "ui.selected" );
		}
	}

	protected void unselectAll()
	{
		nodeSelection.clear();
		nodeList.clearSelection();
		edgeList.clearSelection();
	}
}