/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.ui2.swing;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.util.HashSet;
import java.util.Iterator;

import org.miv.graphstream.graph.Element;
import org.miv.graphstream.ui2.GraphViewer.ScreenshotType;
import org.miv.graphstream.ui2.graphicGraph.GraphicEdge;
import org.miv.graphstream.ui2.graphicGraph.GraphicNode;
import org.miv.graphstream.ui2.graphicGraph.GraphicSprite;
import org.miv.graphstream.ui2.graphicGraph.StyleGroup;
import org.miv.graphstream.ui2.graphicGraph.StyleGroupSet;

/**
 * Graph renderer that tries to be as fast as possible and do not fully implement CSS for speed
 * reasons.
 * 
 * @author Antoine Dutot
 */
public class FastSwingGraphRenderer extends SwingGraphRenderer
{
// Attribute
	
    private static final long serialVersionUID = 1L;
    
// Construction
    
	/**
	 * New "fast" renderer.
	 * @param viewer The parent viewer.
	 */
	public FastSwingGraphRenderer( SwingGraphViewer viewer )
	{
		super( viewer );
		addComponentListener( this );
	}

// Access
	
	@Override
	public ScreenshotType[] getScreenshotTypes()
	{
		ScreenshotType types[] = { ScreenshotType.JPG, ScreenshotType.PNG, ScreenshotType.SVG };
		return types;
	}

// Command
	
	@Override
	public void display( Graphics2D g2 )
	{
		if( fpsInfo )
			fps.beginFrame();
		
		setupG2( g2 );
		renderGraph( g2 );
		renderInfo( g2 );
		
		if( fpsInfo )
			fps.endFrame();
		
		renderFps( g2 );
	}

	@Override
	public void release()
	{
	}

	@Override
	public void screenshot( String fileName )
	{
	}
	
// Rendering

	/**
	 * Setup some rendering hints and general settings in the Swing graphics.
	 */
	protected void setupG2( Graphics2D g2 )
	{
		// It seems that we need to setup the graphics at each frame.
		// The settings are reset between each two frames...
		
		g2.setRenderingHint( RenderingHints.KEY_INTERPOLATION,       RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR );
		g2.setRenderingHint( RenderingHints.KEY_RENDERING,           RenderingHints.VALUE_RENDER_SPEED );
		g2.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING,   RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
	    g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING,        RenderingHints.VALUE_ANTIALIAS_OFF );
	    g2.setRenderingHint( RenderingHints.KEY_COLOR_RENDERING,     RenderingHints.VALUE_COLOR_RENDER_SPEED );
	    g2.setRenderingHint( RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED );
	    g2.setRenderingHint( RenderingHints.KEY_STROKE_CONTROL,      RenderingHints.VALUE_STROKE_PURE );
	}
	
	/**
	 * Render information on the graph.
	 * @param g2 The Swing graphics.
	 */
	protected void renderInfo( Graphics2D g2 )
	{
		
	}
	
	/**
	 * Render the whole graph.
	 * @param g2 The Swing graphics.
	 */
	protected void renderGraph( Graphics2D g2 )
	{
		StyleGroup  style  = graph.getStyle(); 
		Rectangle2D rect   = new Rectangle2D.Float();
		float       px1    = metrics.px1;

		setBackground( graph.getStyle().getColor() );
		camera.pushView( g2 );
		
		if( style.getBorderWidth().value != 0 )
		{
			rect.setFrame( metrics.lo.x, metrics.lo.y+px1, metrics.size.data[0]-px1, metrics.size.data[1]-px1 );
			g2.setStroke( new BasicStroke( metrics.lengthToGu( style.getBorderWidth() ) ) );
			g2.setColor( graph.getStyle().getBorderColor() );
			g2.draw( rect );
		}
		
//		debugVisibleArea( g2 );
		renderGraphElements( g2 );
		camera.popView( g2 );
	}
	
	/**
	 * Render the element of the graph.
	 * @param g2 The Swing graphics.
	 */
	protected void renderGraphElements( Graphics2D g2 )
	{
		StyleGroupSet sgs = graph.getStyleGroups();
		
		Iterator<HashSet<StyleGroup>> z = sgs.getZIterator();
		
		while( z.hasNext() )
		{
			HashSet<StyleGroup> groups = z.next();
			
			for( StyleGroup group: groups )
				renderGroup( g2, group );
		}
	}
	
	/**
	 * Show the frame per second indicator.
	 * @param g2 The Swing graphics.
	 */
	protected void renderFps( Graphics2D g2 )
	{
		if( fpsInfo )
		{
			g2.setColor( Color.GRAY );
			g2.drawString( String.format( "fps %.6f", fps.getAverageFramesPerSecond() ), 10, 15 );
		}
	}

	/**
	 * Render a style group.
	 * @param g2 The Swing graphics.
	 * @param group The group to render.
	 */
	protected void renderGroup( Graphics2D g2, StyleGroup group )
	{
		switch( group.getType() )
		{
			case NODE:
				renderNodeGroup( g2, group );
				break;
			case EDGE:
				renderEdgeGroup( g2, group );
				break;
			case SPRITE:
				renderSpriteGroup( g2, group );
				break;
		}
	}
	
	protected void renderNodeGroup( Graphics2D g2, StyleGroup group )
	{
		setupNodeStyle( g2, group );
		
		Iterator<? extends Element> i = group.getElementIterator();

		Ellipse2D shape = new Ellipse2D.Float();
		float     width = metrics.lengthToGu( group.getWidth() );
		
		while( i.hasNext() )
		{
			GraphicNode node = (GraphicNode) i.next();

			if( metrics.isVisible( node ) )
			{
				float w2 = width/2;
				shape.setFrame( node.x-w2, node.y-w2, width, width );
				g2.fill( shape );
			}
		}
	}
	
	protected void setupNodeStyle( Graphics2D g2, StyleGroup group )
	{
		g2.setColor( group.getColor() );
	}
	
	protected void renderEdgeGroup( Graphics2D g2, StyleGroup group )
	{
		setupEdgeStyle( g2, group );
		
		Iterator<? extends Element> i = group.getElementIterator();

		Line2D shape = new Line2D.Float();
		
		while( i.hasNext() )
		{
			GraphicEdge edge  = (GraphicEdge) i.next();
			GraphicNode node0 = (GraphicNode) edge.getNode0();
			GraphicNode node1 = (GraphicNode) edge.getNode1();
			
			if( metrics.isVisible( node0 ) || metrics.isVisible( node1 ) )
			{
				shape.setLine( node0.x, node0.y, node1.x, node1.y );
				g2.draw( shape );
			}
		}		
	}
	
	protected void setupEdgeStyle( Graphics2D g2, StyleGroup group )
	{
		float width = metrics.lengthToGu( group.getWidth() );

		g2.setColor( group.getColor() );
		g2.setStroke( new BasicStroke( width) );
		
	}

	protected void renderSpriteGroup( Graphics2D g2, StyleGroup group )
	{
		setupSpriteStyle( g2, group );
		
		Iterator<? extends Element> i = group.getElementIterator();

		Ellipse2D shape = new Ellipse2D.Float();
		float     width = metrics.lengthToGu( group.getWidth() );
		
		while( i.hasNext() )
		{
			GraphicSprite sprite = (GraphicSprite) i.next();

			if( metrics.isVisible( sprite ) )
			{
				float w2 = width/2;
				shape.setFrame( sprite.x-w2, sprite.y-w2, width, width );
				g2.fill( shape );
			}
		}		
	}
	
	protected void setupSpriteStyle( Graphics2D g2, StyleGroup group )
	{
		g2.setColor( group.getColor() );
	}
	
// Debug

	/**
	 * Show the centre, the low and high points of the graph, and the visible area (that should
	 * always map to the window borders).
	 */
	protected void debugVisibleArea( Graphics2D g2 )
	{
		Rectangle2D rect = new Rectangle2D.Float();
		
		float x = metrics.loVisible.x;
		float y = metrics.loVisible.y;
		float w = (float) Math.abs( metrics.hiVisible.x - x );
		float h = (float) Math.abs( metrics.hiVisible.y - y );
		
		rect.setFrame( x, y, w, h );
		g2.setStroke( new BasicStroke( metrics.px1 * 4 ) );
		g2.setColor( Color.RED );
		g2.draw( rect );

		g2.setColor( Color.BLUE );
		Ellipse2D ellipse = new Ellipse2D.Float(  );
		float px1 = metrics.px1;
		ellipse.setFrame( camera.center.x-3*px1, camera.center.y-3*px1, px1*6, px1*6 );
		g2.fill( ellipse );
		ellipse.setFrame( metrics.lo.x-3*px1, metrics.lo.y-3*px1, px1*6, px1*6 );
		g2.fill( ellipse );
		ellipse.setFrame( metrics.hi.x-3*px1, metrics.hi.y-3*px1, px1*6, px1*6 );
		g2.fill( ellipse );
	}
}