/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.ui2.swing;

import java.util.Iterator;

import org.miv.graphstream.graph.Node;
import org.miv.graphstream.ui2.graphicGraph.GraphicGraph;
import org.miv.graphstream.ui2.graphicGraph.GraphicNode;
import org.miv.graphstream.ui2.graphicGraph.GraphicSprite;
import org.miv.graphstream.ui2.graphicGraph.stylesheet.Value;
import org.miv.util.geom.Point3;
import org.miv.util.geom.Vector3;

/**
 * Various geometric informations on the graph.
 * 
 * @author Antoine Dutot
 */
public class GraphMetrics
{
// Attribute
	
	/**
	 * Graph lower position (bottom,left,front).
	 */
	public Point3 lo = new Point3();
	
	/**
	 * Graph higher position (top,right,back).
	 */
	public Point3 hi = new Point3(); 
	
	/**
	 * The lowest visible point.
	 */
	public Point3 loVisible = new Point3();
	
	/**
	 * The highest visible point.
	 */
	public Point3 hiVisible = new Point3();
	
	/**
	 * Graph dimension.
	 */
	public Vector3 size = new Vector3();
	
	/**
	 * The graph diagonal.
	 */
	public float diagonal = 1;
	
	/**
	 * The view port size.
	 */
	public Vector3 viewport = new Vector3();
	
	/**
	 * The scaling factor to pass from graph units to pixels.
	 */
	public float ratioPx2Gu;
	
	/**
	 * The length for one pixel, according to the current transformation.
	 */
	public float px1;
	
	/**
	 * The border around the graph.
	 */
	public Value padding;
	
// Construction
	
	public GraphMetrics()
	{
		setDefaults();
	}
	
	/**
	 * Set defaults value in the lo, hi and size fields.
	 */
	protected void setDefaults()
	{
		lo.set( -1, -1, -1 );
		hi.set( 1, 1, 1 );
		size.set( 2, 2, 2 );
		diagonal = 1;
	}
	
// Access
	
	/**
	 * The graph diagonal (the overall width).
	 * @return The diagonal.
	 */
	public float getDiagonal()
	{
		return diagonal;
	}
	
	/**
	 * The graph bounds.
	 * @return The size.
	 */
	public Vector3 getSize()
	{
		return size;
	}
	
	/**
	 * The graph lowest (bottom,left,front) point.
	 * @return The lowest point.
	 */
	public Point3 getLowPoint()
	{
		return lo;
	}
	
	/**
	 * The graph highest (top,right,back) point.
	 * @return The highest point.
	 */
	public Point3 getHighPoint()
	{
		return hi;
	}
	
	/**
	 * True if the node is in the visible area.
	 * @param node The node to test.
	 * @return False if the node is not in the visible area.
	 */
	public boolean isVisible( GraphicNode node )
	{
		float x = node.x;
		float y = node.y;
		
		if( x < loVisible.x )
			return false;
			
		if( x > hiVisible.x )
			return false;
		
		if( y < loVisible.y )
			return false;

		if( y > hiVisible.y )
			return false;
		
		return true;
	}
	
	/**
	 * True if the sprite is in the visible area.
	 * @param sprite The sprite to test.
	 * @return False if the sprite is not in the visible area.
	 */
	public boolean isVisible( GraphicSprite sprite )
	{
		float x = sprite.x;
		float y = sprite.y;
		
		if( x < loVisible.x )
			return false;
			
		if( x > hiVisible.x )
			return false;
		
		if( y < loVisible.y )
			return false;

		if( y > hiVisible.y )
			return false;
		
		return true;
	}
	
// Access -- Convert values
	
	/**
	 * Convert a value in a given units to graph units.
	 * @param value The value to convert (it contains its own units).
	 */
	public float lengthToGu( Value value )
	{
		switch( value.units )
		{
			case PX:
				return (value.value-0.01f) / ratioPx2Gu;
			case PERCENTS:
				return ( diagonal * value.value );
			case GU:
			default:
				return value.value;
		}
	}
	
	/**
	 * Convert a value in pixels to graph units.
	 * @param pixels The value in pixels.
	 * @return The value in graph units.
	 */
	public float pixelsToGu( float pixels )
	{
		return( ( pixels - 0.01f ) / ratioPx2Gu );
	}
	
// Command

	/**
	 * Set the output view port size in pixels. 
	 * @param viewportWidth The width in pixels of the view port.
	 * @param viewportHeight The width in pixels of the view port.
	 */
	public void setViewport( float viewportWidth, float viewportHeight )
	{
		viewport.set( viewportWidth, viewportHeight, 0 );
	}
	
	/**
	 * The ratio to pass by multiplication from pixels to graph units.
	 * @param ratio The ratio.
	 */
	public void setRatioPx2Gu( float ratio )
	{
		ratioPx2Gu = ratio;
		px1        = 0.95f / ratioPx2Gu;
	}
	
	/**
	 * Recompute the graph metrics.
	 * @param graph The graph.
	 */
	public void recompute( GraphicGraph graph )
	{
		computeBounds( graph );
		computeSize( graph );
	}
	
	/**
	 * Compute the graph bounds. 
	 * @param graph The graph.
	 */
	protected void computeBounds( GraphicGraph graph )
	{
		if(  graph.getNodeCount() == 0 && graph.getSpriteCount() == 0 )
		{
			setDefaults();
			return;
		}
		
		Iterator<? extends Node> nodes = graph.getNodeIterator();
		
		lo.x = Float.MAX_VALUE;
		lo.y = Float.MAX_VALUE;
		hi.x = Float.MIN_VALUE;
		hi.y = Float.MIN_VALUE;
		
		while( nodes.hasNext() )
		{
			GraphicNode node = (GraphicNode) nodes.next();
			
			if( node.x < lo.x ) lo.x = node.x;
			if( node.y < lo.y ) lo.y = node.y;

			if( node.x > hi.x ) hi.x = node.x;
			if( node.y > hi.y ) hi.y = node.y;
		}

		Iterator<? extends GraphicSprite> sprites = graph.getSpriteIterator();
		
		while( sprites.hasNext() )
		{
			GraphicSprite sprite = sprites.next();
			
			if( sprite.x < lo.x ) lo.x = sprite.x;
			if( sprite.y < lo.y ) lo.y = sprite.y;

			if( sprite.x > hi.x ) hi.x = sprite.x;
			if( sprite.y > hi.y ) hi.y = sprite.y;
		}
		
		padding = graph.getStyle().getPadding();
	}

	/**
	 * Compute the graph sizes.
	 * @param graph The graph.
	 */
	protected void computeSize( GraphicGraph graph )
	{
		size.data[0] = hi.x - lo.x;
		size.data[1] = hi.y - lo.y;
		diagonal     = (float) Math.sqrt( size.data[0]*size.data[0] + size.data[1] * size.data[1] );
	}
}