/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.ui2.swing;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JPanel;

import org.miv.graphstream.ui2.graphicGraph.GraphicElement;
import org.miv.graphstream.ui2.graphicGraph.GraphicNode;
import org.miv.graphstream.ui2.GraphViewer.ScreenshotType;
import org.miv.graphstream.ui2.graphicGraph.GraphicGraph;
import org.miv.util.geom.Point3;

/**
 * Interchangeable renderer for the swing graph viewer.
 * 
 * @author Antoine Dutot
 */
public abstract class SwingGraphRenderer extends JPanel implements ComponentListener, KeyListener, MouseListener
{
// Attributes

    private static final long serialVersionUID = 1L;
    
    /**
     * The parent viewer.
     */
    protected SwingGraphViewer viewer;

    /**
     * The graphic graph (link to the viewers one).
     */
    protected GraphicGraph graph;
    
    /**
     * The position of the view.
     */
    protected Camera camera;
    
    /**
     * Shortcut to the camera metrics.
     */
    protected GraphMetrics metrics;

    /**
     * Show layout informations ?.
     */
    protected boolean layoutInfo = true;
    
    /**
     * Show frames-per-second informations ?.
     */
    protected boolean fpsInfo = false;
    
    /**
     * Show the current time step.
     */
    protected boolean stepsInfo = false;
    
    /**
     * The quality level.
     */
    protected int quality = 0;

    /**
     * Set to true when the component is not visible to disable rendering.
     */
    protected boolean hidden = false;
    
    /**
     * Set to true each time a redraw is needed.
     */
    protected boolean needRedraw = true;

    /**
     * Count the frames per second.
     */
    protected FpsCounter fps = new FpsCounter();

// Constructor

    /**
     * New renderer.
     * @param viewer The parent viewer.
     */
    public SwingGraphRenderer( SwingGraphViewer viewer )
    {
    	this.viewer  = viewer;
    	this.camera  = new Camera();
    	this.metrics = camera.getMetrics();
    	this.graph   = viewer.getGraph();
    }
    
// Access

    /**
     * The screen shot file types supported.
     */
	public abstract ScreenshotType[] getScreenshotTypes();
	
	/**
	 * The view centre (a point in graph units).
	 * @return The view centre.
	 */
	public Point3 getViewCenter()
	{
		return camera.getViewCenter();
	}
	
	/**
	 * The portion of the graph visible.
	 * @return A real for which value 1 means the graph is fully visible and uses the whole viewport.
	 */
	public float getViewPercent()
	{
		return camera.getViewPercent();
	}
	
	/**
	 * The current rotation angle.
	 * @return The rotation angle in degrees.
	 */
	public float getViewRotation()
	{
		return camera.getViewRotation();
	}
	
	/**
	 * The graph metrics.
	 * @return Metrics.
	 */
	public GraphMetrics getGraphMetrics()
	{
		return metrics;
	}
	
	/**
	 * Are frame-per-second information shown ?.
	 * @return True if the FPS are shown.
	 */
	public boolean getShowFPSInfo()
	{
		return fpsInfo;
	}

// Command
	
	/**
	 * Show or hide the layout completion information.
	 * @param on If true, information is shown.
	 */
	public void setShowLayoutInfo( boolean on )
	{
		layoutInfo = on;
		needRedraw = true;
	}
	
	/**
	 * Show or hide the FPS information.
	 * @param on If true, information is shown.
	 */
	public void setShowFPSInfo( boolean on )
	{
		if( fpsInfo == false && on )
			fps.resetAverages();
		
		fpsInfo = on;
		needRedraw = true;
	}
	
	/**
	 * Show the current time step.
	 * @param on If true show this information.
	 */
	public void setStepsInfo( boolean on )
	{
		stepsInfo = on;
		needRedraw = true;
	}
	
	/**
	 * Set the level of quality [0..4].
	 * @param quality The quality level, one of 0 to 4 integer values.
	 */
	public void setQuality( int quality )
	{
		this.quality = quality;
		needRedraw = true;
	}
	
	/**
	 * Reset the view to the automatic mode.
	 */
	public void resetView()
	{
		camera.setAutoFitView( true );
		camera.setRotation( 0 );
		needRedraw = true;
	}

	/**
	 * Change the view centre.
	 * @param x The new abscissa.
	 * @param y The new ordinate.
	 * @param z The new depth.
	 */
	public void setViewCenter( float x, float y, float z )
	{
		camera.setAutoFitView( false );
		camera.setCenter( x, y /* ignore Z */ );
		needRedraw = true;
	}

	/**
	 * Zoom the view.
	 * @param percent Percent of the graph visible.
	 */
	public void setViewPercent( float percent )
	{
		camera.setAutoFitView( false );
		camera.setZoom( percent );
		needRedraw = true;
	}
	
	/**
	 * Rotate the view around its centre point by a given theta angles (in degrees).
	 * @param theta The rotation angle in degrees.
	 */
	public void setViewRotation( float theta )
	{
		camera.setRotation( theta );
		needRedraw = true;
	}
	
	/**
	 * Store the content of the current frame in a file.
	 * @param fileName The file name.
	 */
	public abstract void screenshot( String fileName );

	/**
	 * Trigger a redraw. The drawing code is not made in the method and therefore you should not need
	 * to override it. The rendering is done in {@link #display(Graphics2D)}. This method only
	 * triggers a redraw using the repaint() method.
	 */
	public void display()
	{
		repaint();
	}
	
	@Override
	public void paintComponent( Graphics g )
	{
		if( hidden )
			return;
		
		if( graph.graphChanged )
		{
			metrics.recompute( graph );
			needRedraw = true;
		}
	
		if( fpsInfo )
			needRedraw = true;	// Force a redraw if in FPS mode.
		
		if( needRedraw )
		{
			super.paintComponent( g );
			Graphics2D g2 = (Graphics2D) g;
			display( g2 );
			
			needRedraw = false;
			graph.graphChanged = false;
		}
	}
	
	/**
	 * Do the rendering of a frame (the graph and additional information). 
	 * @param g2 The graphics of Swing.
	 */
	public abstract void display( Graphics2D g2 );


	/**
	 * Release all used resources. This method is called by the viewer each time
	 * the renderer is changed.
	 */
	public abstract void release();
	
// Events
	
	public void keyPressed( KeyEvent event )
	{
		viewer.getShortcutManager().keyPressed( event );
	}

	public void keyReleased( KeyEvent event )
	{
		viewer.getShortcutManager().keyReleased( event );
	}

	public void keyTyped( KeyEvent event )
	{
		viewer.getShortcutManager().keyTyped( event );
	}

	public void componentHidden( ComponentEvent arg0 )
    {
		hidden = true;
    }

	public void componentShown( ComponentEvent arg0 )
    {
		hidden = false;
    }

	public void componentMoved( ComponentEvent arg0 )
    {
    }

	public void componentResized( ComponentEvent arg0 )
    {
		camera.setViewport( getWidth(), getHeight() );
		needRedraw = true;
    }
	
	public void mouseClicked( MouseEvent e )
    {
    }

	public void mouseEntered( MouseEvent e )
    {
    }

	public void mouseExited( MouseEvent e )
    {
    }

	protected GraphicElement selection;
	protected int mouseStartX = 0;
	protected int mouseStartY = 0;
	
	public void mousePressed( MouseEvent e )
	{
		selection = graph.findNodeOrSprite(
				metrics.pixelsToGu( e.getX() ),
				metrics.pixelsToGu( e.getY() ) );

		if( selection != null )
		{
			mouseStartX = e.getX();
			mouseStartY = e.getY();
			
			if( selection instanceof GraphicNode )
			{
				selection.addAttribute( "ui.clicked" );
				viewer.sendNodeSelectedEvent( selection.getId(), true );
				
				if( viewer.getLayoutRemote() != null )
					viewer.getLayoutRemote().freezeNode( selection.getId(), true );
			}
			else
			{
				selection.addAttribute( "ui.clicked" );
				viewer.sendSpriteSelectedEvent( selection.getId(), true );
			}
		}
		else
		{
			viewer.sendBackgroundClickedEvent(
				metrics.pixelsToGu( e.getX() ),
				metrics.pixelsToGu( e.getY() ),
				e.getButton(), true );
		}
    }

	public void mouseReleased( MouseEvent e )
    {
		if( selection != null )
		{
			if( selection instanceof GraphicNode )
			{
				selection.removeAttribute( "ui.clicked" );
				viewer.sendNodeSelectedEvent( selection.getId(), false );
				
				if( e.getX() != mouseStartX || e.getY() != mouseStartY )
				{
					float oldx = selection.getX();
					float oldy = selection.getY();
				
					selection.move(
						metrics.pixelsToGu( e.getX() ),
						metrics.pixelsToGu( e.getY() ), 0 );
				
					viewer.sendNodeMovedEvent( selection.getId(), selection.getX(), selection.getY(), 0 );
				
					if( viewer.getLayoutRemote() != null )
						viewer.getLayoutRemote().forceMoveNode( selection.getId(),
							selection.getX()-oldx, selection.getY()-oldy, 0 );
				}
			}
			else
			{
				selection.removeAttribute( "ui.clicked" );
				//nodeSelection.move( ctx.pixelsToX( e.getX() ), ctx.pixelsToY( e.getY() ), 0 );
				if( e.getX() != mouseStartX || e.getY() != mouseStartY )
					viewer.sendSpriteMovedEvent( selection.getId(), selection.getX(), selection.getY(), 0 );
				viewer.sendSpriteSelectedEvent( selection.getId(), false );				
			}
			
			selection = null;
		}
		else
		{
			viewer.sendBackgroundClickedEvent(
				metrics.pixelsToGu( e.getX() ),
				metrics.pixelsToGu( e.getY() ),
				e.getButton(), false );
		}
    }

	public void mouseDragged( MouseEvent e )
    {
	    if( selection != null && selection instanceof GraphicNode )
	    {
	    	float oldx = selection.getX();
	    	float oldy = selection.getY();
	    	
			selection.move(
					metrics.pixelsToGu( e.getX() ),
					metrics.pixelsToGu( e.getY() ), 0 );
			
			if( selection instanceof GraphicNode && viewer.getLayoutRemote() != null )
			{
				viewer.getLayoutRemote().forceMoveNode( selection.getId(),
						selection.getX()-oldx, selection.getY()-oldy, 0 );
				viewer.getLayoutRemote().freezeNode( selection.getId(), false );
			}
			
			
	    	viewer.sendNodeMovedEvent( selection.getId(), selection.getX(), selection.getY(), 0 );
	    }
    }
}