/*
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package org.miv.graphstream.ui2.swing;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.Timer;

import org.miv.graphstream.algorithm.layout2.Layout;
import org.miv.graphstream.algorithm.layout2.LayoutRunner;
import org.miv.graphstream.graph.Graph;
import org.miv.graphstream.ui2.GraphViewerBase;
import org.miv.graphstream.ui2.graphicGraph.GraphicGraph;
import org.miv.util.geom.Point3;

/**
 * Implementation of the {@link org.miv.graphstream.ui2.GraphViewer} interface for Swing.
 * 
 * <p>
 * The architecture of the viewer for Swing uses the concept of renderers. There are multiple
 * renderers for the various quality levels. Therefore the role of this class is to create
 * the viewer, associate it with the Swing thread context (so that the viewer runs in the
 * Swing thread) and call graph renderers at regular time intervals.
 * </p>
 * 
 * @author Antoine Dutot
 */
public class SwingGraphViewer extends GraphViewerBase implements ActionListener
{
// Attributes
	
	/**
	 * Optional window to contain the main pane.
	 */
	public JFrame window;
	
	/**
	 * The panel containing everything.
	 */
	public SwingGraphRenderer renderer;
	
	/**
	 * Draw a frame once every while.
	 */
	public Timer timer;
	
	/**
	 * Time allocated for drawing one frame. The timer will fire an
	 * action event that will trigger a redraw every msPerFrame. By
	 * default the value is 40 milliseconds, which makes 25 frames
	 * per second.
	 */
	public int msPerFrame = 40;
	
	/**
	 * The quality levels, driving the renderer to use. There are only
	 * two renderers, the basic one (level 0) and the CSS one that can
	 * use 4 quality levels. 
	 */
	public int quality = 0;
	
	/**
	 * Handle the keyboard and mouse actions.
	 */
	public ShortcutManager shortcutManager;
	
// Constructors
	
	/**
	 * Default constructor that force you to call the {@link #open(Graph)},
	 * {@link #open(Graph, boolean)}, {@link #open(Graph, Layout)} or
	 * {@link #open(Graph, Layout, boolean, boolean)} methods.
	 */
	public SwingGraphViewer()
	{
		// Need to call open().
	}
	
	/**
	 * New graph viewer on the given graph. This creates a graph viewer in its own window.
	 * The {@link #open(Graph)} method is called for you.
	 * @param graph The graph to display.
	 */
	public SwingGraphViewer( Graph graph )
	{
		this( graph, false );
	}
	
	/**
	 * Like {@link #SwingGraphViewer(Graph, boolean)} but allows in addition to specify the layout
	 * class.
	 * @param graph The graph to display.
	 * @param layout The layout class.
	 * @param isPanel If true a the renderer is put in a panel, else a window is created for the renderer.
	 */
	public SwingGraphViewer( Graph graph, Layout layout, boolean isPanel )
	{
		this( graph, layout, isPanel, false );
	}
	
	/**
	 * Like {@link #SwingGraphViewer(Graph, boolean)} but allows in addition to specify if a default
	 * layout should be used.
	 * @param graph The graph to display.
	 * @param autoLayout If true, use the default layout, else no layout is used.
	 * @param isPanel If true a the renderer is put in a panel, else a window is created for the renderer.
	 */
	public SwingGraphViewer( Graph graph, boolean autoLayout, boolean isPanel )
	{
		this( graph, autoLayout, isPanel, false );
	}
	
	/**
	 * New graph viewer on the given graph. This creates a graph viewer in a window or panel. If
	 * the graph viewer is in a panel, you can insert it in your Swing GUI using {@link #getComponent()}.
	 * However remember that due to the fact {@link org.miv.graphstream.ui.GraphViewer} does not
	 * impose any GUI toolkit, you must cast the component to the 'java.awt.Component' class. Also
	 * see {@link #getSwingComponent()}.
	 * @param graph The graph to display.
	 * @param isPanel If true a the renderer is put in a panel, else a window is created for the renderer.
	 */
	public SwingGraphViewer( Graph graph, boolean isPanel )
	{
		open( graph, LayoutRunner.newDefaultLayout(), isPanel, false );
	}

	/**
	 * Like {@link #SwingGraphViewer(Graph, Layout, boolean)} but allows in addition to specify if
	 * the graph is manipulated inside the Swing thread.
	 * @param graph The graph to display.
	 * @param layout The layout class.
	 * @param isPanel If true a the renderer is put in a panel, else a window is created for the renderer.
	 * @param direct Is the graph changed, manipulated, in the Swing thread only.
	 */
	public SwingGraphViewer( Graph graph, Layout layout, boolean isPanel, boolean direct )
	{
		open( graph, layout, isPanel, direct );
	}

	/**
	 * Like {@link #SwingGraphViewer(Graph, boolean, boolean)} but allows in addition to specify if
	 * the graph is manipulated inside the Swing thread.
	 * @param graph The graph to display.
	 * @param autoLayout If true, use the default layout, else no layout is used.
	 * @param isPanel If true a the renderer is put in a panel, else a window is created for the renderer.
	 * @param direct Is the graph changed, manipulated, in the Swing thread only.
	 */
	public SwingGraphViewer( Graph graph, boolean autoLayout, boolean isPanel, boolean direct )
	{
		open( graph, autoLayout, isPanel, direct );
	}
		
	@Override
    public void open( Graph graph )
    {
		open( graph, true );
    }

	@Override
    public void open( Graph graph, boolean autoLayout )
    {
		open( graph, autoLayout ? LayoutRunner.newDefaultLayout() : null );
    }

	@Override
    public void open( Graph graph, Layout layout )
    {
		open( graph, layout, false, false );
    }
	
	public void open( Graph graph, boolean autoLayout, boolean isPanel, boolean direct )
	{
		open( graph, autoLayout ? LayoutRunner.newDefaultLayout() : null, isPanel, direct );
	}

	/**
	 * Open a new viewer with the given input graph. You can specify an instance of layout
	 * algorithm. This parameter can be null to disable the layout thread. It is possible
	 * to activate the creation of a window around the viewer panel is "isPanel" is false.
	 * You specify that the input graph is in the same context (thread) that this viewer
	 * with the "direct" boolean argument. The thread of this viewer is necessarily the swing
	 * thread.
	 * @param graph The input graph to observe and render.
	 * @param layout The layout algorithm to use, pass null to disable layout.
	 * @param isPanel If false a top level window is created.
	 * @param direct If true the input graph is considered in the same thread than this rendered.
	 */
	public void open( Graph graph, Layout layout, boolean isPanel, boolean direct )
	{
		init( graph, direct );
		
		if( layout != null )
			setLayout( layout );
		
		if( ! isPanel )
			window = new JFrame( "GraphStream viewer" );
		
		initPane();
		startTimer();
    }

	@Override
    public void close()
    {
    	super.close();
		timer.stop();
		
		if( window != null )
			window.setVisible( false );
    }

	/**
	 * Initialise the swing GUI around the viewer, eventually create a top level window if needed.
	 */
	protected void initPane()
	{
		shortcutManager = new ShortcutManager( this );
		renderer        = createRenderer();
		timer           = new Timer( msPerFrame, this );

		renderer.setShowLayoutInfo( true );
		
		if( window != null )
			window.addKeyListener( renderer );
		
		if( window != null )
		{
			ImageIcon icon = createImageIcon( "GraphStreamSmallLogo24.png", "GraphStream" );

			window.add( renderer );
			window.setSize( 300, 300 );
			window.setMinimumSize( new Dimension( 200, 200 ) );
			window.setVisible( true );
			window.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
			window.setIconImage( icon.getImage() );
		}
	}
	
	/**
	 * Create and launch the timer that will trigger redraws at regular intervals.
	 */
	protected void startTimer()
	{
		timer = new Timer( msPerFrame, this );

		timer.setCoalesce( true );
		timer.setDelay( msPerFrame );
		timer.setRepeats( true );
		timer.start();
	}
	
// Access
	
	public Object getComponent()
	{
		if( window != null )
			return window;
		
		return renderer;
	}
	
	/**
	 * Like the {@link #getComponent()} method inherited from {@link org.miv.graphstream.ui.GraphViewer},
	 * but returns a swing component. This allows to easily put your renderer into your own GUIs.
	 * @return The Swing component containing the renderer.
	 */
	public Component getSwingComponent()
	{
		if( window != null )
			return window;
	
		return renderer;
	}
	
	public SwingGraphRenderer getRenderer()
	{
		return renderer;
	}
	
	public GraphicGraph getGraph()
	{
		return graph;
	}
	
	/**
	 * The manager for shortcuts and various actions.
	 * @return The shortcut manager.
	 */
	public ShortcutManager getShortcutManager()
	{
		return shortcutManager;
	}

	public ScreenshotType[] getScreenShotTypes()
    {
	    return renderer.getScreenshotTypes();
    }
	
	/**
	 * Create a graph renderer according to the quality level.
	 * @return A new renderer for the given quality.
	 */
	protected SwingGraphRenderer createRenderer()
	{
		switch( quality )
		{
			case 0:
				return new FastSwingGraphRenderer( this );
			case 1:
			case 2:
			case 3:
			case 4:
			default:
				SwingGraphRenderer r = new CssSwingGraphRenderer( this );
				r.setQuality( quality );
				return r;
		}
	}

	public Point3 getViewCenter()
	{
		return renderer.getViewCenter();
	}
	
	public float getViewPercent()
	{
		return renderer.getViewPercent();
	}
	
	public float getViewRotation()
	{
		return renderer.getViewRotation();
	}

	/**
	 * Is the FPS mode on. When on, the viewer tries to draw as many frames as possible and shows
	 * a "frame-per-second" indicator to measure the performance.
	 * @return True if the FPS mode is on.
	 */
	public boolean getModeFPS()
	{
		return renderer.getShowFPSInfo();
	}
	
// Command

	public void actionPerformed( ActionEvent e )
	{
		if( e.getSource() == timer )
		{
			step();
			renderer.display();
		}
	}
	
	@Override
    public void setQuality( int quality )
    {
		if( ( quality == 0 && this.quality != 0 )
		 || ( quality != 0 && this.quality == 0 ) )
		{
			renderer.release();
			this.quality = quality;
			renderer = createRenderer();
		}

		this.quality = quality;
    	renderer.setQuality( quality );
    }
	
	@Override
	public void setStepsVisible( boolean on )
	{
		renderer.setStepsInfo( on );
	}

	@Override
    public void resetView()
    {
		renderer.resetView();
    }

	@Override
    public void setViewCenter( float x, float y, float z )
    {
		renderer.setViewCenter( x, y, z );
    }

	@Override
    public void setViewPercent( float percent )
    {
		renderer.setViewPercent( percent );
    }
	
	@Override
	public void setViewRotation( float theta, float phi )
	{
		renderer.setViewRotation( theta );
	}
	
	/**
	 * If true, the viewer tries to display the most frames per second possible, else a timed mode
	 * is used.
	 * @param on If true, no timer is used and the viewer tries to display as most frames as possible.
	 */
	public void setModeFPS( boolean on )
	{
		renderer.setShowFPSInfo( on );
		
		if( on )
		     timer.setDelay( 1 );
		else timer.setDelay( msPerFrame );
	}

	@Override
	public void screenShot( String fileName, ScreenshotType type )
    {
		switch( type )
		{
			case BMP:
				renderer.screenshot( fileName+".bmp" );
				break;
			case EPS:
				renderer.screenshot( fileName+".eps" );
				break;
			case JPG:
				renderer.screenshot( fileName+".jpg" );
				break;
			case PDF:
				renderer.screenshot( fileName+".pdf" );
				break;
			case PNG:
				renderer.screenshot( fileName+".png" );
				break;
			case SVG:
				renderer.screenshot( fileName+".svg" );
				break;
		}
    }

	/**
	 * Create an icon that must be in the org.miv.graphstream.ui2.swing package.
	 * @param path The path starting from this class package.
	 * @param description The description of the icon.
	 * @return The new icon or null if the image cannot be created. 
	 */
	protected ImageIcon createImageIcon( String path, String description )
	{
	    java.net.URL imgURL = getClass().getResource( path );
	    
	    if( imgURL != null )
	         return new ImageIcon( imgURL, description );
	    else return new ImageIcon( path, description );
	}
}