package toools.io.file.nbs;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.RandomAccessFile;

import it.unimi.dsi.fastutil.longs.LongIterable;
import toools.io.BinaryReader;
import toools.io.Bits;
import toools.io.IORuntimeException;
import toools.io.Utilities;
import toools.io.file.Directory;
import toools.io.file.RegularFile;
import toools.math.MathsUtilities;
import toools.progression.LongProcess;
import toools.thread.MultiThreadProcessing.ThreadSpecifics;
import toools.thread.ParallelIntervalProcessing;
import toools.util.Conversion;

public class NBSFile extends RegularFile implements LongIterable
{

	public static interface ValueExtractor
	{
		long extract(int l);
	}

	public NBSFile(String name)
	{
		super(name);
	}

	public NBSFile(Directory parent, String name)
	{
		super(parent, name);
	}

	public void saveValues(long[] values)
	{
		long max = MathsUtilities.max(values);
		int encoding = Utilities.getNbBytesRequireToEncode(max);

		saveValues(values.length, new ValueExtractor()
		{

			@Override
			public long extract(int l)
			{
				return values[l];
			}
		}, encoding);
	}

	public void saveValues(int[] values)
	{
		int max = MathsUtilities.max(values);
		int encoding = Utilities.getNbBytesRequireToEncode(max);
		saveValues(values.length, new ValueExtractor()
		{
			@Override
			public long extract(int l)
			{
				return values[l];
			}
		}, encoding);
	}

	public void saveValues(int nbItems, ValueExtractor f, int encoding)
	{
		LongProcess writing = new LongProcess(
				"writing " + this + " (" + encoding + "-byte encoding)", "B",
				nbItems * encoding);

		try
		{
			OutputStream os = createWritingStream();
			NBSWriter w = new NBSWriter(os, encoding);

			for (int l = 0; l < nbItems; ++l)
			{
				long value = f.extract(l);
				w.write(value);
				writing.sensor.progressStatus += encoding;
			}

			os.close();
			writing.end();
		}
		catch (IOException e)
		{
			throw new IORuntimeException(e);
		}
	}

	public int readEncoding()
	{
		try
		{
			InputStream is = createReadingStream();
			int encoding = is.read();
			is.close();
			return encoding;
		}
		catch (IOException e)
		{
			throw new IORuntimeException(e);
		}
	}

	public long getNbValues()
	{
		return (getSize() - 1) / readEncoding();
	}

	public long readValue(long i) throws IOException
	{
		return readConsecutiveValues(i, 1)[0];
	}

	public long[] readConsecutiveValues(long startEntry, int n) throws IOException
	{
		InputStream is = createReadingStream(0);
		int encoding = is.read();
		is.skip(startEntry * encoding);
		BinaryReader reader = new BinaryReader(is, n * 8);
		long[] r = new long[n];

		for (int i = 0; i < n; ++i)
		{
			r[i] = reader.next(encoding);
		}

		is.close();
		return r;
	}

	public void writeConsecutiveValues(long startIndex, long[] values, int offset,
			int len)
	{
		if (startIndex < 0)
			throw new IllegalArgumentException("" + startIndex);

		if ( ! exists())
		{
			saveValues(0, null, 8);
		}

		int encoding = readEncoding();

		try
		{
			byte[] b = new byte[values.length * encoding];

			for (int i = 0; i < len; ++i)
			{
				Bits.putLong(b, i * encoding, values[i + offset], encoding);
			}

			RandomAccessFile raf = new RandomAccessFile(getPath(), "rw");
			raf.seek(1 + startIndex * encoding);
			raf.write(b);
			raf.close();
		}
		catch (IOException e)
		{
			throw new IORuntimeException(e);
		}
	}

	public long[] readValues(int nbThreads)
	{
		int nbValues = Conversion.long2int(getNbValues());
		long[] r = new long[Conversion.long2int(nbValues)];
		LongProcess reading = new LongProcess("reading " + this, "B", getSize());

		new ParallelIntervalProcessing(nbValues, nbThreads, reading)
		{
			@Override
			protected void process(ThreadSpecifics s, int lowerBound, int upperBound)
					throws Throwable
			{
				InputStream is = createReadingStream(0);
				int encoding = is.read();

				is.skip(lowerBound * encoding);
				BinaryReader br = new BinaryReader(is, 65536 * 256);

				for (int v = lowerBound; v < upperBound; ++v)
				{
					r[v] = br.next(encoding);
					s.progressStatus += encoding;
				}
			}
		};

		reading.end();
		return r;
	}

	public void decode(PrintStream out, int start, int len, String sep) throws IOException
	{
		InputStream is = createReadingStream();
		int encoding = is.read();
		out.println("values are encoded on " + encoding + " bytes");
		is.skip(start * encoding);

		BinaryReader r = new BinaryReader(is, 65536 * 256);
		int nbValuesInFile = (int) getSize() - 1 / encoding;
		len = Math.min(nbValuesInFile - start, len);

		for (int v = start; v < start + len; ++v)
		{
			long value = r.next(encoding);
			out.print(v + " \t" + value + sep);
		}
	}

	public static void main(String[] args) throws IOException
	{
		for (String filename : args)
		{
			System.out.println("Filename: " + filename);
			NBSFile f = new NBSFile(Directory.getCurrentDirectory(), filename);
			System.out.println("encoding: " + f.readEncoding() + " bytes");
			System.out.println("#entries: " + f.getNbValues());
			f.decode(System.out, 0, 10, "\n");
			System.out.println();
		}
	}

	@Override
	public NBSFileIterator iterator()
	{
		try
		{
			return new NBSFileIterator(this, 0);
		}
		catch (IOException e)
		{
			throw new IORuntimeException(e);
		}
	}
}
